/****************************************************************************
** Meta object code from reading C++ file 'radiobrowsercontroller.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../../fooyin/src/plugins/radiobrowser/radiobrowsercontroller.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'radiobrowsercontroller.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin12RadioBrowser22RadioBrowserControllerE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::RadioBrowser::RadioBrowserController::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin12RadioBrowser22RadioBrowserControllerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::RadioBrowser::RadioBrowserController",
        "searchStarted",
        "",
        "query",
        "searchFailed",
        "error",
        "stationsChanged",
        "Fooyin::RadioBrowser::RadioStationList",
        "stations",
        "reset",
        "stationActionFailed",
        "context",
        "Fooyin::RadioBrowser::RadioStation",
        "station",
        "stationsResolved",
        "resolveId",
        "Fooyin::TrackList",
        "tracks",
        "stationValidated",
        "stationValidationFailed",
        "stationUrlLookupFinished",
        "stationUrlLookupFailed",
        "savedStationsBrowsingChanged",
        "browsing",
        "savedStationsChanged",
        "savedStationsImportFinished",
        "Fooyin::RadioBrowser::SavedStationImportResult",
        "result",
        "savedStationsImportFailed",
        "savedSearchesChanged",
        "Fooyin::RadioBrowser::RadioSavedSearchList",
        "searches",
        "latestSearchChanged",
        "Fooyin::RadioBrowser::RadioSearchRequest",
        "request",
        "searchRequestActivated",
        "currentStationChanged",
        "categoriesStarted",
        "Fooyin::RadioBrowser::RadioCategoryType",
        "type",
        "categoriesChanged",
        "Fooyin::RadioBrowser::RadioCategoryList",
        "categories",
        "categoriesFailed"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'searchStarted'
        QtMocHelpers::SignalData<void(const QString &)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 3 },
        }}),
        // Signal 'searchFailed'
        QtMocHelpers::SignalData<void(const QString &, const QString &)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 3 }, { QMetaType::QString, 5 },
        }}),
        // Signal 'stationsChanged'
        QtMocHelpers::SignalData<void(const Fooyin::RadioBrowser::RadioStationList &, bool)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 7, 8 }, { QMetaType::Bool, 9 },
        }}),
        // Signal 'stationActionFailed'
        QtMocHelpers::SignalData<void(QObject *, const Fooyin::RadioBrowser::RadioStation &, const QString &)>(10, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QObjectStar, 11 }, { 0x80000000 | 12, 13 }, { QMetaType::QString, 5 },
        }}),
        // Signal 'stationsResolved'
        QtMocHelpers::SignalData<void(int, QObject *, const Fooyin::TrackList &)>(14, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 15 }, { QMetaType::QObjectStar, 11 }, { 0x80000000 | 16, 17 },
        }}),
        // Signal 'stationValidated'
        QtMocHelpers::SignalData<void(QObject *, const Fooyin::RadioBrowser::RadioStation &)>(18, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QObjectStar, 11 }, { 0x80000000 | 12, 13 },
        }}),
        // Signal 'stationValidationFailed'
        QtMocHelpers::SignalData<void(QObject *, const QString &)>(19, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QObjectStar, 11 }, { QMetaType::QString, 5 },
        }}),
        // Signal 'stationUrlLookupFinished'
        QtMocHelpers::SignalData<void(QObject *, const Fooyin::RadioBrowser::RadioStationList &)>(20, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QObjectStar, 11 }, { 0x80000000 | 7, 8 },
        }}),
        // Signal 'stationUrlLookupFailed'
        QtMocHelpers::SignalData<void(QObject *, const QString &)>(21, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QObjectStar, 11 }, { QMetaType::QString, 5 },
        }}),
        // Signal 'savedStationsBrowsingChanged'
        QtMocHelpers::SignalData<void(bool)>(22, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 23 },
        }}),
        // Signal 'savedStationsChanged'
        QtMocHelpers::SignalData<void(const Fooyin::RadioBrowser::RadioStationList &)>(24, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 7, 8 },
        }}),
        // Signal 'savedStationsImportFinished'
        QtMocHelpers::SignalData<void(QObject *, Fooyin::RadioBrowser::SavedStationImportResult)>(25, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QObjectStar, 11 }, { 0x80000000 | 26, 27 },
        }}),
        // Signal 'savedStationsImportFailed'
        QtMocHelpers::SignalData<void(QObject *, const QString &)>(28, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QObjectStar, 11 }, { QMetaType::QString, 5 },
        }}),
        // Signal 'savedSearchesChanged'
        QtMocHelpers::SignalData<void(const Fooyin::RadioBrowser::RadioSavedSearchList &)>(29, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 30, 31 },
        }}),
        // Signal 'latestSearchChanged'
        QtMocHelpers::SignalData<void(const Fooyin::RadioBrowser::RadioSearchRequest &)>(32, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 33, 34 },
        }}),
        // Signal 'searchRequestActivated'
        QtMocHelpers::SignalData<void(const Fooyin::RadioBrowser::RadioSearchRequest &)>(35, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 33, 34 },
        }}),
        // Signal 'currentStationChanged'
        QtMocHelpers::SignalData<void(const Fooyin::RadioBrowser::RadioStation &)>(36, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 12, 13 },
        }}),
        // Signal 'categoriesStarted'
        QtMocHelpers::SignalData<void(Fooyin::RadioBrowser::RadioCategoryType)>(37, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 38, 39 },
        }}),
        // Signal 'categoriesChanged'
        QtMocHelpers::SignalData<void(Fooyin::RadioBrowser::RadioCategoryType, const Fooyin::RadioBrowser::RadioCategoryList &)>(40, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 38, 39 }, { 0x80000000 | 41, 42 },
        }}),
        // Signal 'categoriesFailed'
        QtMocHelpers::SignalData<void(Fooyin::RadioBrowser::RadioCategoryType, const QString &)>(43, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 38, 39 }, { QMetaType::QString, 5 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<RadioBrowserController, qt_meta_tag_ZN6Fooyin12RadioBrowser22RadioBrowserControllerE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::RadioBrowser::RadioBrowserController::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin12RadioBrowser22RadioBrowserControllerE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin12RadioBrowser22RadioBrowserControllerE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin12RadioBrowser22RadioBrowserControllerE_t>.metaTypes,
    nullptr
} };

void Fooyin::RadioBrowser::RadioBrowserController::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<RadioBrowserController *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->searchStarted((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 1: _t->searchFailed((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 2: _t->stationsChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioStationList>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 3: _t->stationActionFailed((*reinterpret_cast<std::add_pointer_t<QObject*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioStation>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[3]))); break;
        case 4: _t->stationsResolved((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QObject*>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[3]))); break;
        case 5: _t->stationValidated((*reinterpret_cast<std::add_pointer_t<QObject*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioStation>>(_a[2]))); break;
        case 6: _t->stationValidationFailed((*reinterpret_cast<std::add_pointer_t<QObject*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 7: _t->stationUrlLookupFinished((*reinterpret_cast<std::add_pointer_t<QObject*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioStationList>>(_a[2]))); break;
        case 8: _t->stationUrlLookupFailed((*reinterpret_cast<std::add_pointer_t<QObject*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 9: _t->savedStationsBrowsingChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 10: _t->savedStationsChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioStationList>>(_a[1]))); break;
        case 11: _t->savedStationsImportFinished((*reinterpret_cast<std::add_pointer_t<QObject*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::SavedStationImportResult>>(_a[2]))); break;
        case 12: _t->savedStationsImportFailed((*reinterpret_cast<std::add_pointer_t<QObject*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 13: _t->savedSearchesChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioSavedSearchList>>(_a[1]))); break;
        case 14: _t->latestSearchChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioSearchRequest>>(_a[1]))); break;
        case 15: _t->searchRequestActivated((*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioSearchRequest>>(_a[1]))); break;
        case 16: _t->currentStationChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioStation>>(_a[1]))); break;
        case 17: _t->categoriesStarted((*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioCategoryType>>(_a[1]))); break;
        case 18: _t->categoriesChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioCategoryType>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioCategoryList>>(_a[2]))); break;
        case 19: _t->categoriesFailed((*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioCategoryType>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 2:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::RadioBrowser::RadioStationList >(); break;
            }
            break;
        case 3:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::RadioBrowser::RadioStation >(); break;
            }
            break;
        case 4:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 2:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 5:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::RadioBrowser::RadioStation >(); break;
            }
            break;
        case 7:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::RadioBrowser::RadioStationList >(); break;
            }
            break;
        case 10:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::RadioBrowser::RadioStationList >(); break;
            }
            break;
        case 11:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::RadioBrowser::SavedStationImportResult >(); break;
            }
            break;
        case 13:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::RadioBrowser::RadioSavedSearchList >(); break;
            }
            break;
        case 16:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::RadioBrowser::RadioStation >(); break;
            }
            break;
        case 18:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::RadioBrowser::RadioCategoryList >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(const QString & )>(_a, &RadioBrowserController::searchStarted, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(const QString & , const QString & )>(_a, &RadioBrowserController::searchFailed, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(const Fooyin::RadioBrowser::RadioStationList & , bool )>(_a, &RadioBrowserController::stationsChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(QObject * , const Fooyin::RadioBrowser::RadioStation & , const QString & )>(_a, &RadioBrowserController::stationActionFailed, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(int , QObject * , const Fooyin::TrackList & )>(_a, &RadioBrowserController::stationsResolved, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(QObject * , const Fooyin::RadioBrowser::RadioStation & )>(_a, &RadioBrowserController::stationValidated, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(QObject * , const QString & )>(_a, &RadioBrowserController::stationValidationFailed, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(QObject * , const Fooyin::RadioBrowser::RadioStationList & )>(_a, &RadioBrowserController::stationUrlLookupFinished, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(QObject * , const QString & )>(_a, &RadioBrowserController::stationUrlLookupFailed, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(bool )>(_a, &RadioBrowserController::savedStationsBrowsingChanged, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(const Fooyin::RadioBrowser::RadioStationList & )>(_a, &RadioBrowserController::savedStationsChanged, 10))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(QObject * , Fooyin::RadioBrowser::SavedStationImportResult )>(_a, &RadioBrowserController::savedStationsImportFinished, 11))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(QObject * , const QString & )>(_a, &RadioBrowserController::savedStationsImportFailed, 12))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(const Fooyin::RadioBrowser::RadioSavedSearchList & )>(_a, &RadioBrowserController::savedSearchesChanged, 13))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(const Fooyin::RadioBrowser::RadioSearchRequest & )>(_a, &RadioBrowserController::latestSearchChanged, 14))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(const Fooyin::RadioBrowser::RadioSearchRequest & )>(_a, &RadioBrowserController::searchRequestActivated, 15))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(const Fooyin::RadioBrowser::RadioStation & )>(_a, &RadioBrowserController::currentStationChanged, 16))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(Fooyin::RadioBrowser::RadioCategoryType )>(_a, &RadioBrowserController::categoriesStarted, 17))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(Fooyin::RadioBrowser::RadioCategoryType , const Fooyin::RadioBrowser::RadioCategoryList & )>(_a, &RadioBrowserController::categoriesChanged, 18))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserController::*)(Fooyin::RadioBrowser::RadioCategoryType , const QString & )>(_a, &RadioBrowserController::categoriesFailed, 19))
            return;
    }
}

const QMetaObject *Fooyin::RadioBrowser::RadioBrowserController::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Fooyin::RadioBrowser::RadioBrowserController::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin12RadioBrowser22RadioBrowserControllerE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int Fooyin::RadioBrowser::RadioBrowserController::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 20)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 20;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 20)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 20;
    }
    return _id;
}

// SIGNAL 0
void Fooyin::RadioBrowser::RadioBrowserController::searchStarted(const QString & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void Fooyin::RadioBrowser::RadioBrowserController::searchFailed(const QString & _t1, const QString & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1, _t2);
}

// SIGNAL 2
void Fooyin::RadioBrowser::RadioBrowserController::stationsChanged(const Fooyin::RadioBrowser::RadioStationList & _t1, bool _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1, _t2);
}

// SIGNAL 3
void Fooyin::RadioBrowser::RadioBrowserController::stationActionFailed(QObject * _t1, const Fooyin::RadioBrowser::RadioStation & _t2, const QString & _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1, _t2, _t3);
}

// SIGNAL 4
void Fooyin::RadioBrowser::RadioBrowserController::stationsResolved(int _t1, QObject * _t2, const Fooyin::TrackList & _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1, _t2, _t3);
}

// SIGNAL 5
void Fooyin::RadioBrowser::RadioBrowserController::stationValidated(QObject * _t1, const Fooyin::RadioBrowser::RadioStation & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1, _t2);
}

// SIGNAL 6
void Fooyin::RadioBrowser::RadioBrowserController::stationValidationFailed(QObject * _t1, const QString & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1, _t2);
}

// SIGNAL 7
void Fooyin::RadioBrowser::RadioBrowserController::stationUrlLookupFinished(QObject * _t1, const Fooyin::RadioBrowser::RadioStationList & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1, _t2);
}

// SIGNAL 8
void Fooyin::RadioBrowser::RadioBrowserController::stationUrlLookupFailed(QObject * _t1, const QString & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 8, nullptr, _t1, _t2);
}

// SIGNAL 9
void Fooyin::RadioBrowser::RadioBrowserController::savedStationsBrowsingChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 9, nullptr, _t1);
}

// SIGNAL 10
void Fooyin::RadioBrowser::RadioBrowserController::savedStationsChanged(const Fooyin::RadioBrowser::RadioStationList & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 10, nullptr, _t1);
}

// SIGNAL 11
void Fooyin::RadioBrowser::RadioBrowserController::savedStationsImportFinished(QObject * _t1, Fooyin::RadioBrowser::SavedStationImportResult _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 11, nullptr, _t1, _t2);
}

// SIGNAL 12
void Fooyin::RadioBrowser::RadioBrowserController::savedStationsImportFailed(QObject * _t1, const QString & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 12, nullptr, _t1, _t2);
}

// SIGNAL 13
void Fooyin::RadioBrowser::RadioBrowserController::savedSearchesChanged(const Fooyin::RadioBrowser::RadioSavedSearchList & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 13, nullptr, _t1);
}

// SIGNAL 14
void Fooyin::RadioBrowser::RadioBrowserController::latestSearchChanged(const Fooyin::RadioBrowser::RadioSearchRequest & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 14, nullptr, _t1);
}

// SIGNAL 15
void Fooyin::RadioBrowser::RadioBrowserController::searchRequestActivated(const Fooyin::RadioBrowser::RadioSearchRequest & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 15, nullptr, _t1);
}

// SIGNAL 16
void Fooyin::RadioBrowser::RadioBrowserController::currentStationChanged(const Fooyin::RadioBrowser::RadioStation & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 16, nullptr, _t1);
}

// SIGNAL 17
void Fooyin::RadioBrowser::RadioBrowserController::categoriesStarted(Fooyin::RadioBrowser::RadioCategoryType _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 17, nullptr, _t1);
}

// SIGNAL 18
void Fooyin::RadioBrowser::RadioBrowserController::categoriesChanged(Fooyin::RadioBrowser::RadioCategoryType _t1, const Fooyin::RadioBrowser::RadioCategoryList & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 18, nullptr, _t1, _t2);
}

// SIGNAL 19
void Fooyin::RadioBrowser::RadioBrowserController::categoriesFailed(Fooyin::RadioBrowser::RadioCategoryType _t1, const QString & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 19, nullptr, _t1, _t2);
}
QT_WARNING_POP
