/****************************************************************************
** Meta object code from reading C++ file 'playerdefs.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../fooyin/include/core/player/playerdefs.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'playerdefs.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin6PlayerE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::Player::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin6PlayerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::Player",
        "PlayState",
        "Playing",
        "Paused",
        "Stopped",
        "AdvanceReason",
        "Unknown",
        "StartupRestore",
        "ManualSelection",
        "ManualNext",
        "ManualPrevious",
        "NaturalEnd",
        "PreparedCrossfade",
        "PreparedCommit",
        "PreparedCrossfadeCommit"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'PlayState'
        QtMocHelpers::EnumData<PlayState>(1, 1, QMC::EnumIsScoped).add({
            {    2, PlayState::Playing },
            {    3, PlayState::Paused },
            {    4, PlayState::Stopped },
        }),
        // enum 'AdvanceReason'
        QtMocHelpers::EnumData<AdvanceReason>(5, 5, QMC::EnumIsScoped).add({
            {    6, AdvanceReason::Unknown },
            {    7, AdvanceReason::StartupRestore },
            {    8, AdvanceReason::ManualSelection },
            {    9, AdvanceReason::ManualNext },
            {   10, AdvanceReason::ManualPrevious },
            {   11, AdvanceReason::NaturalEnd },
            {   12, AdvanceReason::PreparedCrossfade },
            {   13, AdvanceReason::PreparedCommit },
            {   14, AdvanceReason::PreparedCrossfadeCommit },
        }),
    };
    return QtMocHelpers::metaObjectData<void, qt_meta_tag_ZN6Fooyin6PlayerE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}

static constexpr auto qt_staticMetaObjectContent_ZN6Fooyin6PlayerE =
    Fooyin::Player::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin6PlayerE_t>();
static constexpr auto qt_staticMetaObjectStaticContent_ZN6Fooyin6PlayerE =
    qt_staticMetaObjectContent_ZN6Fooyin6PlayerE.staticData;
static constexpr auto qt_staticMetaObjectRelocatingContent_ZN6Fooyin6PlayerE =
    qt_staticMetaObjectContent_ZN6Fooyin6PlayerE.relocatingData;

Q_CONSTINIT const QMetaObject Fooyin::Player::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent_ZN6Fooyin6PlayerE.stringdata,
    qt_staticMetaObjectStaticContent_ZN6Fooyin6PlayerE.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent_ZN6Fooyin6PlayerE.metaTypes,
    nullptr
} };

QT_WARNING_POP
