/****************************************************************************
** Meta object code from reading C++ file 'enginecontroller.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../fooyin/include/core/engine/enginecontroller.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'enginecontroller.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin16EngineControllerE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::EngineController::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin16EngineControllerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::EngineController",
        "outputChanged",
        "",
        "output",
        "device",
        "deviceChanged",
        "engineError",
        "error",
        "engineStateChanged",
        "Fooyin::Engine::PlaybackState",
        "state",
        "audiblePauseDrainStarted",
        "audiblePauseDrainCompleted",
        "trackStatusContextChanged",
        "Fooyin::Engine::TrackStatusContext",
        "context",
        "levelReady",
        "Fooyin::LevelFrame",
        "frame",
        "pcmReady",
        "Fooyin::PcmFrame",
        "trackChanged",
        "Fooyin::Track",
        "track",
        "trackCommitted",
        "Fooyin::Engine::TrackCommitContext",
        "trackAboutToFinish",
        "Fooyin::Engine::AboutToFinishContext",
        "trackReadyToSwitch",
        "trackBoundaryReached",
        "nextTrackReadiness",
        "Fooyin::Engine::PlaybackItem",
        "item",
        "ready",
        "uint64_t",
        "requestId",
        "preparedCrossfadeArmResult",
        "generation",
        "armed",
        "preparedGaplessArmResult",
        "finished"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'outputChanged'
        QtMocHelpers::SignalData<void(const QString &, const QString &)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 3 }, { QMetaType::QString, 4 },
        }}),
        // Signal 'deviceChanged'
        QtMocHelpers::SignalData<void(const QString &)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 4 },
        }}),
        // Signal 'engineError'
        QtMocHelpers::SignalData<void(const QString &)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 7 },
        }}),
        // Signal 'engineStateChanged'
        QtMocHelpers::SignalData<void(Fooyin::Engine::PlaybackState)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 9, 10 },
        }}),
        // Signal 'audiblePauseDrainStarted'
        QtMocHelpers::SignalData<void()>(11, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'audiblePauseDrainCompleted'
        QtMocHelpers::SignalData<void()>(12, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'trackStatusContextChanged'
        QtMocHelpers::SignalData<void(const Fooyin::Engine::TrackStatusContext &)>(13, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 14, 15 },
        }}),
        // Signal 'levelReady'
        QtMocHelpers::SignalData<void(const Fooyin::LevelFrame &)>(16, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 17, 18 },
        }}),
        // Signal 'pcmReady'
        QtMocHelpers::SignalData<void(const Fooyin::PcmFrame &)>(19, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 20, 18 },
        }}),
        // Signal 'trackChanged'
        QtMocHelpers::SignalData<void(const Fooyin::Track &)>(21, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 22, 23 },
        }}),
        // Signal 'trackCommitted'
        QtMocHelpers::SignalData<void(const Fooyin::Engine::TrackCommitContext &)>(24, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 25, 15 },
        }}),
        // Signal 'trackAboutToFinish'
        QtMocHelpers::SignalData<void(const Fooyin::Engine::AboutToFinishContext &)>(26, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 27, 15 },
        }}),
        // Signal 'trackReadyToSwitch'
        QtMocHelpers::SignalData<void(const Fooyin::Engine::AboutToFinishContext &)>(28, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 27, 15 },
        }}),
        // Signal 'trackBoundaryReached'
        QtMocHelpers::SignalData<void(const Fooyin::Engine::AboutToFinishContext &)>(29, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 27, 15 },
        }}),
        // Signal 'nextTrackReadiness'
        QtMocHelpers::SignalData<void(const Fooyin::Engine::PlaybackItem &, bool, uint64_t)>(30, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 31, 32 }, { QMetaType::Bool, 33 }, { 0x80000000 | 34, 35 },
        }}),
        // Signal 'preparedCrossfadeArmResult'
        QtMocHelpers::SignalData<void(const Fooyin::Engine::PlaybackItem &, uint64_t, bool)>(36, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 31, 32 }, { 0x80000000 | 34, 37 }, { QMetaType::Bool, 38 },
        }}),
        // Signal 'preparedGaplessArmResult'
        QtMocHelpers::SignalData<void(const Fooyin::Engine::PlaybackItem &, uint64_t, bool)>(39, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 31, 32 }, { 0x80000000 | 34, 37 }, { QMetaType::Bool, 38 },
        }}),
        // Signal 'finished'
        QtMocHelpers::SignalData<void()>(40, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<EngineController, qt_meta_tag_ZN6Fooyin16EngineControllerE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::EngineController::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin16EngineControllerE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin16EngineControllerE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin16EngineControllerE_t>.metaTypes,
    nullptr
} };

void Fooyin::EngineController::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<EngineController *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->outputChanged((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 1: _t->deviceChanged((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 2: _t->engineError((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 3: _t->engineStateChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::PlaybackState>>(_a[1]))); break;
        case 4: _t->audiblePauseDrainStarted(); break;
        case 5: _t->audiblePauseDrainCompleted(); break;
        case 6: _t->trackStatusContextChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::TrackStatusContext>>(_a[1]))); break;
        case 7: _t->levelReady((*reinterpret_cast<std::add_pointer_t<Fooyin::LevelFrame>>(_a[1]))); break;
        case 8: _t->pcmReady((*reinterpret_cast<std::add_pointer_t<Fooyin::PcmFrame>>(_a[1]))); break;
        case 9: _t->trackChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::Track>>(_a[1]))); break;
        case 10: _t->trackCommitted((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::TrackCommitContext>>(_a[1]))); break;
        case 11: _t->trackAboutToFinish((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::AboutToFinishContext>>(_a[1]))); break;
        case 12: _t->trackReadyToSwitch((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::AboutToFinishContext>>(_a[1]))); break;
        case 13: _t->trackBoundaryReached((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::AboutToFinishContext>>(_a[1]))); break;
        case 14: _t->nextTrackReadiness((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::PlaybackItem>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[3]))); break;
        case 15: _t->preparedCrossfadeArmResult((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::PlaybackItem>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[3]))); break;
        case 16: _t->preparedGaplessArmResult((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::PlaybackItem>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[3]))); break;
        case 17: _t->finished(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 6:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::TrackStatusContext >(); break;
            }
            break;
        case 7:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::LevelFrame >(); break;
            }
            break;
        case 8:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::PcmFrame >(); break;
            }
            break;
        case 10:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::TrackCommitContext >(); break;
            }
            break;
        case 11:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::AboutToFinishContext >(); break;
            }
            break;
        case 12:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::AboutToFinishContext >(); break;
            }
            break;
        case 13:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::AboutToFinishContext >(); break;
            }
            break;
        case 14:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::PlaybackItem >(); break;
            }
            break;
        case 15:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::PlaybackItem >(); break;
            }
            break;
        case 16:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::PlaybackItem >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (EngineController::*)(const QString & , const QString & )>(_a, &EngineController::outputChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (EngineController::*)(const QString & )>(_a, &EngineController::deviceChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (EngineController::*)(const QString & )>(_a, &EngineController::engineError, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (EngineController::*)(Fooyin::Engine::PlaybackState )>(_a, &EngineController::engineStateChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (EngineController::*)()>(_a, &EngineController::audiblePauseDrainStarted, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (EngineController::*)()>(_a, &EngineController::audiblePauseDrainCompleted, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (EngineController::*)(const Fooyin::Engine::TrackStatusContext & )>(_a, &EngineController::trackStatusContextChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (EngineController::*)(const Fooyin::LevelFrame & )>(_a, &EngineController::levelReady, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (EngineController::*)(const Fooyin::PcmFrame & )>(_a, &EngineController::pcmReady, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (EngineController::*)(const Fooyin::Track & )>(_a, &EngineController::trackChanged, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (EngineController::*)(const Fooyin::Engine::TrackCommitContext & )>(_a, &EngineController::trackCommitted, 10))
            return;
        if (QtMocHelpers::indexOfMethod<void (EngineController::*)(const Fooyin::Engine::AboutToFinishContext & )>(_a, &EngineController::trackAboutToFinish, 11))
            return;
        if (QtMocHelpers::indexOfMethod<void (EngineController::*)(const Fooyin::Engine::AboutToFinishContext & )>(_a, &EngineController::trackReadyToSwitch, 12))
            return;
        if (QtMocHelpers::indexOfMethod<void (EngineController::*)(const Fooyin::Engine::AboutToFinishContext & )>(_a, &EngineController::trackBoundaryReached, 13))
            return;
        if (QtMocHelpers::indexOfMethod<void (EngineController::*)(const Fooyin::Engine::PlaybackItem & , bool , uint64_t )>(_a, &EngineController::nextTrackReadiness, 14))
            return;
        if (QtMocHelpers::indexOfMethod<void (EngineController::*)(const Fooyin::Engine::PlaybackItem & , uint64_t , bool )>(_a, &EngineController::preparedCrossfadeArmResult, 15))
            return;
        if (QtMocHelpers::indexOfMethod<void (EngineController::*)(const Fooyin::Engine::PlaybackItem & , uint64_t , bool )>(_a, &EngineController::preparedGaplessArmResult, 16))
            return;
        if (QtMocHelpers::indexOfMethod<void (EngineController::*)()>(_a, &EngineController::finished, 17))
            return;
    }
}

const QMetaObject *Fooyin::EngineController::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Fooyin::EngineController::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin16EngineControllerE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int Fooyin::EngineController::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 18)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 18;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 18)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 18;
    }
    return _id;
}

// SIGNAL 0
void Fooyin::EngineController::outputChanged(const QString & _t1, const QString & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1, _t2);
}

// SIGNAL 1
void Fooyin::EngineController::deviceChanged(const QString & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void Fooyin::EngineController::engineError(const QString & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void Fooyin::EngineController::engineStateChanged(Fooyin::Engine::PlaybackState _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void Fooyin::EngineController::audiblePauseDrainStarted()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void Fooyin::EngineController::audiblePauseDrainCompleted()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}

// SIGNAL 6
void Fooyin::EngineController::trackStatusContextChanged(const Fooyin::Engine::TrackStatusContext & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1);
}

// SIGNAL 7
void Fooyin::EngineController::levelReady(const Fooyin::LevelFrame & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1);
}

// SIGNAL 8
void Fooyin::EngineController::pcmReady(const Fooyin::PcmFrame & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 8, nullptr, _t1);
}

// SIGNAL 9
void Fooyin::EngineController::trackChanged(const Fooyin::Track & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 9, nullptr, _t1);
}

// SIGNAL 10
void Fooyin::EngineController::trackCommitted(const Fooyin::Engine::TrackCommitContext & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 10, nullptr, _t1);
}

// SIGNAL 11
void Fooyin::EngineController::trackAboutToFinish(const Fooyin::Engine::AboutToFinishContext & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 11, nullptr, _t1);
}

// SIGNAL 12
void Fooyin::EngineController::trackReadyToSwitch(const Fooyin::Engine::AboutToFinishContext & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 12, nullptr, _t1);
}

// SIGNAL 13
void Fooyin::EngineController::trackBoundaryReached(const Fooyin::Engine::AboutToFinishContext & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 13, nullptr, _t1);
}

// SIGNAL 14
void Fooyin::EngineController::nextTrackReadiness(const Fooyin::Engine::PlaybackItem & _t1, bool _t2, uint64_t _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 14, nullptr, _t1, _t2, _t3);
}

// SIGNAL 15
void Fooyin::EngineController::preparedCrossfadeArmResult(const Fooyin::Engine::PlaybackItem & _t1, uint64_t _t2, bool _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 15, nullptr, _t1, _t2, _t3);
}

// SIGNAL 16
void Fooyin::EngineController::preparedGaplessArmResult(const Fooyin::Engine::PlaybackItem & _t1, uint64_t _t2, bool _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 16, nullptr, _t1, _t2, _t3);
}

// SIGNAL 17
void Fooyin::EngineController::finished()
{
    QMetaObject::activate(this, &staticMetaObject, 17, nullptr);
}
QT_WARNING_POP
