/****************************************************************************
** Meta object code from reading C++ file 'playercontroller.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../../../fooyin/include/core/player/playercontroller.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'playercontroller.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin16PlayerControllerE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::PlayerController::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin16PlayerControllerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::PlayerController",
        "playStateChanged",
        "",
        "Fooyin::Player::PlayState",
        "state",
        "previous",
        "playModeChanged",
        "Fooyin::Playlist::PlayModes",
        "mode",
        "transportPlayRequested",
        "transportPauseRequested",
        "transportStopRequested",
        "positionChanged",
        "uint64_t",
        "ms",
        "positionChangedSeconds",
        "seconds",
        "bitrateChanged",
        "bitrate",
        "positionMoved",
        "currentTrackChanged",
        "Fooyin::Track",
        "track",
        "currentTrackUpdated",
        "currentTrackSeekableChanged",
        "seekable",
        "playlistTrackChanged",
        "Fooyin::PlaylistTrack",
        "playlistTrackUpdated",
        "trackPlayed",
        "trackChangeRequested",
        "Fooyin::Player::TrackChangeRequest",
        "request",
        "upcomingTrackChanged",
        "Fooyin::Player::UpcomingTrack",
        "upcomingTrack",
        "trackEndAutoTransitionsEnabledChanged",
        "enabled",
        "tracksQueued",
        "Fooyin::QueueTracks",
        "tracks",
        "index",
        "tracksDequeued",
        "trackIndexesDequeued",
        "Fooyin::PlaylistIndexes",
        "indexes",
        "trackQueueChanged",
        "removed",
        "added",
        "playbackSnapshotChanged",
        "Fooyin::Player::PlaybackSnapshot",
        "snapshot"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'playStateChanged'
        QtMocHelpers::SignalData<void(Fooyin::Player::PlayState, Fooyin::Player::PlayState)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 }, { 0x80000000 | 3, 5 },
        }}),
        // Signal 'playModeChanged'
        QtMocHelpers::SignalData<void(Fooyin::Playlist::PlayModes)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 7, 8 },
        }}),
        // Signal 'transportPlayRequested'
        QtMocHelpers::SignalData<void()>(9, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'transportPauseRequested'
        QtMocHelpers::SignalData<void()>(10, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'transportStopRequested'
        QtMocHelpers::SignalData<void()>(11, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'positionChanged'
        QtMocHelpers::SignalData<void(uint64_t)>(12, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 13, 14 },
        }}),
        // Signal 'positionChangedSeconds'
        QtMocHelpers::SignalData<void(uint64_t)>(15, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 13, 16 },
        }}),
        // Signal 'bitrateChanged'
        QtMocHelpers::SignalData<void(int)>(17, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 18 },
        }}),
        // Signal 'positionMoved'
        QtMocHelpers::SignalData<void(uint64_t)>(19, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 13, 14 },
        }}),
        // Signal 'currentTrackChanged'
        QtMocHelpers::SignalData<void(const Fooyin::Track &)>(20, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 21, 22 },
        }}),
        // Signal 'currentTrackUpdated'
        QtMocHelpers::SignalData<void(const Fooyin::Track &)>(23, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 21, 22 },
        }}),
        // Signal 'currentTrackSeekableChanged'
        QtMocHelpers::SignalData<void(bool)>(24, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 25 },
        }}),
        // Signal 'playlistTrackChanged'
        QtMocHelpers::SignalData<void(const Fooyin::PlaylistTrack &)>(26, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 27, 22 },
        }}),
        // Signal 'playlistTrackUpdated'
        QtMocHelpers::SignalData<void(const Fooyin::PlaylistTrack &)>(28, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 27, 22 },
        }}),
        // Signal 'trackPlayed'
        QtMocHelpers::SignalData<void(const Fooyin::Track &)>(29, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 21, 22 },
        }}),
        // Signal 'trackChangeRequested'
        QtMocHelpers::SignalData<void(const Fooyin::Player::TrackChangeRequest &)>(30, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 31, 32 },
        }}),
        // Signal 'upcomingTrackChanged'
        QtMocHelpers::SignalData<void(const Fooyin::Player::UpcomingTrack &)>(33, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 34, 35 },
        }}),
        // Signal 'trackEndAutoTransitionsEnabledChanged'
        QtMocHelpers::SignalData<void(bool)>(36, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 37 },
        }}),
        // Signal 'tracksQueued'
        QtMocHelpers::SignalData<void(const Fooyin::QueueTracks &, int)>(38, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 39, 40 }, { QMetaType::Int, 41 },
        }}),
        // Signal 'tracksDequeued'
        QtMocHelpers::SignalData<void(const Fooyin::QueueTracks &)>(42, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 39, 40 },
        }}),
        // Signal 'trackIndexesDequeued'
        QtMocHelpers::SignalData<void(const Fooyin::PlaylistIndexes &)>(43, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 44, 45 },
        }}),
        // Signal 'trackQueueChanged'
        QtMocHelpers::SignalData<void(const Fooyin::QueueTracks &, const Fooyin::QueueTracks &)>(46, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 39, 47 }, { 0x80000000 | 39, 48 },
        }}),
        // Signal 'playbackSnapshotChanged'
        QtMocHelpers::SignalData<void(const Fooyin::Player::PlaybackSnapshot &)>(49, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 50, 51 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<PlayerController, qt_meta_tag_ZN6Fooyin16PlayerControllerE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::PlayerController::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin16PlayerControllerE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin16PlayerControllerE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin16PlayerControllerE_t>.metaTypes,
    nullptr
} };

void Fooyin::PlayerController::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<PlayerController *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->playStateChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::Player::PlayState>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::Player::PlayState>>(_a[2]))); break;
        case 1: _t->playModeChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::Playlist::PlayModes>>(_a[1]))); break;
        case 2: _t->transportPlayRequested(); break;
        case 3: _t->transportPauseRequested(); break;
        case 4: _t->transportStopRequested(); break;
        case 5: _t->positionChanged((*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[1]))); break;
        case 6: _t->positionChangedSeconds((*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[1]))); break;
        case 7: _t->bitrateChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 8: _t->positionMoved((*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[1]))); break;
        case 9: _t->currentTrackChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::Track>>(_a[1]))); break;
        case 10: _t->currentTrackUpdated((*reinterpret_cast<std::add_pointer_t<Fooyin::Track>>(_a[1]))); break;
        case 11: _t->currentTrackSeekableChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 12: _t->playlistTrackChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::PlaylistTrack>>(_a[1]))); break;
        case 13: _t->playlistTrackUpdated((*reinterpret_cast<std::add_pointer_t<Fooyin::PlaylistTrack>>(_a[1]))); break;
        case 14: _t->trackPlayed((*reinterpret_cast<std::add_pointer_t<Fooyin::Track>>(_a[1]))); break;
        case 15: _t->trackChangeRequested((*reinterpret_cast<std::add_pointer_t<Fooyin::Player::TrackChangeRequest>>(_a[1]))); break;
        case 16: _t->upcomingTrackChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::Player::UpcomingTrack>>(_a[1]))); break;
        case 17: _t->trackEndAutoTransitionsEnabledChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 18: _t->tracksQueued((*reinterpret_cast<std::add_pointer_t<Fooyin::QueueTracks>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[2]))); break;
        case 19: _t->tracksDequeued((*reinterpret_cast<std::add_pointer_t<Fooyin::QueueTracks>>(_a[1]))); break;
        case 20: _t->trackIndexesDequeued((*reinterpret_cast<std::add_pointer_t<Fooyin::PlaylistIndexes>>(_a[1]))); break;
        case 21: _t->trackQueueChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::QueueTracks>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::QueueTracks>>(_a[2]))); break;
        case 22: _t->playbackSnapshotChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::Player::PlaybackSnapshot>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 15:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Player::TrackChangeRequest >(); break;
            }
            break;
        case 16:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Player::UpcomingTrack >(); break;
            }
            break;
        case 22:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Player::PlaybackSnapshot >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(Fooyin::Player::PlayState , Fooyin::Player::PlayState )>(_a, &PlayerController::playStateChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(Fooyin::Playlist::PlayModes )>(_a, &PlayerController::playModeChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)()>(_a, &PlayerController::transportPlayRequested, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)()>(_a, &PlayerController::transportPauseRequested, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)()>(_a, &PlayerController::transportStopRequested, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(uint64_t )>(_a, &PlayerController::positionChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(uint64_t )>(_a, &PlayerController::positionChangedSeconds, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(int )>(_a, &PlayerController::bitrateChanged, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(uint64_t )>(_a, &PlayerController::positionMoved, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(const Fooyin::Track & )>(_a, &PlayerController::currentTrackChanged, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(const Fooyin::Track & )>(_a, &PlayerController::currentTrackUpdated, 10))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(bool )>(_a, &PlayerController::currentTrackSeekableChanged, 11))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(const Fooyin::PlaylistTrack & )>(_a, &PlayerController::playlistTrackChanged, 12))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(const Fooyin::PlaylistTrack & )>(_a, &PlayerController::playlistTrackUpdated, 13))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(const Fooyin::Track & )>(_a, &PlayerController::trackPlayed, 14))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(const Fooyin::Player::TrackChangeRequest & )>(_a, &PlayerController::trackChangeRequested, 15))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(const Fooyin::Player::UpcomingTrack & )>(_a, &PlayerController::upcomingTrackChanged, 16))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(bool )>(_a, &PlayerController::trackEndAutoTransitionsEnabledChanged, 17))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(const Fooyin::QueueTracks & , int )>(_a, &PlayerController::tracksQueued, 18))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(const Fooyin::QueueTracks & )>(_a, &PlayerController::tracksDequeued, 19))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(const Fooyin::PlaylistIndexes & )>(_a, &PlayerController::trackIndexesDequeued, 20))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(const Fooyin::QueueTracks & , const Fooyin::QueueTracks & )>(_a, &PlayerController::trackQueueChanged, 21))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlayerController::*)(const Fooyin::Player::PlaybackSnapshot & )>(_a, &PlayerController::playbackSnapshotChanged, 22))
            return;
    }
}

const QMetaObject *Fooyin::PlayerController::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Fooyin::PlayerController::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin16PlayerControllerE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int Fooyin::PlayerController::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 23)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 23;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 23)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 23;
    }
    return _id;
}

// SIGNAL 0
void Fooyin::PlayerController::playStateChanged(Fooyin::Player::PlayState _t1, Fooyin::Player::PlayState _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1, _t2);
}

// SIGNAL 1
void Fooyin::PlayerController::playModeChanged(Fooyin::Playlist::PlayModes _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void Fooyin::PlayerController::transportPlayRequested()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void Fooyin::PlayerController::transportPauseRequested()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void Fooyin::PlayerController::transportStopRequested()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void Fooyin::PlayerController::positionChanged(uint64_t _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1);
}

// SIGNAL 6
void Fooyin::PlayerController::positionChangedSeconds(uint64_t _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1);
}

// SIGNAL 7
void Fooyin::PlayerController::bitrateChanged(int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1);
}

// SIGNAL 8
void Fooyin::PlayerController::positionMoved(uint64_t _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 8, nullptr, _t1);
}

// SIGNAL 9
void Fooyin::PlayerController::currentTrackChanged(const Fooyin::Track & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 9, nullptr, _t1);
}

// SIGNAL 10
void Fooyin::PlayerController::currentTrackUpdated(const Fooyin::Track & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 10, nullptr, _t1);
}

// SIGNAL 11
void Fooyin::PlayerController::currentTrackSeekableChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 11, nullptr, _t1);
}

// SIGNAL 12
void Fooyin::PlayerController::playlistTrackChanged(const Fooyin::PlaylistTrack & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 12, nullptr, _t1);
}

// SIGNAL 13
void Fooyin::PlayerController::playlistTrackUpdated(const Fooyin::PlaylistTrack & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 13, nullptr, _t1);
}

// SIGNAL 14
void Fooyin::PlayerController::trackPlayed(const Fooyin::Track & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 14, nullptr, _t1);
}

// SIGNAL 15
void Fooyin::PlayerController::trackChangeRequested(const Fooyin::Player::TrackChangeRequest & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 15, nullptr, _t1);
}

// SIGNAL 16
void Fooyin::PlayerController::upcomingTrackChanged(const Fooyin::Player::UpcomingTrack & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 16, nullptr, _t1);
}

// SIGNAL 17
void Fooyin::PlayerController::trackEndAutoTransitionsEnabledChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 17, nullptr, _t1);
}

// SIGNAL 18
void Fooyin::PlayerController::tracksQueued(const Fooyin::QueueTracks & _t1, int _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 18, nullptr, _t1, _t2);
}

// SIGNAL 19
void Fooyin::PlayerController::tracksDequeued(const Fooyin::QueueTracks & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 19, nullptr, _t1);
}

// SIGNAL 20
void Fooyin::PlayerController::trackIndexesDequeued(const Fooyin::PlaylistIndexes & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 20, nullptr, _t1);
}

// SIGNAL 21
void Fooyin::PlayerController::trackQueueChanged(const Fooyin::QueueTracks & _t1, const Fooyin::QueueTracks & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 21, nullptr, _t1, _t2);
}

// SIGNAL 22
void Fooyin::PlayerController::playbackSnapshotChanged(const Fooyin::Player::PlaybackSnapshot & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 22, nullptr, _t1);
}
QT_WARNING_POP
