/****************************************************************************
** Meta object code from reading C++ file 'audioengine.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../fooyin/src/core/engine/audioengine.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'audioengine.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin11AudioEngineE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::AudioEngine::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin11AudioEngineE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::AudioEngine",
        "stateChanged",
        "",
        "Fooyin::Engine::PlaybackState",
        "state",
        "audiblePauseDrainStarted",
        "audiblePauseDrainCompleted",
        "trackStatusContextChanged",
        "Fooyin::Engine::TrackStatus",
        "status",
        "Fooyin::Track",
        "track",
        "uint64_t",
        "generation",
        "seekable",
        "positionChanged",
        "positionMs",
        "positionChangedWithContext",
        "trackGeneration",
        "timelineEpoch",
        "seekRequestId",
        "positionContextChanged",
        "seekPositionApplied",
        "requestId",
        "bitrateChanged",
        "bitrate",
        "volumeChanged",
        "volume",
        "levelReady",
        "Fooyin::LevelFrame",
        "frame",
        "pcmReady",
        "Fooyin::PcmFrame",
        "trackAboutToFinish",
        "trackReadyToSwitch",
        "trackBoundaryReached",
        "remainingOutputMs",
        "engineOwnsTransition",
        "nextTrackReadiness",
        "Fooyin::Engine::PlaybackItem",
        "item",
        "ready",
        "finished",
        "deviceError",
        "error",
        "trackChanged",
        "trackCommitted",
        "Fooyin::Engine::TrackCommitContext",
        "context",
        "loadTrack",
        "manualChange",
        "setTrackEndAutoTransitionEnabled",
        "enabled",
        "setUpcomingTrackCandidate",
        "prepareNextTrack",
        "armPreparedCrossfadeTransition",
        "commitPreparedCrossfadeTransition",
        "shouldEnableTimelineTransitionHints",
        "Fooyin::AudioDecoder::PlaybackHints",
        "playbackHints",
        "armPreparedGaplessTransition",
        "commitPreparedGaplessTransition",
        "play",
        "pause",
        "stop",
        "stopImmediate",
        "restorePosition",
        "seek",
        "seekWithRequest",
        "setVolume",
        "handleOutputVolumeChange",
        "updateCurrentTrackMetadata",
        "setAudioOutput",
        "Fooyin::OutputCreator",
        "output",
        "device",
        "applyOutputProfile",
        "Fooyin::SampleFormat",
        "bitdepth",
        "dither",
        "Fooyin::Engine::DspChains",
        "chain",
        "setOutputDevice",
        "setAutomaticResampling",
        "preferredDspNames",
        "setAnalysisDataSubscriptions",
        "Fooyin::Engine::AnalysisDataTypes",
        "subscriptions",
        "setVisualisationAnalysisEnabled",
        "updateLiveDspSettings",
        "Fooyin::Engine::LiveDspSettingsUpdate",
        "update",
        "event",
        "QEvent*",
        "PhaseChangeReason",
        "Unknown",
        "PlaybackStatePlaying",
        "PlaybackStatePaused",
        "PlaybackStateStopped",
        "PlaybackStateError",
        "TrackStatusLoading",
        "ManualChangeFadeQueued",
        "AutoTransitionCrossfadeActive",
        "AutoTransitionPlayingNoOrphan",
        "SeekCrossfadeActive",
        "SeekSimpleActive",
        "SeekRestorePlaying",
        "SeekRestorePaused",
        "SeekRestoreStopped",
        "CrossfadeCleanupResumePlaying",
        "TransportPauseFadeQueued",
        "TransportStopFadeQueued",
        "DrainFillPrepareDiagnosticReason",
        "None",
        "NoUpcomingCandidate",
        "CandidateMatchesCurrent",
        "MultiTrackFileTransition",
        "AutoTransitionDisabled",
        "WaitingForEndOfInput",
        "PreparedCrossfadeAlreadyActive",
        "PreparedGaplessAlreadyActive",
        "AlreadyBuffered",
        "AlreadyRequested",
        "Enqueued"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'stateChanged'
        QtMocHelpers::SignalData<void(Fooyin::Engine::PlaybackState)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'audiblePauseDrainStarted'
        QtMocHelpers::SignalData<void()>(5, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'audiblePauseDrainCompleted'
        QtMocHelpers::SignalData<void()>(6, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'trackStatusContextChanged'
        QtMocHelpers::SignalData<void(Fooyin::Engine::TrackStatus, const Fooyin::Track &, uint64_t, bool)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 8, 9 }, { 0x80000000 | 10, 11 }, { 0x80000000 | 12, 13 }, { QMetaType::Bool, 14 },
        }}),
        // Signal 'positionChanged'
        QtMocHelpers::SignalData<void(uint64_t)>(15, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 12, 16 },
        }}),
        // Signal 'positionChangedWithContext'
        QtMocHelpers::SignalData<void(uint64_t, uint64_t, uint64_t, uint64_t)>(17, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 12, 16 }, { 0x80000000 | 12, 18 }, { 0x80000000 | 12, 19 }, { 0x80000000 | 12, 20 },
        }}),
        // Signal 'positionContextChanged'
        QtMocHelpers::SignalData<void(uint64_t, uint64_t, uint64_t)>(21, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 12, 18 }, { 0x80000000 | 12, 19 }, { 0x80000000 | 12, 20 },
        }}),
        // Signal 'seekPositionApplied'
        QtMocHelpers::SignalData<void(uint64_t, uint64_t)>(22, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 12, 16 }, { 0x80000000 | 12, 23 },
        }}),
        // Signal 'bitrateChanged'
        QtMocHelpers::SignalData<void(int)>(24, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 25 },
        }}),
        // Signal 'volumeChanged'
        QtMocHelpers::SignalData<void(double)>(26, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Double, 27 },
        }}),
        // Signal 'levelReady'
        QtMocHelpers::SignalData<void(const Fooyin::LevelFrame &)>(28, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 29, 30 },
        }}),
        // Signal 'pcmReady'
        QtMocHelpers::SignalData<void(const Fooyin::PcmFrame &)>(31, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 32, 30 },
        }}),
        // Signal 'trackAboutToFinish'
        QtMocHelpers::SignalData<void(const Fooyin::Track &, uint64_t)>(33, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 10, 11 }, { 0x80000000 | 12, 13 },
        }}),
        // Signal 'trackReadyToSwitch'
        QtMocHelpers::SignalData<void(const Fooyin::Track &, uint64_t)>(34, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 10, 11 }, { 0x80000000 | 12, 13 },
        }}),
        // Signal 'trackBoundaryReached'
        QtMocHelpers::SignalData<void(const Fooyin::Track &, uint64_t, uint64_t, bool)>(35, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 10, 11 }, { 0x80000000 | 12, 13 }, { 0x80000000 | 12, 36 }, { QMetaType::Bool, 37 },
        }}),
        // Signal 'nextTrackReadiness'
        QtMocHelpers::SignalData<void(const Fooyin::Engine::PlaybackItem &, bool, uint64_t)>(38, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 39, 40 }, { QMetaType::Bool, 41 }, { 0x80000000 | 12, 23 },
        }}),
        // Signal 'finished'
        QtMocHelpers::SignalData<void()>(42, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'deviceError'
        QtMocHelpers::SignalData<void(const QString &)>(43, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 44 },
        }}),
        // Signal 'trackChanged'
        QtMocHelpers::SignalData<void(const Fooyin::Track &)>(45, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 10, 11 },
        }}),
        // Signal 'trackCommitted'
        QtMocHelpers::SignalData<void(const Fooyin::Engine::TrackCommitContext &)>(46, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 47, 48 },
        }}),
        // Slot 'loadTrack'
        QtMocHelpers::SlotData<void(const Fooyin::Engine::PlaybackItem &, bool)>(49, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 39, 40 }, { QMetaType::Bool, 50 },
        }}),
        // Slot 'loadTrack'
        QtMocHelpers::SlotData<void(const Fooyin::Engine::PlaybackItem &)>(49, 2, QMC::AccessPublic | QMC::MethodCloned, QMetaType::Void, {{
            { 0x80000000 | 39, 40 },
        }}),
        // Slot 'setTrackEndAutoTransitionEnabled'
        QtMocHelpers::SlotData<void(bool)>(51, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 52 },
        }}),
        // Slot 'setUpcomingTrackCandidate'
        QtMocHelpers::SlotData<void(const Fooyin::Engine::PlaybackItem &)>(53, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 39, 40 },
        }}),
        // Slot 'prepareNextTrack'
        QtMocHelpers::SlotData<void(const Fooyin::Engine::PlaybackItem &, uint64_t)>(54, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 39, 40 }, { 0x80000000 | 12, 23 },
        }}),
        // Slot 'prepareNextTrack'
        QtMocHelpers::SlotData<void(const Fooyin::Engine::PlaybackItem &)>(54, 2, QMC::AccessPublic | QMC::MethodCloned, QMetaType::Void, {{
            { 0x80000000 | 39, 40 },
        }}),
        // Slot 'armPreparedCrossfadeTransition'
        QtMocHelpers::SlotData<bool(const Fooyin::Engine::PlaybackItem &, uint64_t)>(55, 2, QMC::AccessPublic, QMetaType::Bool, {{
            { 0x80000000 | 39, 40 }, { 0x80000000 | 12, 13 },
        }}),
        // Slot 'commitPreparedCrossfadeTransition'
        QtMocHelpers::SlotData<bool(const Fooyin::Engine::PlaybackItem &)>(56, 2, QMC::AccessPublic, QMetaType::Bool, {{
            { 0x80000000 | 39, 40 },
        }}),
        // Slot 'shouldEnableTimelineTransitionHints'
        QtMocHelpers::SlotData<bool(const Fooyin::Track &, Fooyin::AudioDecoder::PlaybackHints)>(57, 2, QMC::AccessPublic, QMetaType::Bool, {{
            { 0x80000000 | 10, 11 }, { 0x80000000 | 58, 59 },
        }}),
        // Slot 'armPreparedGaplessTransition'
        QtMocHelpers::SlotData<bool(const Fooyin::Engine::PlaybackItem &, uint64_t)>(60, 2, QMC::AccessPublic, QMetaType::Bool, {{
            { 0x80000000 | 39, 40 }, { 0x80000000 | 12, 13 },
        }}),
        // Slot 'commitPreparedGaplessTransition'
        QtMocHelpers::SlotData<bool(const Fooyin::Engine::PlaybackItem &)>(61, 2, QMC::AccessPublic, QMetaType::Bool, {{
            { 0x80000000 | 39, 40 },
        }}),
        // Slot 'play'
        QtMocHelpers::SlotData<void()>(62, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'pause'
        QtMocHelpers::SlotData<void()>(63, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'stop'
        QtMocHelpers::SlotData<void()>(64, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'stopImmediate'
        QtMocHelpers::SlotData<void()>(65, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'restorePosition'
        QtMocHelpers::SlotData<void(uint64_t, bool)>(66, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 12, 16 }, { QMetaType::Bool, 63 },
        }}),
        // Slot 'seek'
        QtMocHelpers::SlotData<void(uint64_t)>(67, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 12, 16 },
        }}),
        // Slot 'seekWithRequest'
        QtMocHelpers::SlotData<void(uint64_t, uint64_t)>(68, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 12, 16 }, { 0x80000000 | 12, 23 },
        }}),
        // Slot 'setVolume'
        QtMocHelpers::SlotData<void(double)>(69, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Double, 27 },
        }}),
        // Slot 'handleOutputVolumeChange'
        QtMocHelpers::SlotData<void(double)>(70, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Double, 27 },
        }}),
        // Slot 'updateCurrentTrackMetadata'
        QtMocHelpers::SlotData<void(const Fooyin::Track &)>(71, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 10, 11 },
        }}),
        // Slot 'setAudioOutput'
        QtMocHelpers::SlotData<void(const Fooyin::OutputCreator &, const QString &)>(72, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 73, 74 }, { QMetaType::QString, 75 },
        }}),
        // Slot 'applyOutputProfile'
        QtMocHelpers::SlotData<void(const Fooyin::OutputCreator &, const QString &, Fooyin::SampleFormat, bool, const Fooyin::Engine::DspChains &)>(76, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 73, 74 }, { QMetaType::QString, 75 }, { 0x80000000 | 77, 78 }, { QMetaType::Bool, 79 },
            { 0x80000000 | 80, 81 },
        }}),
        // Slot 'setOutputDevice'
        QtMocHelpers::SlotData<void(const QString &)>(82, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 75 },
        }}),
        // Slot 'setAutomaticResampling'
        QtMocHelpers::SlotData<void(bool, const QStringList &)>(83, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 52 }, { QMetaType::QStringList, 84 },
        }}),
        // Slot 'setAnalysisDataSubscriptions'
        QtMocHelpers::SlotData<void(Fooyin::Engine::AnalysisDataTypes)>(85, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 86, 87 },
        }}),
        // Slot 'setVisualisationAnalysisEnabled'
        QtMocHelpers::SlotData<void(bool)>(88, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 52 },
        }}),
        // Slot 'updateLiveDspSettings'
        QtMocHelpers::SlotData<void(const Fooyin::Engine::LiveDspSettingsUpdate &)>(89, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 90, 91 },
        }}),
        // Slot 'event'
        QtMocHelpers::SlotData<bool(QEvent *)>(92, 2, QMC::AccessPublic, QMetaType::Bool, {{
            { 0x80000000 | 93, 92 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'PhaseChangeReason'
        QtMocHelpers::EnumData<enum PhaseChangeReason>(94, 94, QMC::EnumIsScoped).add({
            {   95, PhaseChangeReason::Unknown },
            {   96, PhaseChangeReason::PlaybackStatePlaying },
            {   97, PhaseChangeReason::PlaybackStatePaused },
            {   98, PhaseChangeReason::PlaybackStateStopped },
            {   99, PhaseChangeReason::PlaybackStateError },
            {  100, PhaseChangeReason::TrackStatusLoading },
            {  101, PhaseChangeReason::ManualChangeFadeQueued },
            {  102, PhaseChangeReason::AutoTransitionCrossfadeActive },
            {  103, PhaseChangeReason::AutoTransitionPlayingNoOrphan },
            {  104, PhaseChangeReason::SeekCrossfadeActive },
            {  105, PhaseChangeReason::SeekSimpleActive },
            {  106, PhaseChangeReason::SeekRestorePlaying },
            {  107, PhaseChangeReason::SeekRestorePaused },
            {  108, PhaseChangeReason::SeekRestoreStopped },
            {  109, PhaseChangeReason::CrossfadeCleanupResumePlaying },
            {  110, PhaseChangeReason::TransportPauseFadeQueued },
            {  111, PhaseChangeReason::TransportStopFadeQueued },
        }),
        // enum 'DrainFillPrepareDiagnosticReason'
        QtMocHelpers::EnumData<enum DrainFillPrepareDiagnosticReason>(112, 112, QMC::EnumIsScoped).add({
            {  113, DrainFillPrepareDiagnosticReason::None },
            {  114, DrainFillPrepareDiagnosticReason::NoUpcomingCandidate },
            {  115, DrainFillPrepareDiagnosticReason::CandidateMatchesCurrent },
            {  116, DrainFillPrepareDiagnosticReason::MultiTrackFileTransition },
            {  117, DrainFillPrepareDiagnosticReason::AutoTransitionDisabled },
            {  118, DrainFillPrepareDiagnosticReason::WaitingForEndOfInput },
            {  119, DrainFillPrepareDiagnosticReason::PreparedCrossfadeAlreadyActive },
            {  120, DrainFillPrepareDiagnosticReason::PreparedGaplessAlreadyActive },
            {  121, DrainFillPrepareDiagnosticReason::AlreadyBuffered },
            {  122, DrainFillPrepareDiagnosticReason::AlreadyRequested },
            {  123, DrainFillPrepareDiagnosticReason::Enqueued },
        }),
    };
    return QtMocHelpers::metaObjectData<AudioEngine, qt_meta_tag_ZN6Fooyin11AudioEngineE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::AudioEngine::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin11AudioEngineE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin11AudioEngineE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin11AudioEngineE_t>.metaTypes,
    nullptr
} };

void Fooyin::AudioEngine::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<AudioEngine *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->stateChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::PlaybackState>>(_a[1]))); break;
        case 1: _t->audiblePauseDrainStarted(); break;
        case 2: _t->audiblePauseDrainCompleted(); break;
        case 3: _t->trackStatusContextChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::TrackStatus>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::Track>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[4]))); break;
        case 4: _t->positionChanged((*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[1]))); break;
        case 5: _t->positionChangedWithContext((*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[4]))); break;
        case 6: _t->positionContextChanged((*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[3]))); break;
        case 7: _t->seekPositionApplied((*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[2]))); break;
        case 8: _t->bitrateChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 9: _t->volumeChanged((*reinterpret_cast<std::add_pointer_t<double>>(_a[1]))); break;
        case 10: _t->levelReady((*reinterpret_cast<std::add_pointer_t<Fooyin::LevelFrame>>(_a[1]))); break;
        case 11: _t->pcmReady((*reinterpret_cast<std::add_pointer_t<Fooyin::PcmFrame>>(_a[1]))); break;
        case 12: _t->trackAboutToFinish((*reinterpret_cast<std::add_pointer_t<Fooyin::Track>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[2]))); break;
        case 13: _t->trackReadyToSwitch((*reinterpret_cast<std::add_pointer_t<Fooyin::Track>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[2]))); break;
        case 14: _t->trackBoundaryReached((*reinterpret_cast<std::add_pointer_t<Fooyin::Track>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[4]))); break;
        case 15: _t->nextTrackReadiness((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::PlaybackItem>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[3]))); break;
        case 16: _t->finished(); break;
        case 17: _t->deviceError((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 18: _t->trackChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::Track>>(_a[1]))); break;
        case 19: _t->trackCommitted((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::TrackCommitContext>>(_a[1]))); break;
        case 20: _t->loadTrack((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::PlaybackItem>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 21: _t->loadTrack((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::PlaybackItem>>(_a[1]))); break;
        case 22: _t->setTrackEndAutoTransitionEnabled((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 23: _t->setUpcomingTrackCandidate((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::PlaybackItem>>(_a[1]))); break;
        case 24: _t->prepareNextTrack((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::PlaybackItem>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[2]))); break;
        case 25: _t->prepareNextTrack((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::PlaybackItem>>(_a[1]))); break;
        case 26: { bool _r = _t->armPreparedCrossfadeTransition((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::PlaybackItem>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[2])));
            if (_a[0]) *reinterpret_cast<bool*>(_a[0]) = std::move(_r); }  break;
        case 27: { bool _r = _t->commitPreparedCrossfadeTransition((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::PlaybackItem>>(_a[1])));
            if (_a[0]) *reinterpret_cast<bool*>(_a[0]) = std::move(_r); }  break;
        case 28: { bool _r = _t->shouldEnableTimelineTransitionHints((*reinterpret_cast<std::add_pointer_t<Fooyin::Track>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::AudioDecoder::PlaybackHints>>(_a[2])));
            if (_a[0]) *reinterpret_cast<bool*>(_a[0]) = std::move(_r); }  break;
        case 29: { bool _r = _t->armPreparedGaplessTransition((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::PlaybackItem>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[2])));
            if (_a[0]) *reinterpret_cast<bool*>(_a[0]) = std::move(_r); }  break;
        case 30: { bool _r = _t->commitPreparedGaplessTransition((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::PlaybackItem>>(_a[1])));
            if (_a[0]) *reinterpret_cast<bool*>(_a[0]) = std::move(_r); }  break;
        case 31: _t->play(); break;
        case 32: _t->pause(); break;
        case 33: _t->stop(); break;
        case 34: _t->stopImmediate(); break;
        case 35: _t->restorePosition((*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 36: _t->seek((*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[1]))); break;
        case 37: _t->seekWithRequest((*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<uint64_t>>(_a[2]))); break;
        case 38: _t->setVolume((*reinterpret_cast<std::add_pointer_t<double>>(_a[1]))); break;
        case 39: _t->handleOutputVolumeChange((*reinterpret_cast<std::add_pointer_t<double>>(_a[1]))); break;
        case 40: _t->updateCurrentTrackMetadata((*reinterpret_cast<std::add_pointer_t<Fooyin::Track>>(_a[1]))); break;
        case 41: _t->setAudioOutput((*reinterpret_cast<std::add_pointer_t<Fooyin::OutputCreator>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 42: _t->applyOutputProfile((*reinterpret_cast<std::add_pointer_t<Fooyin::OutputCreator>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<Fooyin::SampleFormat>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[4])),(*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::DspChains>>(_a[5]))); break;
        case 43: _t->setOutputDevice((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 44: _t->setAutomaticResampling((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QStringList>>(_a[2]))); break;
        case 45: _t->setAnalysisDataSubscriptions((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::AnalysisDataTypes>>(_a[1]))); break;
        case 46: _t->setVisualisationAnalysisEnabled((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 47: _t->updateLiveDspSettings((*reinterpret_cast<std::add_pointer_t<Fooyin::Engine::LiveDspSettingsUpdate>>(_a[1]))); break;
        case 48: { bool _r = _t->event((*reinterpret_cast<std::add_pointer_t<QEvent*>>(_a[1])));
            if (_a[0]) *reinterpret_cast<bool*>(_a[0]) = std::move(_r); }  break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 10:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::LevelFrame >(); break;
            }
            break;
        case 11:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::PcmFrame >(); break;
            }
            break;
        case 15:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::PlaybackItem >(); break;
            }
            break;
        case 19:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::TrackCommitContext >(); break;
            }
            break;
        case 20:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::PlaybackItem >(); break;
            }
            break;
        case 21:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::PlaybackItem >(); break;
            }
            break;
        case 23:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::PlaybackItem >(); break;
            }
            break;
        case 24:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::PlaybackItem >(); break;
            }
            break;
        case 25:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::PlaybackItem >(); break;
            }
            break;
        case 26:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::PlaybackItem >(); break;
            }
            break;
        case 27:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::PlaybackItem >(); break;
            }
            break;
        case 29:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::PlaybackItem >(); break;
            }
            break;
        case 30:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::PlaybackItem >(); break;
            }
            break;
        case 45:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::AnalysisDataTypes >(); break;
            }
            break;
        case 47:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::Engine::LiveDspSettingsUpdate >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)(Fooyin::Engine::PlaybackState )>(_a, &AudioEngine::stateChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)()>(_a, &AudioEngine::audiblePauseDrainStarted, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)()>(_a, &AudioEngine::audiblePauseDrainCompleted, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)(Fooyin::Engine::TrackStatus , const Fooyin::Track & , uint64_t , bool )>(_a, &AudioEngine::trackStatusContextChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)(uint64_t )>(_a, &AudioEngine::positionChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)(uint64_t , uint64_t , uint64_t , uint64_t )>(_a, &AudioEngine::positionChangedWithContext, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)(uint64_t , uint64_t , uint64_t )>(_a, &AudioEngine::positionContextChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)(uint64_t , uint64_t )>(_a, &AudioEngine::seekPositionApplied, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)(int )>(_a, &AudioEngine::bitrateChanged, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)(double )>(_a, &AudioEngine::volumeChanged, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)(const Fooyin::LevelFrame & )>(_a, &AudioEngine::levelReady, 10))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)(const Fooyin::PcmFrame & )>(_a, &AudioEngine::pcmReady, 11))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)(const Fooyin::Track & , uint64_t )>(_a, &AudioEngine::trackAboutToFinish, 12))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)(const Fooyin::Track & , uint64_t )>(_a, &AudioEngine::trackReadyToSwitch, 13))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)(const Fooyin::Track & , uint64_t , uint64_t , bool )>(_a, &AudioEngine::trackBoundaryReached, 14))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)(const Fooyin::Engine::PlaybackItem & , bool , uint64_t )>(_a, &AudioEngine::nextTrackReadiness, 15))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)()>(_a, &AudioEngine::finished, 16))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)(const QString & )>(_a, &AudioEngine::deviceError, 17))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)(const Fooyin::Track & )>(_a, &AudioEngine::trackChanged, 18))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioEngine::*)(const Fooyin::Engine::TrackCommitContext & )>(_a, &AudioEngine::trackCommitted, 19))
            return;
    }
}

const QMetaObject *Fooyin::AudioEngine::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Fooyin::AudioEngine::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin11AudioEngineE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int Fooyin::AudioEngine::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 49)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 49;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 49)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 49;
    }
    return _id;
}

// SIGNAL 0
void Fooyin::AudioEngine::stateChanged(Fooyin::Engine::PlaybackState _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void Fooyin::AudioEngine::audiblePauseDrainStarted()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void Fooyin::AudioEngine::audiblePauseDrainCompleted()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void Fooyin::AudioEngine::trackStatusContextChanged(Fooyin::Engine::TrackStatus _t1, const Fooyin::Track & _t2, uint64_t _t3, bool _t4)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1, _t2, _t3, _t4);
}

// SIGNAL 4
void Fooyin::AudioEngine::positionChanged(uint64_t _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void Fooyin::AudioEngine::positionChangedWithContext(uint64_t _t1, uint64_t _t2, uint64_t _t3, uint64_t _t4)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1, _t2, _t3, _t4);
}

// SIGNAL 6
void Fooyin::AudioEngine::positionContextChanged(uint64_t _t1, uint64_t _t2, uint64_t _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1, _t2, _t3);
}

// SIGNAL 7
void Fooyin::AudioEngine::seekPositionApplied(uint64_t _t1, uint64_t _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1, _t2);
}

// SIGNAL 8
void Fooyin::AudioEngine::bitrateChanged(int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 8, nullptr, _t1);
}

// SIGNAL 9
void Fooyin::AudioEngine::volumeChanged(double _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 9, nullptr, _t1);
}

// SIGNAL 10
void Fooyin::AudioEngine::levelReady(const Fooyin::LevelFrame & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 10, nullptr, _t1);
}

// SIGNAL 11
void Fooyin::AudioEngine::pcmReady(const Fooyin::PcmFrame & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 11, nullptr, _t1);
}

// SIGNAL 12
void Fooyin::AudioEngine::trackAboutToFinish(const Fooyin::Track & _t1, uint64_t _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 12, nullptr, _t1, _t2);
}

// SIGNAL 13
void Fooyin::AudioEngine::trackReadyToSwitch(const Fooyin::Track & _t1, uint64_t _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 13, nullptr, _t1, _t2);
}

// SIGNAL 14
void Fooyin::AudioEngine::trackBoundaryReached(const Fooyin::Track & _t1, uint64_t _t2, uint64_t _t3, bool _t4)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 14, nullptr, _t1, _t2, _t3, _t4);
}

// SIGNAL 15
void Fooyin::AudioEngine::nextTrackReadiness(const Fooyin::Engine::PlaybackItem & _t1, bool _t2, uint64_t _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 15, nullptr, _t1, _t2, _t3);
}

// SIGNAL 16
void Fooyin::AudioEngine::finished()
{
    QMetaObject::activate(this, &staticMetaObject, 16, nullptr);
}

// SIGNAL 17
void Fooyin::AudioEngine::deviceError(const QString & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 17, nullptr, _t1);
}

// SIGNAL 18
void Fooyin::AudioEngine::trackChanged(const Fooyin::Track & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 18, nullptr, _t1);
}

// SIGNAL 19
void Fooyin::AudioEngine::trackCommitted(const Fooyin::Engine::TrackCommitContext & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 19, nullptr, _t1);
}
QT_WARNING_POP
