// This file is autogenerated. Changes will be overwritten.
#include "EWIEGA46WW/moc_lyricsconfigwidget.cpp"
#include "EWIEGA46WW/moc_lyricsdelegate.cpp"
#include "EWIEGA46WW/moc_lyricseditor.cpp"
#include "EWIEGA46WW/moc_lyricseditordialog.cpp"
#include "EWIEGA46WW/moc_lyricsfinder.cpp"
#include "EWIEGA46WW/moc_lyricsmodel.cpp"
#include "EWIEGA46WW/moc_lyricspropertiestab.cpp"
#include "EWIEGA46WW/moc_lyricssaver.cpp"
#include "EWIEGA46WW/moc_lyricssearchdialog.cpp"
#include "EWIEGA46WW/moc_lyricssearchmodel.cpp"
#include "EWIEGA46WW/moc_lyricsview.cpp"
#include "YPB6MEUM2W/moc_lyricssourcesmodel.cpp"
#include "EXY53EAHJJ/moc_darklyrics.cpp"
#include "EXY53EAHJJ/moc_kugoulyrics.cpp"
#include "EXY53EAHJJ/moc_locallyrics.cpp"
#include "EXY53EAHJJ/moc_lrcliblyrics.cpp"
#include "EXY53EAHJJ/moc_lyricsource.cpp"
#include "EXY53EAHJJ/moc_neteaselyrics.cpp"
#include "EXY53EAHJJ/moc_qqlyrics.cpp"
#include "EXY53EAHJJ/moc_taglyrics.cpp"
