/****************************************************************************
** Meta object code from reading C++ file 'radiobrowserservice.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../../fooyin/src/plugins/radiobrowser/radiobrowserservice.h"
#include <QtNetwork/QSslError>
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'radiobrowserservice.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin12RadioBrowser19RadioBrowserServiceE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::RadioBrowser::RadioBrowserService::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin12RadioBrowser19RadioBrowserServiceE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::RadioBrowser::RadioBrowserService",
        "searchStarted",
        "",
        "query",
        "searchFinished",
        "Fooyin::RadioBrowser::RadioStationList",
        "stations",
        "searchFailed",
        "error",
        "categoriesStarted",
        "Fooyin::RadioBrowser::RadioCategoryType",
        "type",
        "categoriesFinished",
        "Fooyin::RadioBrowser::RadioCategoryList",
        "categories",
        "categoriesFailed",
        "stationLookupFinished",
        "requestId",
        "stationLookupFailed"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'searchStarted'
        QtMocHelpers::SignalData<void(const QString &)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 3 },
        }}),
        // Signal 'searchFinished'
        QtMocHelpers::SignalData<void(const QString &, const Fooyin::RadioBrowser::RadioStationList &)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 3 }, { 0x80000000 | 5, 6 },
        }}),
        // Signal 'searchFailed'
        QtMocHelpers::SignalData<void(const QString &, const QString &)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 3 }, { QMetaType::QString, 8 },
        }}),
        // Signal 'categoriesStarted'
        QtMocHelpers::SignalData<void(Fooyin::RadioBrowser::RadioCategoryType)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 10, 11 },
        }}),
        // Signal 'categoriesFinished'
        QtMocHelpers::SignalData<void(Fooyin::RadioBrowser::RadioCategoryType, const Fooyin::RadioBrowser::RadioCategoryList &)>(12, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 10, 11 }, { 0x80000000 | 13, 14 },
        }}),
        // Signal 'categoriesFailed'
        QtMocHelpers::SignalData<void(Fooyin::RadioBrowser::RadioCategoryType, const QString &)>(15, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 10, 11 }, { QMetaType::QString, 8 },
        }}),
        // Signal 'stationLookupFinished'
        QtMocHelpers::SignalData<void(int, const Fooyin::RadioBrowser::RadioStationList &)>(16, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 17 }, { 0x80000000 | 5, 6 },
        }}),
        // Signal 'stationLookupFailed'
        QtMocHelpers::SignalData<void(int, const QString &)>(18, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 17 }, { QMetaType::QString, 8 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<RadioBrowserService, qt_meta_tag_ZN6Fooyin12RadioBrowser19RadioBrowserServiceE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::RadioBrowser::RadioBrowserService::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin12RadioBrowser19RadioBrowserServiceE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin12RadioBrowser19RadioBrowserServiceE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin12RadioBrowser19RadioBrowserServiceE_t>.metaTypes,
    nullptr
} };

void Fooyin::RadioBrowser::RadioBrowserService::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<RadioBrowserService *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->searchStarted((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 1: _t->searchFinished((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioStationList>>(_a[2]))); break;
        case 2: _t->searchFailed((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 3: _t->categoriesStarted((*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioCategoryType>>(_a[1]))); break;
        case 4: _t->categoriesFinished((*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioCategoryType>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioCategoryList>>(_a[2]))); break;
        case 5: _t->categoriesFailed((*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioCategoryType>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 6: _t->stationLookupFinished((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::RadioBrowser::RadioStationList>>(_a[2]))); break;
        case 7: _t->stationLookupFailed((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 1:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::RadioBrowser::RadioStationList >(); break;
            }
            break;
        case 4:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::RadioBrowser::RadioCategoryList >(); break;
            }
            break;
        case 6:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::RadioBrowser::RadioStationList >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserService::*)(const QString & )>(_a, &RadioBrowserService::searchStarted, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserService::*)(const QString & , const Fooyin::RadioBrowser::RadioStationList & )>(_a, &RadioBrowserService::searchFinished, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserService::*)(const QString & , const QString & )>(_a, &RadioBrowserService::searchFailed, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserService::*)(Fooyin::RadioBrowser::RadioCategoryType )>(_a, &RadioBrowserService::categoriesStarted, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserService::*)(Fooyin::RadioBrowser::RadioCategoryType , const Fooyin::RadioBrowser::RadioCategoryList & )>(_a, &RadioBrowserService::categoriesFinished, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserService::*)(Fooyin::RadioBrowser::RadioCategoryType , const QString & )>(_a, &RadioBrowserService::categoriesFailed, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserService::*)(int , const Fooyin::RadioBrowser::RadioStationList & )>(_a, &RadioBrowserService::stationLookupFinished, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (RadioBrowserService::*)(int , const QString & )>(_a, &RadioBrowserService::stationLookupFailed, 7))
            return;
    }
}

const QMetaObject *Fooyin::RadioBrowser::RadioBrowserService::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Fooyin::RadioBrowser::RadioBrowserService::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin12RadioBrowser19RadioBrowserServiceE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int Fooyin::RadioBrowser::RadioBrowserService::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 8)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 8;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 8)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 8;
    }
    return _id;
}

// SIGNAL 0
void Fooyin::RadioBrowser::RadioBrowserService::searchStarted(const QString & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void Fooyin::RadioBrowser::RadioBrowserService::searchFinished(const QString & _t1, const Fooyin::RadioBrowser::RadioStationList & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1, _t2);
}

// SIGNAL 2
void Fooyin::RadioBrowser::RadioBrowserService::searchFailed(const QString & _t1, const QString & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1, _t2);
}

// SIGNAL 3
void Fooyin::RadioBrowser::RadioBrowserService::categoriesStarted(Fooyin::RadioBrowser::RadioCategoryType _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void Fooyin::RadioBrowser::RadioBrowserService::categoriesFinished(Fooyin::RadioBrowser::RadioCategoryType _t1, const Fooyin::RadioBrowser::RadioCategoryList & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1, _t2);
}

// SIGNAL 5
void Fooyin::RadioBrowser::RadioBrowserService::categoriesFailed(Fooyin::RadioBrowser::RadioCategoryType _t1, const QString & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1, _t2);
}

// SIGNAL 6
void Fooyin::RadioBrowser::RadioBrowserService::stationLookupFinished(int _t1, const Fooyin::RadioBrowser::RadioStationList & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1, _t2);
}

// SIGNAL 7
void Fooyin::RadioBrowser::RadioBrowserService::stationLookupFailed(int _t1, const QString & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1, _t2);
}
QT_WARNING_POP
