/****************************************************************************
** Meta object code from reading C++ file 'openmptdefs.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../../fooyin/src/plugins/openmpt/openmptdefs.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'openmptdefs.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin8Settings7OpenMptE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::Settings::OpenMpt::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin8Settings7OpenMptE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::Settings::OpenMpt",
        "OpenMptSettings",
        "Gain",
        "Separation",
        "VolumeRamping",
        "InterpolationFilter",
        "EmulateAmiga",
        "LoopCount"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'OpenMptSettings'
        QtMocHelpers::EnumData<OpenMptSettings>(1, 1, QMC::EnumFlags{}).add({
            {    2, OpenMptSettings::Gain },
            {    3, OpenMptSettings::Separation },
            {    4, OpenMptSettings::VolumeRamping },
            {    5, OpenMptSettings::InterpolationFilter },
            {    6, OpenMptSettings::EmulateAmiga },
            {    7, OpenMptSettings::LoopCount },
        }),
    };
    return QtMocHelpers::metaObjectData<void, qt_meta_tag_ZN6Fooyin8Settings7OpenMptE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}

static constexpr auto qt_staticMetaObjectContent_ZN6Fooyin8Settings7OpenMptE =
    Fooyin::Settings::OpenMpt::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin8Settings7OpenMptE_t>();
static constexpr auto qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings7OpenMptE =
    qt_staticMetaObjectContent_ZN6Fooyin8Settings7OpenMptE.staticData;
static constexpr auto qt_staticMetaObjectRelocatingContent_ZN6Fooyin8Settings7OpenMptE =
    qt_staticMetaObjectContent_ZN6Fooyin8Settings7OpenMptE.relocatingData;

Q_CONSTINIT const QMetaObject Fooyin::Settings::OpenMpt::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings7OpenMptE.stringdata,
    qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings7OpenMptE.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent_ZN6Fooyin8Settings7OpenMptE.metaTypes,
    nullptr
} };

QT_WARNING_POP
