/*
 * Fooyin
 * Copyright © 2026, Luke Taylor <luket@pm.me>
 *
 * Fooyin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fooyin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fooyin.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#pragma once

#include <core/player/playerdefs.h>

#include <cstdint>
#include <optional>

namespace Fooyin::PlaybackState {
void saveActivePlaylistDbId(int dbId);
void clearActivePlaylistDbId();
std::optional<int> activePlaylistDbId();

void saveActiveTrackIndex(int index);
void clearActiveTrackIndex();
std::optional<int> activeTrackIndex();

void savePlaybackPosition(uint64_t position);
void clearPlaybackPosition();
std::optional<uint64_t> playbackPosition();

void savePlaybackTimeListened(uint64_t timeListenedMs);
void clearPlaybackTimeListened();
std::optional<uint64_t> playbackTimeListened();

void savePlaybackState(Player::PlayState state);
void clearPlaybackState();
std::optional<Player::PlayState> playbackState();
} // namespace Fooyin::PlaybackState
