/****************************************************************************
** Meta object code from reading C++ file 'playlistmodel.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../fooyin/src/gui/playlist/playlistmodel.h"
#include <QtCore/qmetatype.h>
#include <QtCore/QList>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'playlistmodel.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin13PlaylistModelE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::PlaylistModel::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin13PlaylistModelE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::PlaylistModel",
        "metadataWriteRequested",
        "",
        "Fooyin::TrackList",
        "tracks",
        "loadingStateChanged",
        "playlistLoaded",
        "filesDropped",
        "QList<QUrl>",
        "urls",
        "index",
        "tracksInserted",
        "Fooyin::TrackGroups",
        "groups",
        "tracksMoved",
        "Fooyin::MoveOperation",
        "operation",
        "trackGroupApplied",
        "playingTrackChanged",
        "Fooyin::PlaylistTrack",
        "track",
        "playStateChanged",
        "Fooyin::Player::PlayState",
        "state",
        "refreshPlayingTrackPositionData",
        "refreshPlayingTrackBitrateData",
        "PlaybackDependencies",
        "PlaybackDependency",
        "None",
        "Position",
        "Bitrate",
        "PlaybackState",
        "All"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'metadataWriteRequested'
        QtMocHelpers::SignalData<void(const Fooyin::TrackList &)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'loadingStateChanged'
        QtMocHelpers::SignalData<void()>(5, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'playlistLoaded'
        QtMocHelpers::SignalData<void()>(6, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'filesDropped'
        QtMocHelpers::SignalData<void(const QList<QUrl> &, int)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 8, 9 }, { QMetaType::Int, 10 },
        }}),
        // Signal 'tracksInserted'
        QtMocHelpers::SignalData<void(const Fooyin::TrackGroups &)>(11, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 12, 13 },
        }}),
        // Signal 'tracksMoved'
        QtMocHelpers::SignalData<void(const Fooyin::MoveOperation &)>(14, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 15, 16 },
        }}),
        // Signal 'trackGroupApplied'
        QtMocHelpers::SignalData<void()>(17, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'playingTrackChanged'
        QtMocHelpers::SlotData<void(const Fooyin::PlaylistTrack &)>(18, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 19, 20 },
        }}),
        // Slot 'playStateChanged'
        QtMocHelpers::SlotData<void(Fooyin::Player::PlayState)>(21, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 22, 23 },
        }}),
        // Slot 'refreshPlayingTrackPositionData'
        QtMocHelpers::SlotData<void()>(24, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'refreshPlayingTrackBitrateData'
        QtMocHelpers::SlotData<void()>(25, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // flag 'PlaybackDependencies'
        QtMocHelpers::EnumData<PlaybackDependencies>(26, 27, QMC::EnumIsFlag).add({
            {   28, PlaybackDependency::None },
            {   29, PlaybackDependency::Position },
            {   30, PlaybackDependency::Bitrate },
            {   31, PlaybackDependency::PlaybackState },
            {   32, PlaybackDependency::All },
        }),
    };
    return QtMocHelpers::metaObjectData<PlaylistModel, qt_meta_tag_ZN6Fooyin13PlaylistModelE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::PlaylistModel::staticMetaObject = { {
    QMetaObject::SuperData::link<TreeModel<PlaylistItem>::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin13PlaylistModelE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin13PlaylistModelE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin13PlaylistModelE_t>.metaTypes,
    nullptr
} };

void Fooyin::PlaylistModel::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<PlaylistModel *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->metadataWriteRequested((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1]))); break;
        case 1: _t->loadingStateChanged(); break;
        case 2: _t->playlistLoaded(); break;
        case 3: _t->filesDropped((*reinterpret_cast<std::add_pointer_t<QList<QUrl>>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[2]))); break;
        case 4: _t->tracksInserted((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackGroups>>(_a[1]))); break;
        case 5: _t->tracksMoved((*reinterpret_cast<std::add_pointer_t<Fooyin::MoveOperation>>(_a[1]))); break;
        case 6: _t->trackGroupApplied(); break;
        case 7: _t->playingTrackChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::PlaylistTrack>>(_a[1]))); break;
        case 8: _t->playStateChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::Player::PlayState>>(_a[1]))); break;
        case 9: _t->refreshPlayingTrackPositionData(); break;
        case 10: _t->refreshPlayingTrackBitrateData(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 0:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (PlaylistModel::*)(const Fooyin::TrackList & )>(_a, &PlaylistModel::metadataWriteRequested, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistModel::*)()>(_a, &PlaylistModel::loadingStateChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistModel::*)()>(_a, &PlaylistModel::playlistLoaded, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistModel::*)(const QList<QUrl> & , int )>(_a, &PlaylistModel::filesDropped, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistModel::*)(const Fooyin::TrackGroups & )>(_a, &PlaylistModel::tracksInserted, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistModel::*)(const Fooyin::MoveOperation & )>(_a, &PlaylistModel::tracksMoved, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistModel::*)()>(_a, &PlaylistModel::trackGroupApplied, 6))
            return;
    }
}

const QMetaObject *Fooyin::PlaylistModel::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Fooyin::PlaylistModel::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin13PlaylistModelE_t>.strings))
        return static_cast<void*>(this);
    return TreeModel<PlaylistItem>::qt_metacast(_clname);
}

int Fooyin::PlaylistModel::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = TreeModel<PlaylistItem>::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 11)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 11;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 11)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 11;
    }
    return _id;
}

// SIGNAL 0
void Fooyin::PlaylistModel::metadataWriteRequested(const Fooyin::TrackList & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void Fooyin::PlaylistModel::loadingStateChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void Fooyin::PlaylistModel::playlistLoaded()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void Fooyin::PlaylistModel::filesDropped(const QList<QUrl> & _t1, int _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1, _t2);
}

// SIGNAL 4
void Fooyin::PlaylistModel::tracksInserted(const Fooyin::TrackGroups & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void Fooyin::PlaylistModel::tracksMoved(const Fooyin::MoveOperation & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1);
}

// SIGNAL 6
void Fooyin::PlaylistModel::trackGroupApplied()
{
    QMetaObject::activate(this, &staticMetaObject, 6, nullptr);
}
QT_WARNING_POP
