/*
 * Fooyin
 * Copyright © 2024, Luke Taylor <luket@pm.me>
 *
 * Fooyin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fooyin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fooyin.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <utils/database/dbconnectionprovider.h>

#include <QDebug>
#include <QLoggingCategory>

Q_LOGGING_CATEGORY(DB_CONPROV, "fy.db")

namespace Fooyin {
DbConnectionProvider::DbConnectionProvider()
    : DbConnectionProvider{nullptr}
{ }

DbConnectionProvider::DbConnectionProvider(DbConnectionPoolPtr pool)
    : m_connectionPool{std::move(pool)}
{ }

QSqlDatabase DbConnectionProvider::db() const
{
    if(!m_connectionPool) {
        qCWarning(DB_CONPROV) << "No connection pool";
        return {};
    }

    const DbConnection* connection = m_connectionPool->threadConnection();

    if(!connection) {
        qCWarning(DB_CONPROV) << "Thread connection not found";
        return {};
    }

    if(!connection->isOpen() && !connection->db().open()) {
        qCWarning(DB_CONPROV) << "Thread connection could not be opened";
        return {};
    }

    return connection->db();
}
} // namespace Fooyin
