/****************************************************************************
** Meta object code from reading C++ file 'playbackphase.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../fooyin/src/core/engine/control/playbackphase.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'playbackphase.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin8PlaybackE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::Playback::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin8PlaybackE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::Playback",
        "Phase",
        "Stopped",
        "Loading",
        "Playing",
        "Paused",
        "Seeking",
        "SeekCrossfading",
        "TrackCrossfading",
        "FadingToPause",
        "FadingToStop",
        "FadingToManualChange",
        "Error"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'Phase'
        QtMocHelpers::EnumData<Phase>(1, 1, QMC::EnumIsScoped).add({
            {    2, Phase::Stopped },
            {    3, Phase::Loading },
            {    4, Phase::Playing },
            {    5, Phase::Paused },
            {    6, Phase::Seeking },
            {    7, Phase::SeekCrossfading },
            {    8, Phase::TrackCrossfading },
            {    9, Phase::FadingToPause },
            {   10, Phase::FadingToStop },
            {   11, Phase::FadingToManualChange },
            {   12, Phase::Error },
        }),
    };
    return QtMocHelpers::metaObjectData<void, qt_meta_tag_ZN6Fooyin8PlaybackE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}

static constexpr auto qt_staticMetaObjectContent_ZN6Fooyin8PlaybackE =
    Fooyin::Playback::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin8PlaybackE_t>();
static constexpr auto qt_staticMetaObjectStaticContent_ZN6Fooyin8PlaybackE =
    qt_staticMetaObjectContent_ZN6Fooyin8PlaybackE.staticData;
static constexpr auto qt_staticMetaObjectRelocatingContent_ZN6Fooyin8PlaybackE =
    qt_staticMetaObjectContent_ZN6Fooyin8PlaybackE.relocatingData;

Q_CONSTINIT const QMetaObject Fooyin::Playback::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent_ZN6Fooyin8PlaybackE.stringdata,
    qt_staticMetaObjectStaticContent_ZN6Fooyin8PlaybackE.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent_ZN6Fooyin8PlaybackE.metaTypes,
    nullptr
} };

QT_WARNING_POP
