/****************************************************************************
** Meta object code from reading C++ file 'internalcoresettings.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../fooyin/src/core/internalcoresettings.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'internalcoresettings.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin8Settings4Core8InternalE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::Settings::Core::Internal::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin8Settings4Core8InternalE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::Settings::Core::Internal",
        "CoreInternalSettings",
        "MonitorLibraryDirectories",
        "MuteVolume",
        "DisabledPlugins",
        "EngineFading",
        "FadingValues",
        "EngineCrossfading",
        "CrossfadingValues",
        "VBRUpdateInterval",
        "ProxyMode",
        "ProxyType",
        "ProxyHost",
        "ProxyPort",
        "ProxyAuth",
        "ProxyUsername",
        "ProxyPassword",
        "DecodeLowWatermarkRatio",
        "DecodeHighWatermarkRatio",
        "CrossfadeSwitchPolicy",
        "OutputDeviceProfiles",
        "OpusHeaderWriteMode",
        "RemoteReadAheadKb",
        "RemoteBufferLengthMs",
        "RemotePrebufferMs",
        "RemoteOpenTimeoutMs",
        "MonitorTrackFiles",
        "OutputAutoResample",
        "OutputResamplerPreference"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'CoreInternalSettings'
        QtMocHelpers::EnumData<CoreInternalSettings>(1, 1, QMC::EnumFlags{}).add({
            {    2, CoreInternalSettings::MonitorLibraryDirectories },
            {    3, CoreInternalSettings::MuteVolume },
            {    4, CoreInternalSettings::DisabledPlugins },
            {    5, CoreInternalSettings::EngineFading },
            {    6, CoreInternalSettings::FadingValues },
            {    7, CoreInternalSettings::EngineCrossfading },
            {    8, CoreInternalSettings::CrossfadingValues },
            {    9, CoreInternalSettings::VBRUpdateInterval },
            {   10, CoreInternalSettings::ProxyMode },
            {   11, CoreInternalSettings::ProxyType },
            {   12, CoreInternalSettings::ProxyHost },
            {   13, CoreInternalSettings::ProxyPort },
            {   14, CoreInternalSettings::ProxyAuth },
            {   15, CoreInternalSettings::ProxyUsername },
            {   16, CoreInternalSettings::ProxyPassword },
            {   17, CoreInternalSettings::DecodeLowWatermarkRatio },
            {   18, CoreInternalSettings::DecodeHighWatermarkRatio },
            {   19, CoreInternalSettings::CrossfadeSwitchPolicy },
            {   20, CoreInternalSettings::OutputDeviceProfiles },
            {   21, CoreInternalSettings::OpusHeaderWriteMode },
            {   22, CoreInternalSettings::RemoteReadAheadKb },
            {   23, CoreInternalSettings::RemoteBufferLengthMs },
            {   24, CoreInternalSettings::RemotePrebufferMs },
            {   25, CoreInternalSettings::RemoteOpenTimeoutMs },
            {   26, CoreInternalSettings::MonitorTrackFiles },
            {   27, CoreInternalSettings::OutputAutoResample },
            {   28, CoreInternalSettings::OutputResamplerPreference },
        }),
    };
    return QtMocHelpers::metaObjectData<void, qt_meta_tag_ZN6Fooyin8Settings4Core8InternalE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}

static constexpr auto qt_staticMetaObjectContent_ZN6Fooyin8Settings4Core8InternalE =
    Fooyin::Settings::Core::Internal::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin8Settings4Core8InternalE_t>();
static constexpr auto qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings4Core8InternalE =
    qt_staticMetaObjectContent_ZN6Fooyin8Settings4Core8InternalE.staticData;
static constexpr auto qt_staticMetaObjectRelocatingContent_ZN6Fooyin8Settings4Core8InternalE =
    qt_staticMetaObjectContent_ZN6Fooyin8Settings4Core8InternalE.relocatingData;

Q_CONSTINIT const QMetaObject Fooyin::Settings::Core::Internal::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings4Core8InternalE.stringdata,
    qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings4Core8InternalE.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent_ZN6Fooyin8Settings4Core8InternalE.metaTypes,
    nullptr
} };

QT_WARNING_POP
