/*
 * Fooyin
 * Copyright © 2023, Luke Taylor <luket@pm.me>
 *
 * Fooyin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fooyin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fooyin.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#pragma once

#include <QObject>
#include <QPersistentModelIndex>
#include <QTreeView>

class QEvent;

namespace Fooyin {
class LibraryTreeView : public QTreeView
{
    Q_OBJECT

public:
    explicit LibraryTreeView(QWidget* parent = nullptr);

    void setLoading(bool isLoading);
    void setExpandsOnSingleClick(bool enabled);

Q_SIGNALS:
    void middleClicked(const QModelIndex& index);
    void displayChanged();

protected:
    void changeEvent(QEvent* event) override;
    void mousePressEvent(QMouseEvent* event) override;
    void mouseReleaseEvent(QMouseEvent* event) override;
    void mouseDoubleClickEvent(QMouseEvent* event) override;
    void wheelEvent(QWheelEvent* event) override;
    void paintEvent(QPaintEvent* event) override;
    QModelIndex moveCursor(CursorAction cursorAction, Qt::KeyboardModifiers modifiers) override;

private:
    bool m_isLoading;
    bool m_expandsOnSingleClick;
    QPersistentModelIndex m_singleClickToggleIndex;
};
} // namespace Fooyin
