/*
 * This file was generated by qdbusxml2cpp version 0.8
 * Source file was org.mpris.MediaPlayer2.Player.xml
 *
 * qdbusxml2cpp is Copyright (C) The Qt Company Ltd. and other contributors.
 *
 * This is an auto-generated file.
 * Do not edit! All changes made to it will be lost.
 */

#include "mprisplayer.h"
#include <QtCore/QMetaObject>
#include <QtCore/QByteArray>
#include <QtCore/QList>
#include <QtCore/QMap>
#include <QtCore/QString>
#include <QtCore/QStringList>
#include <QtCore/QVariant>

/*
 * Implementation of adaptor class MprisPlayer
 */

MprisPlayer::MprisPlayer(Fooyin::Mpris::MprisPlugin *parent)
    : QDBusAbstractAdaptor(parent)
{
    // constructor
    setAutoRelaySignals(true);
}

MprisPlayer::~MprisPlayer()
{
    // destructor
}

bool MprisPlayer::canControl() const
{
    // get the value of property CanControl
    return qvariant_cast< bool >(parent()->property("CanControl"));
}

bool MprisPlayer::canGoNext() const
{
    // get the value of property CanGoNext
    return qvariant_cast< bool >(parent()->property("CanGoNext"));
}

bool MprisPlayer::canGoPrevious() const
{
    // get the value of property CanGoPrevious
    return qvariant_cast< bool >(parent()->property("CanGoPrevious"));
}

bool MprisPlayer::canPause() const
{
    // get the value of property CanPause
    return qvariant_cast< bool >(parent()->property("CanPause"));
}

bool MprisPlayer::canPlay() const
{
    // get the value of property CanPlay
    return qvariant_cast< bool >(parent()->property("CanPlay"));
}

bool MprisPlayer::canSeek() const
{
    // get the value of property CanSeek
    return qvariant_cast< bool >(parent()->property("CanSeek"));
}

QString MprisPlayer::loopStatus() const
{
    // get the value of property LoopStatus
    return qvariant_cast< QString >(parent()->property("LoopStatus"));
}

void MprisPlayer::setLoopStatus(const QString &value)
{
    // set the value of property LoopStatus
    parent()->setProperty("LoopStatus", QVariant::fromValue(value));
}

double MprisPlayer::maximumRate() const
{
    // get the value of property MaximumRate
    return qvariant_cast< double >(parent()->property("MaximumRate"));
}

QVariantMap MprisPlayer::metadata() const
{
    // get the value of property Metadata
    return qvariant_cast< QVariantMap >(parent()->property("Metadata"));
}

double MprisPlayer::minimumRate() const
{
    // get the value of property MinimumRate
    return qvariant_cast< double >(parent()->property("MinimumRate"));
}

QString MprisPlayer::playbackStatus() const
{
    // get the value of property PlaybackStatus
    return qvariant_cast< QString >(parent()->property("PlaybackStatus"));
}

qlonglong MprisPlayer::position() const
{
    // get the value of property Position
    return qvariant_cast< qlonglong >(parent()->property("Position"));
}

bool MprisPlayer::shuffle() const
{
    // get the value of property Shuffle
    return qvariant_cast< bool >(parent()->property("Shuffle"));
}

void MprisPlayer::setShuffle(bool value)
{
    // set the value of property Shuffle
    parent()->setProperty("Shuffle", QVariant::fromValue(value));
}

double MprisPlayer::volume() const
{
    // get the value of property Volume
    return qvariant_cast< double >(parent()->property("Volume"));
}

void MprisPlayer::setVolume(double value)
{
    // set the value of property Volume
    parent()->setProperty("Volume", QVariant::fromValue(value));
}

void MprisPlayer::Next()
{
    // handle method call org.mpris.MediaPlayer2.Player.Next
    parent()->Next();
}

void MprisPlayer::Pause()
{
    // handle method call org.mpris.MediaPlayer2.Player.Pause
    parent()->Pause();
}

void MprisPlayer::Play()
{
    // handle method call org.mpris.MediaPlayer2.Player.Play
    parent()->Play();
}

void MprisPlayer::PlayPause()
{
    // handle method call org.mpris.MediaPlayer2.Player.PlayPause
    parent()->PlayPause();
}

void MprisPlayer::Previous()
{
    // handle method call org.mpris.MediaPlayer2.Player.Previous
    parent()->Previous();
}

void MprisPlayer::Seek(qlonglong Offset)
{
    // handle method call org.mpris.MediaPlayer2.Player.Seek
    parent()->Seek(Offset);
}

void MprisPlayer::SetPosition(const QDBusObjectPath &TrackId, qlonglong Position)
{
    // handle method call org.mpris.MediaPlayer2.Player.SetPosition
    parent()->SetPosition(TrackId, Position);
}

void MprisPlayer::Stop()
{
    // handle method call org.mpris.MediaPlayer2.Player.Stop
    parent()->Stop();
}


#include "moc_mprisplayer.cpp"
