// This file is autogenerated. Changes will be overwritten.
#include "CSWKE6ESUK/moc_freedesktopnotificationbackend.cpp"
#include "CSWKE6ESUK/moc_portalnotificationbackend.cpp"
#include "EWIEGA46WW/moc_notificationbackend.cpp"
#include "YPB6MEUM2W/moc_notifysettings.cpp"
