/****************************************************************************
** Meta object code from reading C++ file 'filemenu.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../fooyin/src/gui/menubar/filemenu.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'filemenu.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin8FileMenuE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::FileMenu::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin8FileMenuE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::FileMenu",
        "requestExit",
        "",
        "requestAddFiles",
        "requestAddFolders",
        "requestAddStreamUrl",
        "requestNewPlaylist",
        "requestNewAutoPlaylist",
        "requestLoadPlaylist",
        "requestSavePlaylist",
        "requestSaveAllPlaylists"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'requestExit'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'requestAddFiles'
        QtMocHelpers::SignalData<void()>(3, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'requestAddFolders'
        QtMocHelpers::SignalData<void()>(4, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'requestAddStreamUrl'
        QtMocHelpers::SignalData<void()>(5, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'requestNewPlaylist'
        QtMocHelpers::SignalData<void()>(6, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'requestNewAutoPlaylist'
        QtMocHelpers::SignalData<void()>(7, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'requestLoadPlaylist'
        QtMocHelpers::SignalData<void()>(8, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'requestSavePlaylist'
        QtMocHelpers::SignalData<void()>(9, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'requestSaveAllPlaylists'
        QtMocHelpers::SignalData<void()>(10, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<FileMenu, qt_meta_tag_ZN6Fooyin8FileMenuE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::FileMenu::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin8FileMenuE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin8FileMenuE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin8FileMenuE_t>.metaTypes,
    nullptr
} };

void Fooyin::FileMenu::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<FileMenu *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->requestExit(); break;
        case 1: _t->requestAddFiles(); break;
        case 2: _t->requestAddFolders(); break;
        case 3: _t->requestAddStreamUrl(); break;
        case 4: _t->requestNewPlaylist(); break;
        case 5: _t->requestNewAutoPlaylist(); break;
        case 6: _t->requestLoadPlaylist(); break;
        case 7: _t->requestSavePlaylist(); break;
        case 8: _t->requestSaveAllPlaylists(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (FileMenu::*)()>(_a, &FileMenu::requestExit, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (FileMenu::*)()>(_a, &FileMenu::requestAddFiles, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (FileMenu::*)()>(_a, &FileMenu::requestAddFolders, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (FileMenu::*)()>(_a, &FileMenu::requestAddStreamUrl, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (FileMenu::*)()>(_a, &FileMenu::requestNewPlaylist, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (FileMenu::*)()>(_a, &FileMenu::requestNewAutoPlaylist, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (FileMenu::*)()>(_a, &FileMenu::requestLoadPlaylist, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (FileMenu::*)()>(_a, &FileMenu::requestSavePlaylist, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (FileMenu::*)()>(_a, &FileMenu::requestSaveAllPlaylists, 8))
            return;
    }
}

const QMetaObject *Fooyin::FileMenu::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Fooyin::FileMenu::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin8FileMenuE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int Fooyin::FileMenu::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 9)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 9;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 9)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 9;
    }
    return _id;
}

// SIGNAL 0
void Fooyin::FileMenu::requestExit()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void Fooyin::FileMenu::requestAddFiles()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void Fooyin::FileMenu::requestAddFolders()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void Fooyin::FileMenu::requestAddStreamUrl()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void Fooyin::FileMenu::requestNewPlaylist()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void Fooyin::FileMenu::requestNewAutoPlaylist()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}

// SIGNAL 6
void Fooyin::FileMenu::requestLoadPlaylist()
{
    QMetaObject::activate(this, &staticMetaObject, 6, nullptr);
}

// SIGNAL 7
void Fooyin::FileMenu::requestSavePlaylist()
{
    QMetaObject::activate(this, &staticMetaObject, 7, nullptr);
}

// SIGNAL 8
void Fooyin::FileMenu::requestSaveAllPlaylists()
{
    QMetaObject::activate(this, &staticMetaObject, 8, nullptr);
}
QT_WARNING_POP
