/*
 * Fooyin
 * Copyright © 2026, Luke Taylor <luket@pm.me>
 * Copyright © 2026, Gustav Oechler <gustavoechler@gmail.com>
 *
 * Fooyin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fooyin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fooyin.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#pragma once

#include "inhibitor.h"

#include <QPointer>
#include <QtDBus/QDBusInterface>
#include <QtDBus/QDBusObjectPath>
#include <QtDBus/QDBusPendingCallWatcher>

namespace Fooyin::SleepInhibitor {
class InhibitorDbus : public InhibitorPrivate
{
    Q_OBJECT

public:
    explicit InhibitorDbus(QObject* parent = nullptr);

    void inhibitSleep() override;
    void uninhibitSleep() override;

private:
    enum class Interface : uint8_t
    {
        None,
        GnomeSessionManager,
        FreedesktopPower,
        FreedesktopPortal,
    };

    void onInhibitCallFinished(QDBusPendingCallWatcher* watcher);
    void onUninhibitCallFinished(QDBusPendingCallWatcher* watcher);

    QPointer<QDBusInterface> m_busInterface;
    Interface m_interface{Interface::None};
    uint32_t m_inhibitCookie{0};     // Used by GnomeSessionManager and FreedesktopPower
    QDBusObjectPath m_inhibitHandle; // Used by FreedesktopPortal
};
} // namespace Fooyin::SleepInhibitor
