/****************************************************************************
** Meta object code from reading C++ file 'coresettings.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../fooyin/include/core/coresettings.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'coresettings.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin8Settings4CoreE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::Settings::Core::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin8Settings4CoreE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::Settings::Core",
        "CoreSettings",
        "Version",
        "FirstRun",
        "PlayMode",
        "AutoRefresh",
        "LibrarySortScript",
        "AudioOutput",
        "OutputVolume",
        "RewindPreviousTrack",
        "GaplessPlayback",
        "Language",
        "BufferLength",
        "OpenFilesPlaylist",
        "OpenFilesSendTo",
        "SaveRatingToMetadata",
        "SavePlaycountToMetadata",
        "PlayedThreshold",
        "ExternalSortScript",
        "Shutdown",
        "StopAfterCurrent",
        "RGMode",
        "RGType",
        "RGPreAmp",
        "NonRGPreAmp",
        "UseVariousForCompilations",
        "ShuffleAlbumsGroupScript",
        "ShuffleAlbumsSortScript",
        "ActiveTrack",
        "ActiveTrackId",
        "FollowPlaybackQueue",
        "StopIfActivePlaylistDeleted",
        "ResetStopAfterCurrent",
        "PreserveTimestamps",
        "PlaylistSkipMissing",
        "PlaybackQueueStopWhenFinished",
        "OutputBitDepth",
        "OutputDither",
        "ClearPlaybackQueueOnExit",
        "LibraryViewPlaylistSortScript",
        "OverwriteRatingOnReload",
        "OverwritePlaycountOnReload",
        "OpenFileAddDirectory",
        "AddFoldersIgnorePlaylists",
        "PlaylistPreventDuplicates",
        "PlayedThresholdTime"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'CoreSettings'
        QtMocHelpers::EnumData<CoreSettings>(1, 1, QMC::EnumFlags{}).add({
            {    2, CoreSettings::Version },
            {    3, CoreSettings::FirstRun },
            {    4, CoreSettings::PlayMode },
            {    5, CoreSettings::AutoRefresh },
            {    6, CoreSettings::LibrarySortScript },
            {    7, CoreSettings::AudioOutput },
            {    8, CoreSettings::OutputVolume },
            {    9, CoreSettings::RewindPreviousTrack },
            {   10, CoreSettings::GaplessPlayback },
            {   11, CoreSettings::Language },
            {   12, CoreSettings::BufferLength },
            {   13, CoreSettings::OpenFilesPlaylist },
            {   14, CoreSettings::OpenFilesSendTo },
            {   15, CoreSettings::SaveRatingToMetadata },
            {   16, CoreSettings::SavePlaycountToMetadata },
            {   17, CoreSettings::PlayedThreshold },
            {   18, CoreSettings::ExternalSortScript },
            {   19, CoreSettings::Shutdown },
            {   20, CoreSettings::StopAfterCurrent },
            {   21, CoreSettings::RGMode },
            {   22, CoreSettings::RGType },
            {   23, CoreSettings::RGPreAmp },
            {   24, CoreSettings::NonRGPreAmp },
            {   25, CoreSettings::UseVariousForCompilations },
            {   26, CoreSettings::ShuffleAlbumsGroupScript },
            {   27, CoreSettings::ShuffleAlbumsSortScript },
            {   28, CoreSettings::ActiveTrack },
            {   29, CoreSettings::ActiveTrackId },
            {   30, CoreSettings::FollowPlaybackQueue },
            {   31, CoreSettings::StopIfActivePlaylistDeleted },
            {   32, CoreSettings::ResetStopAfterCurrent },
            {   33, CoreSettings::PreserveTimestamps },
            {   34, CoreSettings::PlaylistSkipMissing },
            {   35, CoreSettings::PlaybackQueueStopWhenFinished },
            {   36, CoreSettings::OutputBitDepth },
            {   37, CoreSettings::OutputDither },
            {   38, CoreSettings::ClearPlaybackQueueOnExit },
            {   39, CoreSettings::LibraryViewPlaylistSortScript },
            {   40, CoreSettings::OverwriteRatingOnReload },
            {   41, CoreSettings::OverwritePlaycountOnReload },
            {   42, CoreSettings::OpenFileAddDirectory },
            {   43, CoreSettings::AddFoldersIgnorePlaylists },
            {   44, CoreSettings::PlaylistPreventDuplicates },
            {   45, CoreSettings::PlayedThresholdTime },
        }),
    };
    return QtMocHelpers::metaObjectData<void, qt_meta_tag_ZN6Fooyin8Settings4CoreE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}

static constexpr auto qt_staticMetaObjectContent_ZN6Fooyin8Settings4CoreE =
    Fooyin::Settings::Core::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin8Settings4CoreE_t>();
static constexpr auto qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings4CoreE =
    qt_staticMetaObjectContent_ZN6Fooyin8Settings4CoreE.staticData;
static constexpr auto qt_staticMetaObjectRelocatingContent_ZN6Fooyin8Settings4CoreE =
    qt_staticMetaObjectContent_ZN6Fooyin8Settings4CoreE.relocatingData;

Q_CONSTINIT const QMetaObject Fooyin::Settings::Core::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings4CoreE.stringdata,
    qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings4CoreE.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent_ZN6Fooyin8Settings4CoreE.metaTypes,
    nullptr
} };

QT_WARNING_POP
