/****************************************************************************
** Meta object code from reading C++ file 'librarythreadhandler.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../fooyin/src/core/library/librarythreadhandler.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'librarythreadhandler.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin20LibraryThreadHandlerE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::LibraryThreadHandler::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin20LibraryThreadHandlerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::LibraryThreadHandler",
        "progressChanged",
        "",
        "Fooyin::ScanProgress",
        "progress",
        "scannedTracks",
        "id",
        "Fooyin::TrackList",
        "tracks",
        "playlistLoaded",
        "scanFinished",
        "Fooyin::ScanRequest::Type",
        "type",
        "cancelled",
        "statusChanged",
        "Fooyin::LibraryInfo",
        "library",
        "scanUpdate",
        "Fooyin::ScanResult",
        "result",
        "tracksUpdated",
        "tracksAvailabilityUpdated",
        "tracksStatsUpdated",
        "tracksRemoved",
        "gotTracks"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'progressChanged'
        QtMocHelpers::SignalData<void(const Fooyin::ScanProgress &)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'scannedTracks'
        QtMocHelpers::SignalData<void(int, const Fooyin::TrackList &)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 6 }, { 0x80000000 | 7, 8 },
        }}),
        // Signal 'playlistLoaded'
        QtMocHelpers::SignalData<void(int, const Fooyin::TrackList &)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 6 }, { 0x80000000 | 7, 8 },
        }}),
        // Signal 'scanFinished'
        QtMocHelpers::SignalData<void(int, Fooyin::ScanRequest::Type, bool)>(10, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 6 }, { 0x80000000 | 11, 12 }, { QMetaType::Bool, 13 },
        }}),
        // Signal 'statusChanged'
        QtMocHelpers::SignalData<void(const Fooyin::LibraryInfo &)>(14, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 15, 16 },
        }}),
        // Signal 'scanUpdate'
        QtMocHelpers::SignalData<void(int, Fooyin::ScanRequest::Type, const Fooyin::ScanResult &)>(17, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 6 }, { 0x80000000 | 11, 12 }, { 0x80000000 | 18, 19 },
        }}),
        // Signal 'tracksUpdated'
        QtMocHelpers::SignalData<void(const Fooyin::TrackList &)>(20, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 7, 8 },
        }}),
        // Signal 'tracksAvailabilityUpdated'
        QtMocHelpers::SignalData<void(const Fooyin::TrackList &)>(21, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 7, 8 },
        }}),
        // Signal 'tracksStatsUpdated'
        QtMocHelpers::SignalData<void(const Fooyin::TrackList &)>(22, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 7, 8 },
        }}),
        // Signal 'tracksRemoved'
        QtMocHelpers::SignalData<void(const Fooyin::TrackList &)>(23, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 7, 8 },
        }}),
        // Signal 'gotTracks'
        QtMocHelpers::SignalData<void(Fooyin::TrackList)>(24, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 7, 19 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<LibraryThreadHandler, qt_meta_tag_ZN6Fooyin20LibraryThreadHandlerE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::LibraryThreadHandler::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin20LibraryThreadHandlerE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin20LibraryThreadHandlerE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin20LibraryThreadHandlerE_t>.metaTypes,
    nullptr
} };

void Fooyin::LibraryThreadHandler::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<LibraryThreadHandler *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->progressChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::ScanProgress>>(_a[1]))); break;
        case 1: _t->scannedTracks((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[2]))); break;
        case 2: _t->playlistLoaded((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[2]))); break;
        case 3: _t->scanFinished((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::ScanRequest::Type>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[3]))); break;
        case 4: _t->statusChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::LibraryInfo>>(_a[1]))); break;
        case 5: _t->scanUpdate((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::ScanRequest::Type>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<Fooyin::ScanResult>>(_a[3]))); break;
        case 6: _t->tracksUpdated((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1]))); break;
        case 7: _t->tracksAvailabilityUpdated((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1]))); break;
        case 8: _t->tracksStatsUpdated((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1]))); break;
        case 9: _t->tracksRemoved((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1]))); break;
        case 10: _t->gotTracks((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 0:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::ScanProgress >(); break;
            }
            break;
        case 1:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 2:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 6:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 7:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 8:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 9:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 10:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (LibraryThreadHandler::*)(const Fooyin::ScanProgress & )>(_a, &LibraryThreadHandler::progressChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (LibraryThreadHandler::*)(int , const Fooyin::TrackList & )>(_a, &LibraryThreadHandler::scannedTracks, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (LibraryThreadHandler::*)(int , const Fooyin::TrackList & )>(_a, &LibraryThreadHandler::playlistLoaded, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (LibraryThreadHandler::*)(int , Fooyin::ScanRequest::Type , bool )>(_a, &LibraryThreadHandler::scanFinished, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (LibraryThreadHandler::*)(const Fooyin::LibraryInfo & )>(_a, &LibraryThreadHandler::statusChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (LibraryThreadHandler::*)(int , Fooyin::ScanRequest::Type , const Fooyin::ScanResult & )>(_a, &LibraryThreadHandler::scanUpdate, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (LibraryThreadHandler::*)(const Fooyin::TrackList & )>(_a, &LibraryThreadHandler::tracksUpdated, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (LibraryThreadHandler::*)(const Fooyin::TrackList & )>(_a, &LibraryThreadHandler::tracksAvailabilityUpdated, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (LibraryThreadHandler::*)(const Fooyin::TrackList & )>(_a, &LibraryThreadHandler::tracksStatsUpdated, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (LibraryThreadHandler::*)(const Fooyin::TrackList & )>(_a, &LibraryThreadHandler::tracksRemoved, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (LibraryThreadHandler::*)(Fooyin::TrackList )>(_a, &LibraryThreadHandler::gotTracks, 10))
            return;
    }
}

const QMetaObject *Fooyin::LibraryThreadHandler::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Fooyin::LibraryThreadHandler::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin20LibraryThreadHandlerE_t>.strings))
        return static_cast<void*>(this);
    if (!strcmp(_clname, "PendingTrackCoverProvider"))
        return static_cast< PendingTrackCoverProvider*>(this);
    return QObject::qt_metacast(_clname);
}

int Fooyin::LibraryThreadHandler::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 11)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 11;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 11)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 11;
    }
    return _id;
}

// SIGNAL 0
void Fooyin::LibraryThreadHandler::progressChanged(const Fooyin::ScanProgress & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void Fooyin::LibraryThreadHandler::scannedTracks(int _t1, const Fooyin::TrackList & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1, _t2);
}

// SIGNAL 2
void Fooyin::LibraryThreadHandler::playlistLoaded(int _t1, const Fooyin::TrackList & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1, _t2);
}

// SIGNAL 3
void Fooyin::LibraryThreadHandler::scanFinished(int _t1, Fooyin::ScanRequest::Type _t2, bool _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1, _t2, _t3);
}

// SIGNAL 4
void Fooyin::LibraryThreadHandler::statusChanged(const Fooyin::LibraryInfo & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void Fooyin::LibraryThreadHandler::scanUpdate(int _t1, Fooyin::ScanRequest::Type _t2, const Fooyin::ScanResult & _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1, _t2, _t3);
}

// SIGNAL 6
void Fooyin::LibraryThreadHandler::tracksUpdated(const Fooyin::TrackList & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1);
}

// SIGNAL 7
void Fooyin::LibraryThreadHandler::tracksAvailabilityUpdated(const Fooyin::TrackList & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1);
}

// SIGNAL 8
void Fooyin::LibraryThreadHandler::tracksStatsUpdated(const Fooyin::TrackList & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 8, nullptr, _t1);
}

// SIGNAL 9
void Fooyin::LibraryThreadHandler::tracksRemoved(const Fooyin::TrackList & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 9, nullptr, _t1);
}

// SIGNAL 10
void Fooyin::LibraryThreadHandler::gotTracks(Fooyin::TrackList _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 10, nullptr, _t1);
}
QT_WARNING_POP
