/****************************************************************************
** Meta object code from reading C++ file 'notifysettings.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../../fooyin/src/plugins/notify/settings/notifysettings.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'notifysettings.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin8Settings6NotifyE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::Settings::Notify::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin8Settings6NotifyE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::Settings::Notify",
        "NotifySettings",
        "Enabled",
        "TitleField",
        "BodyField",
        "ShowAlbumArt",
        "Timeout",
        "Controls"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'NotifySettings'
        QtMocHelpers::EnumData<NotifySettings>(1, 1, QMC::EnumFlags{}).add({
            {    2, NotifySettings::Enabled },
            {    3, NotifySettings::TitleField },
            {    4, NotifySettings::BodyField },
            {    5, NotifySettings::ShowAlbumArt },
            {    6, NotifySettings::Timeout },
            {    7, NotifySettings::Controls },
        }),
    };
    return QtMocHelpers::metaObjectData<void, qt_meta_tag_ZN6Fooyin8Settings6NotifyE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}

static constexpr auto qt_staticMetaObjectContent_ZN6Fooyin8Settings6NotifyE =
    Fooyin::Settings::Notify::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin8Settings6NotifyE_t>();
static constexpr auto qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings6NotifyE =
    qt_staticMetaObjectContent_ZN6Fooyin8Settings6NotifyE.staticData;
static constexpr auto qt_staticMetaObjectRelocatingContent_ZN6Fooyin8Settings6NotifyE =
    qt_staticMetaObjectContent_ZN6Fooyin8Settings6NotifyE.relocatingData;

Q_CONSTINIT const QMetaObject Fooyin::Settings::Notify::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings6NotifyE.stringdata,
    qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings6NotifyE.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent_ZN6Fooyin8Settings6NotifyE.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN6Fooyin6NotifyE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::Notify::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin6NotifyE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::Notify",
        "PlaybackControls",
        "PlaybackControlFlag",
        "None",
        "Previous",
        "PlayPause",
        "Next",
        "Default"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // flag 'PlaybackControls'
        QtMocHelpers::EnumData<PlaybackControls>(1, 2, QMC::EnumIsFlag | QMC::EnumIsScoped).add({
            {    3, PlaybackControlFlag::None },
            {    4, PlaybackControlFlag::Previous },
            {    5, PlaybackControlFlag::PlayPause },
            {    6, PlaybackControlFlag::Next },
            {    7, PlaybackControlFlag::Default },
        }),
    };
    return QtMocHelpers::metaObjectData<void, qt_meta_tag_ZN6Fooyin6NotifyE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}

static constexpr auto qt_staticMetaObjectContent_ZN6Fooyin6NotifyE =
    Fooyin::Notify::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin6NotifyE_t>();
static constexpr auto qt_staticMetaObjectStaticContent_ZN6Fooyin6NotifyE =
    qt_staticMetaObjectContent_ZN6Fooyin6NotifyE.staticData;
static constexpr auto qt_staticMetaObjectRelocatingContent_ZN6Fooyin6NotifyE =
    qt_staticMetaObjectContent_ZN6Fooyin6NotifyE.relocatingData;

Q_CONSTINIT const QMetaObject Fooyin::Notify::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent_ZN6Fooyin6NotifyE.stringdata,
    qt_staticMetaObjectStaticContent_ZN6Fooyin6NotifyE.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent_ZN6Fooyin6NotifyE.metaTypes,
    nullptr
} };

QT_WARNING_POP
