/****************************************************************************
** Meta object code from reading C++ file 'playlistcontroller.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../fooyin/src/gui/playlist/playlistcontroller.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'playlistcontroller.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin18PlaylistControllerE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::PlaylistController::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin18PlaylistControllerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::PlaylistController",
        "playlistsLoaded",
        "",
        "currentPlaylistUpdated",
        "Fooyin::Playlist*",
        "playlist",
        "currentPlaylistTracksPatched",
        "Fooyin::PlaylistChangeset",
        "changeSet",
        "currentPlaylistTracksChanged",
        "std::vector<int>",
        "indexes",
        "allNew",
        "currentPlaylistTracksUpdated",
        "currentPlaylistTracksRemoved",
        "currentPlaylistQueueChanged",
        "tracks",
        "playStateChanged",
        "Fooyin::Player::PlayState",
        "state",
        "clipboardChanged",
        "playlistHistoryChanged",
        "playingTrackChanged",
        "Fooyin::PlaylistTrack",
        "track",
        "handleTrackSelectionAction",
        "Fooyin::TrackAction",
        "action"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'playlistsLoaded'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'currentPlaylistUpdated'
        QtMocHelpers::SignalData<void(Fooyin::Playlist *)>(3, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 },
        }}),
        // Signal 'currentPlaylistTracksPatched'
        QtMocHelpers::SignalData<void(const Fooyin::PlaylistChangeset &)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 7, 8 },
        }}),
        // Signal 'currentPlaylistTracksChanged'
        QtMocHelpers::SignalData<void(const std::vector<int> &, bool)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 10, 11 }, { QMetaType::Bool, 12 },
        }}),
        // Signal 'currentPlaylistTracksUpdated'
        QtMocHelpers::SignalData<void(const std::vector<int> &)>(13, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 10, 11 },
        }}),
        // Signal 'currentPlaylistTracksRemoved'
        QtMocHelpers::SignalData<void(const std::vector<int> &)>(14, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 10, 11 },
        }}),
        // Signal 'currentPlaylistQueueChanged'
        QtMocHelpers::SignalData<void(const std::vector<int> &)>(15, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 10, 16 },
        }}),
        // Signal 'playStateChanged'
        QtMocHelpers::SignalData<void(Fooyin::Player::PlayState)>(17, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 18, 19 },
        }}),
        // Signal 'clipboardChanged'
        QtMocHelpers::SignalData<void()>(20, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'playlistHistoryChanged'
        QtMocHelpers::SignalData<void()>(21, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'playingTrackChanged'
        QtMocHelpers::SignalData<void(const Fooyin::PlaylistTrack &)>(22, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 23, 24 },
        }}),
        // Slot 'handleTrackSelectionAction'
        QtMocHelpers::SlotData<void(Fooyin::TrackAction)>(25, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 26, 27 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<PlaylistController, qt_meta_tag_ZN6Fooyin18PlaylistControllerE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::PlaylistController::staticMetaObject = { {
    QMetaObject::SuperData::link<CurrentPlaylistController::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin18PlaylistControllerE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin18PlaylistControllerE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin18PlaylistControllerE_t>.metaTypes,
    nullptr
} };

void Fooyin::PlaylistController::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<PlaylistController *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->playlistsLoaded(); break;
        case 1: _t->currentPlaylistUpdated((*reinterpret_cast<std::add_pointer_t<Fooyin::Playlist*>>(_a[1]))); break;
        case 2: _t->currentPlaylistTracksPatched((*reinterpret_cast<std::add_pointer_t<Fooyin::PlaylistChangeset>>(_a[1]))); break;
        case 3: _t->currentPlaylistTracksChanged((*reinterpret_cast<std::add_pointer_t<std::vector<int>>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 4: _t->currentPlaylistTracksUpdated((*reinterpret_cast<std::add_pointer_t<std::vector<int>>>(_a[1]))); break;
        case 5: _t->currentPlaylistTracksRemoved((*reinterpret_cast<std::add_pointer_t<std::vector<int>>>(_a[1]))); break;
        case 6: _t->currentPlaylistQueueChanged((*reinterpret_cast<std::add_pointer_t<std::vector<int>>>(_a[1]))); break;
        case 7: _t->playStateChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::Player::PlayState>>(_a[1]))); break;
        case 8: _t->clipboardChanged(); break;
        case 9: _t->playlistHistoryChanged(); break;
        case 10: _t->playingTrackChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::PlaylistTrack>>(_a[1]))); break;
        case 11: _t->handleTrackSelectionAction((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackAction>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 2:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::PlaylistChangeset >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (PlaylistController::*)()>(_a, &PlaylistController::playlistsLoaded, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistController::*)(Fooyin::Playlist * )>(_a, &PlaylistController::currentPlaylistUpdated, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistController::*)(const Fooyin::PlaylistChangeset & )>(_a, &PlaylistController::currentPlaylistTracksPatched, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistController::*)(const std::vector<int> & , bool )>(_a, &PlaylistController::currentPlaylistTracksChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistController::*)(const std::vector<int> & )>(_a, &PlaylistController::currentPlaylistTracksUpdated, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistController::*)(const std::vector<int> & )>(_a, &PlaylistController::currentPlaylistTracksRemoved, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistController::*)(const std::vector<int> & )>(_a, &PlaylistController::currentPlaylistQueueChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistController::*)(Fooyin::Player::PlayState )>(_a, &PlaylistController::playStateChanged, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistController::*)()>(_a, &PlaylistController::clipboardChanged, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistController::*)()>(_a, &PlaylistController::playlistHistoryChanged, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistController::*)(const Fooyin::PlaylistTrack & )>(_a, &PlaylistController::playingTrackChanged, 10))
            return;
    }
}

const QMetaObject *Fooyin::PlaylistController::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Fooyin::PlaylistController::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin18PlaylistControllerE_t>.strings))
        return static_cast<void*>(this);
    return CurrentPlaylistController::qt_metacast(_clname);
}

int Fooyin::PlaylistController::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = CurrentPlaylistController::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 12)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 12;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 12)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 12;
    }
    return _id;
}

// SIGNAL 0
void Fooyin::PlaylistController::playlistsLoaded()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void Fooyin::PlaylistController::currentPlaylistUpdated(Fooyin::Playlist * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void Fooyin::PlaylistController::currentPlaylistTracksPatched(const Fooyin::PlaylistChangeset & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void Fooyin::PlaylistController::currentPlaylistTracksChanged(const std::vector<int> & _t1, bool _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1, _t2);
}

// SIGNAL 4
void Fooyin::PlaylistController::currentPlaylistTracksUpdated(const std::vector<int> & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void Fooyin::PlaylistController::currentPlaylistTracksRemoved(const std::vector<int> & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1);
}

// SIGNAL 6
void Fooyin::PlaylistController::currentPlaylistQueueChanged(const std::vector<int> & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1);
}

// SIGNAL 7
void Fooyin::PlaylistController::playStateChanged(Fooyin::Player::PlayState _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1);
}

// SIGNAL 8
void Fooyin::PlaylistController::clipboardChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 8, nullptr);
}

// SIGNAL 9
void Fooyin::PlaylistController::playlistHistoryChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 9, nullptr);
}

// SIGNAL 10
void Fooyin::PlaylistController::playingTrackChanged(const Fooyin::PlaylistTrack & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 10, nullptr, _t1);
}
QT_WARNING_POP
