/*
 * Fooyin
 * Copyright © 2025, Luke Taylor <luket@pm.me>
 *
 * Fooyin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fooyin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fooyin.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#pragma once

#include "fileopsdefs.h"

#include <map>

namespace Fooyin::FileOps {
namespace Settings {
inline constexpr auto ConfirmDelete      = "FileOps/ConfirmDelete";
inline constexpr auto ImmediateDelete    = "FileOps/ImmediateDelete";
inline constexpr auto DeleteEmptyFolders = "FileOps/DeleteEmptyFolders";
} // namespace Settings

std::vector<FileOpPreset> getPresets();
std::map<Operation, std::vector<FileOpPreset>> getMappedPresets();
void savePresets(const std::vector<FileOpPreset>& presets);
} // namespace Fooyin::FileOps
