/****************************************************************************
** Meta object code from reading C++ file 'internalguisettings.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../fooyin/src/gui/internalguisettings.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'internalguisettings.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin8Settings3Gui8InternalE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::Settings::Gui::Internal::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin8Settings3Gui8InternalE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::Settings::Gui::Internal",
        "GuiInternalSettings",
        "EditingMenuLevels",
        "PlaylistAltColours",
        "PlaylistHeader",
        "PlaylistScrollBar",
        "StatusPlayingScript",
        "StatusSelectionScript",
        "StatusShowIcon",
        "StatusShowSelection",
        "SystemIconTheme",
        "SystemStyle",
        "SystemFont",
        "SystemPalette",
        "WindowTitleTrackScript",
        "TrackCoverPaths",
        "TrackCoverDisplayOption",
        "PlaylistImagePadding",
        "PlaylistImagePaddingTop",
        "PixmapCacheSize",
        "EditableLayoutMargin",
        "PlaylistTabsAddButton",
        "ShowTrayIcon",
        "TrayOnClose",
        "PlaylistTabsCloseButton",
        "PlaylistTabsMiddleClose",
        "PlaylistTabsExpand",
        "PlaylistTabsClearButton",
        "PlaylistMiddleClick",
        "InfoDisplayPrefer",
        "LibTreeIconSize",
        "ArtworkSaveMethods",
        "ArtworkAutoSearch",
        "ArtworkTitleField",
        "ArtworkAlbumField",
        "ArtworkArtistField",
        "ArtworkMatchThreshold",
        "ArtworkDownloadThumbSize",
        "ImageAllocationLimit",
        "PlaylistTrackPreloadCount",
        "TrackCoverSourcePreference",
        "PlaylistInlineTagEditing",
        "ContextMenuTrackDisabledSections",
        "ContextMenuPlaylistDisabledSections",
        "ContextMenuTrackLayout",
        "ContextMenuPlaylistLayout",
        "ContextMenuLibraryTreeDisabledSections",
        "ContextMenuLibraryTreeLayout",
        "PropertiesSidebarTrackScript",
        "StatusPlaylistScript",
        "StatusShowPlaylist",
        "TrackCoverThumbnailGroupScript",
        "ContextMenuDirBrowserDisabledSections",
        "ContextMenuDirBrowserLayout",
        "ContextMenuLayoutEditingDisabledSections",
        "ContextMenuLayoutEditingLayout",
        "PlaylistBackgroundImageMode",
        "PlaylistBackgroundCustomImage",
        "PlaylistBackgroundScaling",
        "PlaylistBackgroundPosition",
        "PlaylistBackgroundMaxSize",
        "PlaylistBackgroundBlur",
        "PlaylistBackgroundOpacity",
        "PlaylistBackgroundFadeDuration",
        "PlaylistBackgroundCoverType",
        "NowPlayingOutputEnabled",
        "NowPlayingOutputScript",
        "NowPlayingOutputUpdateEvents",
        "NowPlayingOutputTargets",
        "NowPlayingOutputFilePath",
        "NowPlayingOutputOptions",
        "NowPlayingOutputAppendLineLimit",
        "OutputDeviceRefreshMs"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'GuiInternalSettings'
        QtMocHelpers::EnumData<GuiInternalSettings>(1, 1, QMC::EnumFlags{}).add({
            {    2, GuiInternalSettings::EditingMenuLevels },
            {    3, GuiInternalSettings::PlaylistAltColours },
            {    4, GuiInternalSettings::PlaylistHeader },
            {    5, GuiInternalSettings::PlaylistScrollBar },
            {    6, GuiInternalSettings::StatusPlayingScript },
            {    7, GuiInternalSettings::StatusSelectionScript },
            {    8, GuiInternalSettings::StatusShowIcon },
            {    9, GuiInternalSettings::StatusShowSelection },
            {   10, GuiInternalSettings::SystemIconTheme },
            {   11, GuiInternalSettings::SystemStyle },
            {   12, GuiInternalSettings::SystemFont },
            {   13, GuiInternalSettings::SystemPalette },
            {   14, GuiInternalSettings::WindowTitleTrackScript },
            {   15, GuiInternalSettings::TrackCoverPaths },
            {   16, GuiInternalSettings::TrackCoverDisplayOption },
            {   17, GuiInternalSettings::PlaylistImagePadding },
            {   18, GuiInternalSettings::PlaylistImagePaddingTop },
            {   19, GuiInternalSettings::PixmapCacheSize },
            {   20, GuiInternalSettings::EditableLayoutMargin },
            {   21, GuiInternalSettings::PlaylistTabsAddButton },
            {   22, GuiInternalSettings::ShowTrayIcon },
            {   23, GuiInternalSettings::TrayOnClose },
            {   24, GuiInternalSettings::PlaylistTabsCloseButton },
            {   25, GuiInternalSettings::PlaylistTabsMiddleClose },
            {   26, GuiInternalSettings::PlaylistTabsExpand },
            {   27, GuiInternalSettings::PlaylistTabsClearButton },
            {   28, GuiInternalSettings::PlaylistMiddleClick },
            {   29, GuiInternalSettings::InfoDisplayPrefer },
            {   30, GuiInternalSettings::LibTreeIconSize },
            {   31, GuiInternalSettings::ArtworkSaveMethods },
            {   32, GuiInternalSettings::ArtworkAutoSearch },
            {   33, GuiInternalSettings::ArtworkTitleField },
            {   34, GuiInternalSettings::ArtworkAlbumField },
            {   35, GuiInternalSettings::ArtworkArtistField },
            {   36, GuiInternalSettings::ArtworkMatchThreshold },
            {   37, GuiInternalSettings::ArtworkDownloadThumbSize },
            {   38, GuiInternalSettings::ImageAllocationLimit },
            {   39, GuiInternalSettings::PlaylistTrackPreloadCount },
            {   40, GuiInternalSettings::TrackCoverSourcePreference },
            {   41, GuiInternalSettings::PlaylistInlineTagEditing },
            {   42, GuiInternalSettings::ContextMenuTrackDisabledSections },
            {   43, GuiInternalSettings::ContextMenuPlaylistDisabledSections },
            {   44, GuiInternalSettings::ContextMenuTrackLayout },
            {   45, GuiInternalSettings::ContextMenuPlaylistLayout },
            {   46, GuiInternalSettings::ContextMenuLibraryTreeDisabledSections },
            {   47, GuiInternalSettings::ContextMenuLibraryTreeLayout },
            {   48, GuiInternalSettings::PropertiesSidebarTrackScript },
            {   49, GuiInternalSettings::StatusPlaylistScript },
            {   50, GuiInternalSettings::StatusShowPlaylist },
            {   51, GuiInternalSettings::TrackCoverThumbnailGroupScript },
            {   52, GuiInternalSettings::ContextMenuDirBrowserDisabledSections },
            {   53, GuiInternalSettings::ContextMenuDirBrowserLayout },
            {   54, GuiInternalSettings::ContextMenuLayoutEditingDisabledSections },
            {   55, GuiInternalSettings::ContextMenuLayoutEditingLayout },
            {   56, GuiInternalSettings::PlaylistBackgroundImageMode },
            {   57, GuiInternalSettings::PlaylistBackgroundCustomImage },
            {   58, GuiInternalSettings::PlaylistBackgroundScaling },
            {   59, GuiInternalSettings::PlaylistBackgroundPosition },
            {   60, GuiInternalSettings::PlaylistBackgroundMaxSize },
            {   61, GuiInternalSettings::PlaylistBackgroundBlur },
            {   62, GuiInternalSettings::PlaylistBackgroundOpacity },
            {   63, GuiInternalSettings::PlaylistBackgroundFadeDuration },
            {   64, GuiInternalSettings::PlaylistBackgroundCoverType },
            {   65, GuiInternalSettings::NowPlayingOutputEnabled },
            {   66, GuiInternalSettings::NowPlayingOutputScript },
            {   67, GuiInternalSettings::NowPlayingOutputUpdateEvents },
            {   68, GuiInternalSettings::NowPlayingOutputTargets },
            {   69, GuiInternalSettings::NowPlayingOutputFilePath },
            {   70, GuiInternalSettings::NowPlayingOutputOptions },
            {   71, GuiInternalSettings::NowPlayingOutputAppendLineLimit },
            {   72, GuiInternalSettings::OutputDeviceRefreshMs },
        }),
    };
    return QtMocHelpers::metaObjectData<void, qt_meta_tag_ZN6Fooyin8Settings3Gui8InternalE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}

static constexpr auto qt_staticMetaObjectContent_ZN6Fooyin8Settings3Gui8InternalE =
    Fooyin::Settings::Gui::Internal::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin8Settings3Gui8InternalE_t>();
static constexpr auto qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings3Gui8InternalE =
    qt_staticMetaObjectContent_ZN6Fooyin8Settings3Gui8InternalE.staticData;
static constexpr auto qt_staticMetaObjectRelocatingContent_ZN6Fooyin8Settings3Gui8InternalE =
    qt_staticMetaObjectContent_ZN6Fooyin8Settings3Gui8InternalE.relocatingData;

Q_CONSTINIT const QMetaObject Fooyin::Settings::Gui::Internal::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings3Gui8InternalE.stringdata,
    qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings3Gui8InternalE.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent_ZN6Fooyin8Settings3Gui8InternalE.metaTypes,
    nullptr
} };

QT_WARNING_POP
