/****************************************************************************
** Meta object code from reading C++ file 'playlisthandler.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../../../fooyin/include/core/playlist/playlisthandler.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'playlisthandler.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin15PlaylistHandlerE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::PlaylistHandler::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin15PlaylistHandlerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::PlaylistHandler",
        "playlistsPopulated",
        "",
        "playlistAdded",
        "Fooyin::Playlist*",
        "playlist",
        "playlistRemoved",
        "playlistRenamed",
        "playlistUpdated",
        "playlistIndexChanged",
        "activePlaylistChanged",
        "activePlaylistDeleted",
        "playlistReferencesRemapRequested",
        "Fooyin::UId",
        "fromPlaylistId",
        "toPlaylistId",
        "tracksAdded",
        "Fooyin::TrackList",
        "tracks",
        "index",
        "tracksPatched",
        "Fooyin::PlaylistChangeset",
        "changeSet",
        "Fooyin::PlaylistTrackChangeSource",
        "source",
        "tracksChanged",
        "std::vector<int>",
        "indexes",
        "tracksUpdated",
        "tracksRemoved"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'playlistsPopulated'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'playlistAdded'
        QtMocHelpers::SignalData<void(Fooyin::Playlist *)>(3, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 },
        }}),
        // Signal 'playlistRemoved'
        QtMocHelpers::SignalData<void(Fooyin::Playlist *)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 },
        }}),
        // Signal 'playlistRenamed'
        QtMocHelpers::SignalData<void(Fooyin::Playlist *)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 },
        }}),
        // Signal 'playlistUpdated'
        QtMocHelpers::SignalData<void(Fooyin::Playlist *)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 },
        }}),
        // Signal 'playlistIndexChanged'
        QtMocHelpers::SignalData<void(Fooyin::Playlist *)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 },
        }}),
        // Signal 'activePlaylistChanged'
        QtMocHelpers::SignalData<void(Fooyin::Playlist *)>(10, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 },
        }}),
        // Signal 'activePlaylistDeleted'
        QtMocHelpers::SignalData<void()>(11, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'playlistReferencesRemapRequested'
        QtMocHelpers::SignalData<void(const Fooyin::UId &, const Fooyin::UId &)>(12, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 13, 14 }, { 0x80000000 | 13, 15 },
        }}),
        // Signal 'tracksAdded'
        QtMocHelpers::SignalData<void(Fooyin::Playlist *, const Fooyin::TrackList &, int)>(16, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 }, { 0x80000000 | 17, 18 }, { QMetaType::Int, 19 },
        }}),
        // Signal 'tracksPatched'
        QtMocHelpers::SignalData<void(Fooyin::Playlist *, const Fooyin::PlaylistChangeset &, Fooyin::PlaylistTrackChangeSource)>(20, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 }, { 0x80000000 | 21, 22 }, { 0x80000000 | 23, 24 },
        }}),
        // Signal 'tracksChanged'
        QtMocHelpers::SignalData<void(Fooyin::Playlist *, const std::vector<int> &, Fooyin::PlaylistTrackChangeSource)>(25, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 }, { 0x80000000 | 26, 27 }, { 0x80000000 | 23, 24 },
        }}),
        // Signal 'tracksUpdated'
        QtMocHelpers::SignalData<void(Fooyin::Playlist *, const std::vector<int> &)>(28, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 }, { 0x80000000 | 26, 27 },
        }}),
        // Signal 'tracksRemoved'
        QtMocHelpers::SignalData<void(Fooyin::Playlist *, const std::vector<int> &)>(29, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 }, { 0x80000000 | 26, 27 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<PlaylistHandler, qt_meta_tag_ZN6Fooyin15PlaylistHandlerE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::PlaylistHandler::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin15PlaylistHandlerE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin15PlaylistHandlerE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin15PlaylistHandlerE_t>.metaTypes,
    nullptr
} };

void Fooyin::PlaylistHandler::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<PlaylistHandler *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->playlistsPopulated(); break;
        case 1: _t->playlistAdded((*reinterpret_cast<std::add_pointer_t<Fooyin::Playlist*>>(_a[1]))); break;
        case 2: _t->playlistRemoved((*reinterpret_cast<std::add_pointer_t<Fooyin::Playlist*>>(_a[1]))); break;
        case 3: _t->playlistRenamed((*reinterpret_cast<std::add_pointer_t<Fooyin::Playlist*>>(_a[1]))); break;
        case 4: _t->playlistUpdated((*reinterpret_cast<std::add_pointer_t<Fooyin::Playlist*>>(_a[1]))); break;
        case 5: _t->playlistIndexChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::Playlist*>>(_a[1]))); break;
        case 6: _t->activePlaylistChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::Playlist*>>(_a[1]))); break;
        case 7: _t->activePlaylistDeleted(); break;
        case 8: _t->playlistReferencesRemapRequested((*reinterpret_cast<std::add_pointer_t<Fooyin::UId>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::UId>>(_a[2]))); break;
        case 9: _t->tracksAdded((*reinterpret_cast<std::add_pointer_t<Fooyin::Playlist*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[3]))); break;
        case 10: _t->tracksPatched((*reinterpret_cast<std::add_pointer_t<Fooyin::Playlist*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::PlaylistChangeset>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<Fooyin::PlaylistTrackChangeSource>>(_a[3]))); break;
        case 11: _t->tracksChanged((*reinterpret_cast<std::add_pointer_t<Fooyin::Playlist*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<std::vector<int>>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<Fooyin::PlaylistTrackChangeSource>>(_a[3]))); break;
        case 12: _t->tracksUpdated((*reinterpret_cast<std::add_pointer_t<Fooyin::Playlist*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<std::vector<int>>>(_a[2]))); break;
        case 13: _t->tracksRemoved((*reinterpret_cast<std::add_pointer_t<Fooyin::Playlist*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<std::vector<int>>>(_a[2]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 9:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 10:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::PlaylistChangeset >(); break;
            case 2:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::PlaylistTrackChangeSource >(); break;
            }
            break;
        case 11:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 2:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::PlaylistTrackChangeSource >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (PlaylistHandler::*)()>(_a, &PlaylistHandler::playlistsPopulated, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistHandler::*)(Fooyin::Playlist * )>(_a, &PlaylistHandler::playlistAdded, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistHandler::*)(Fooyin::Playlist * )>(_a, &PlaylistHandler::playlistRemoved, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistHandler::*)(Fooyin::Playlist * )>(_a, &PlaylistHandler::playlistRenamed, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistHandler::*)(Fooyin::Playlist * )>(_a, &PlaylistHandler::playlistUpdated, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistHandler::*)(Fooyin::Playlist * )>(_a, &PlaylistHandler::playlistIndexChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistHandler::*)(Fooyin::Playlist * )>(_a, &PlaylistHandler::activePlaylistChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistHandler::*)()>(_a, &PlaylistHandler::activePlaylistDeleted, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistHandler::*)(const Fooyin::UId & , const Fooyin::UId & )>(_a, &PlaylistHandler::playlistReferencesRemapRequested, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistHandler::*)(Fooyin::Playlist * , const Fooyin::TrackList & , int )>(_a, &PlaylistHandler::tracksAdded, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistHandler::*)(Fooyin::Playlist * , const Fooyin::PlaylistChangeset & , Fooyin::PlaylistTrackChangeSource )>(_a, &PlaylistHandler::tracksPatched, 10))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistHandler::*)(Fooyin::Playlist * , const std::vector<int> & , Fooyin::PlaylistTrackChangeSource )>(_a, &PlaylistHandler::tracksChanged, 11))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistHandler::*)(Fooyin::Playlist * , const std::vector<int> & )>(_a, &PlaylistHandler::tracksUpdated, 12))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistHandler::*)(Fooyin::Playlist * , const std::vector<int> & )>(_a, &PlaylistHandler::tracksRemoved, 13))
            return;
    }
}

const QMetaObject *Fooyin::PlaylistHandler::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Fooyin::PlaylistHandler::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin15PlaylistHandlerE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int Fooyin::PlaylistHandler::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 14)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 14;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 14)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 14;
    }
    return _id;
}

// SIGNAL 0
void Fooyin::PlaylistHandler::playlistsPopulated()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void Fooyin::PlaylistHandler::playlistAdded(Fooyin::Playlist * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void Fooyin::PlaylistHandler::playlistRemoved(Fooyin::Playlist * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void Fooyin::PlaylistHandler::playlistRenamed(Fooyin::Playlist * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void Fooyin::PlaylistHandler::playlistUpdated(Fooyin::Playlist * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void Fooyin::PlaylistHandler::playlistIndexChanged(Fooyin::Playlist * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1);
}

// SIGNAL 6
void Fooyin::PlaylistHandler::activePlaylistChanged(Fooyin::Playlist * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1);
}

// SIGNAL 7
void Fooyin::PlaylistHandler::activePlaylistDeleted()
{
    QMetaObject::activate(this, &staticMetaObject, 7, nullptr);
}

// SIGNAL 8
void Fooyin::PlaylistHandler::playlistReferencesRemapRequested(const Fooyin::UId & _t1, const Fooyin::UId & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 8, nullptr, _t1, _t2);
}

// SIGNAL 9
void Fooyin::PlaylistHandler::tracksAdded(Fooyin::Playlist * _t1, const Fooyin::TrackList & _t2, int _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 9, nullptr, _t1, _t2, _t3);
}

// SIGNAL 10
void Fooyin::PlaylistHandler::tracksPatched(Fooyin::Playlist * _t1, const Fooyin::PlaylistChangeset & _t2, Fooyin::PlaylistTrackChangeSource _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 10, nullptr, _t1, _t2, _t3);
}

// SIGNAL 11
void Fooyin::PlaylistHandler::tracksChanged(Fooyin::Playlist * _t1, const std::vector<int> & _t2, Fooyin::PlaylistTrackChangeSource _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 11, nullptr, _t1, _t2, _t3);
}

// SIGNAL 12
void Fooyin::PlaylistHandler::tracksUpdated(Fooyin::Playlist * _t1, const std::vector<int> & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 12, nullptr, _t1, _t2);
}

// SIGNAL 13
void Fooyin::PlaylistHandler::tracksRemoved(Fooyin::Playlist * _t1, const std::vector<int> & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 13, nullptr, _t1, _t2);
}
QT_WARNING_POP
