/****************************************************************************
** Meta object code from reading C++ file 'trackselectioncontroller.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../../fooyin/include/gui/trackselectioncontroller.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'trackselectioncontroller.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin24TrackSelectionControllerE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::TrackSelectionController::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin24TrackSelectionControllerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::TrackSelectionController",
        "actionExecuted",
        "",
        "Fooyin::TrackAction",
        "action",
        "selectionChanged",
        "displaySelectionChanged",
        "requestPropertiesDialog",
        "Fooyin::TrackList",
        "tracks",
        "requestArtworkSearch",
        "quick",
        "requestArtworkAttach",
        "Fooyin::Track::Cover",
        "type",
        "filepath",
        "requestArtworkRemoval",
        "tracksUpdated",
        "tracksRemoved"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'actionExecuted'
        QtMocHelpers::SignalData<void(Fooyin::TrackAction)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'selectionChanged'
        QtMocHelpers::SignalData<void()>(5, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'displaySelectionChanged'
        QtMocHelpers::SignalData<void()>(6, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'requestPropertiesDialog'
        QtMocHelpers::SignalData<void(const Fooyin::TrackList &)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 8, 9 },
        }}),
        // Signal 'requestArtworkSearch'
        QtMocHelpers::SignalData<void(const Fooyin::TrackList &, bool)>(10, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 8, 9 }, { QMetaType::Bool, 11 },
        }}),
        // Signal 'requestArtworkAttach'
        QtMocHelpers::SignalData<void(const Fooyin::TrackList &, Fooyin::Track::Cover, const QString &)>(12, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 8, 9 }, { 0x80000000 | 13, 14 }, { QMetaType::QString, 15 },
        }}),
        // Signal 'requestArtworkRemoval'
        QtMocHelpers::SignalData<void(const Fooyin::TrackList &)>(16, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 8, 9 },
        }}),
        // Slot 'tracksUpdated'
        QtMocHelpers::SlotData<void(const Fooyin::TrackList &)>(17, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 8, 9 },
        }}),
        // Slot 'tracksRemoved'
        QtMocHelpers::SlotData<void(const Fooyin::TrackList &)>(18, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 8, 9 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<TrackSelectionController, qt_meta_tag_ZN6Fooyin24TrackSelectionControllerE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::TrackSelectionController::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin24TrackSelectionControllerE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin24TrackSelectionControllerE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin24TrackSelectionControllerE_t>.metaTypes,
    nullptr
} };

void Fooyin::TrackSelectionController::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<TrackSelectionController *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->actionExecuted((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackAction>>(_a[1]))); break;
        case 1: _t->selectionChanged(); break;
        case 2: _t->displaySelectionChanged(); break;
        case 3: _t->requestPropertiesDialog((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1]))); break;
        case 4: _t->requestArtworkSearch((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 5: _t->requestArtworkAttach((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::Track::Cover>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[3]))); break;
        case 6: _t->requestArtworkRemoval((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1]))); break;
        case 7: _t->tracksUpdated((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1]))); break;
        case 8: _t->tracksRemoved((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 3:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 4:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 5:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 6:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 7:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 8:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (TrackSelectionController::*)(Fooyin::TrackAction )>(_a, &TrackSelectionController::actionExecuted, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (TrackSelectionController::*)()>(_a, &TrackSelectionController::selectionChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (TrackSelectionController::*)()>(_a, &TrackSelectionController::displaySelectionChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (TrackSelectionController::*)(const Fooyin::TrackList & )>(_a, &TrackSelectionController::requestPropertiesDialog, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (TrackSelectionController::*)(const Fooyin::TrackList & , bool )>(_a, &TrackSelectionController::requestArtworkSearch, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (TrackSelectionController::*)(const Fooyin::TrackList & , Fooyin::Track::Cover , const QString & )>(_a, &TrackSelectionController::requestArtworkAttach, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (TrackSelectionController::*)(const Fooyin::TrackList & )>(_a, &TrackSelectionController::requestArtworkRemoval, 6))
            return;
    }
}

const QMetaObject *Fooyin::TrackSelectionController::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Fooyin::TrackSelectionController::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin24TrackSelectionControllerE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int Fooyin::TrackSelectionController::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 9)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 9;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 9)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 9;
    }
    return _id;
}

// SIGNAL 0
void Fooyin::TrackSelectionController::actionExecuted(Fooyin::TrackAction _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void Fooyin::TrackSelectionController::selectionChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void Fooyin::TrackSelectionController::displaySelectionChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void Fooyin::TrackSelectionController::requestPropertiesDialog(const Fooyin::TrackList & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void Fooyin::TrackSelectionController::requestArtworkSearch(const Fooyin::TrackList & _t1, bool _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1, _t2);
}

// SIGNAL 5
void Fooyin::TrackSelectionController::requestArtworkAttach(const Fooyin::TrackList & _t1, Fooyin::Track::Cover _t2, const QString & _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1, _t2, _t3);
}

// SIGNAL 6
void Fooyin::TrackSelectionController::requestArtworkRemoval(const Fooyin::TrackList & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1);
}
QT_WARNING_POP
