/****************************************************************************
** Meta object code from reading C++ file 'discordsettings.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../../fooyin/src/plugins/discord/settings/discordsettings.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'discordsettings.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin8Settings7DiscordE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::Settings::Discord::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin8Settings7DiscordE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::Settings::Discord",
        "DiscordSettings",
        "ClientId",
        "ArtistField",
        "AlbumField",
        "TitleField",
        "ShowStateIcon",
        "DiscordEnabled",
        "ClearOnPause"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'DiscordSettings'
        QtMocHelpers::EnumData<DiscordSettings>(1, 1, QMC::EnumFlags{}).add({
            {    2, DiscordSettings::ClientId },
            {    3, DiscordSettings::ArtistField },
            {    4, DiscordSettings::AlbumField },
            {    5, DiscordSettings::TitleField },
            {    6, DiscordSettings::ShowStateIcon },
            {    7, DiscordSettings::DiscordEnabled },
            {    8, DiscordSettings::ClearOnPause },
        }),
    };
    return QtMocHelpers::metaObjectData<void, qt_meta_tag_ZN6Fooyin8Settings7DiscordE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}

static constexpr auto qt_staticMetaObjectContent_ZN6Fooyin8Settings7DiscordE =
    Fooyin::Settings::Discord::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin8Settings7DiscordE_t>();
static constexpr auto qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings7DiscordE =
    qt_staticMetaObjectContent_ZN6Fooyin8Settings7DiscordE.staticData;
static constexpr auto qt_staticMetaObjectRelocatingContent_ZN6Fooyin8Settings7DiscordE =
    qt_staticMetaObjectContent_ZN6Fooyin8Settings7DiscordE.relocatingData;

Q_CONSTINIT const QMetaObject Fooyin::Settings::Discord::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings7DiscordE.stringdata,
    qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings7DiscordE.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent_ZN6Fooyin8Settings7DiscordE.metaTypes,
    nullptr
} };

QT_WARNING_POP
