/****************************************************************************
** Meta object code from reading C++ file 'playlistorganisermodel.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../fooyin/src/gui/playlist/organiser/playlistorganisermodel.h"
#include <QtCore/qmetatype.h>
#include <QtCore/QList>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'playlistorganisermodel.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin22PlaylistOrganiserModelE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::PlaylistOrganiserModel::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin22PlaylistOrganiserModelE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::PlaylistOrganiserModel",
        "filesDroppedOnPlaylist",
        "",
        "QList<QUrl>",
        "urls",
        "Fooyin::UId",
        "id",
        "filesDroppedOnGroup",
        "group",
        "index",
        "tracksDroppedOnPlaylist",
        "std::vector<int>",
        "trackIds",
        "tracksDroppedOnGroup"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'filesDroppedOnPlaylist'
        QtMocHelpers::SignalData<void(const QList<QUrl> &, const Fooyin::UId &)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 }, { 0x80000000 | 5, 6 },
        }}),
        // Signal 'filesDroppedOnGroup'
        QtMocHelpers::SignalData<void(const QList<QUrl> &, const QString &, int)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 }, { QMetaType::QString, 8 }, { QMetaType::Int, 9 },
        }}),
        // Signal 'tracksDroppedOnPlaylist'
        QtMocHelpers::SignalData<void(const std::vector<int> &, const Fooyin::UId &)>(10, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 11, 12 }, { 0x80000000 | 5, 6 },
        }}),
        // Signal 'tracksDroppedOnGroup'
        QtMocHelpers::SignalData<void(const std::vector<int> &, const QString &, int)>(13, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 11, 12 }, { QMetaType::QString, 8 }, { QMetaType::Int, 9 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<PlaylistOrganiserModel, qt_meta_tag_ZN6Fooyin22PlaylistOrganiserModelE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::PlaylistOrganiserModel::staticMetaObject = { {
    QMetaObject::SuperData::link<TreeModel<PlaylistOrganiserItem>::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin22PlaylistOrganiserModelE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin22PlaylistOrganiserModelE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin22PlaylistOrganiserModelE_t>.metaTypes,
    nullptr
} };

void Fooyin::PlaylistOrganiserModel::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<PlaylistOrganiserModel *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->filesDroppedOnPlaylist((*reinterpret_cast<std::add_pointer_t<QList<QUrl>>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::UId>>(_a[2]))); break;
        case 1: _t->filesDroppedOnGroup((*reinterpret_cast<std::add_pointer_t<QList<QUrl>>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[3]))); break;
        case 2: _t->tracksDroppedOnPlaylist((*reinterpret_cast<std::add_pointer_t<std::vector<int>>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::UId>>(_a[2]))); break;
        case 3: _t->tracksDroppedOnGroup((*reinterpret_cast<std::add_pointer_t<std::vector<int>>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[3]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (PlaylistOrganiserModel::*)(const QList<QUrl> & , const Fooyin::UId & )>(_a, &PlaylistOrganiserModel::filesDroppedOnPlaylist, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistOrganiserModel::*)(const QList<QUrl> & , const QString & , int )>(_a, &PlaylistOrganiserModel::filesDroppedOnGroup, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistOrganiserModel::*)(const std::vector<int> & , const Fooyin::UId & )>(_a, &PlaylistOrganiserModel::tracksDroppedOnPlaylist, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistOrganiserModel::*)(const std::vector<int> & , const QString & , int )>(_a, &PlaylistOrganiserModel::tracksDroppedOnGroup, 3))
            return;
    }
}

const QMetaObject *Fooyin::PlaylistOrganiserModel::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Fooyin::PlaylistOrganiserModel::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin22PlaylistOrganiserModelE_t>.strings))
        return static_cast<void*>(this);
    return TreeModel<PlaylistOrganiserItem>::qt_metacast(_clname);
}

int Fooyin::PlaylistOrganiserModel::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = TreeModel<PlaylistOrganiserItem>::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 4)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 4;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 4)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 4;
    }
    return _id;
}

// SIGNAL 0
void Fooyin::PlaylistOrganiserModel::filesDroppedOnPlaylist(const QList<QUrl> & _t1, const Fooyin::UId & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1, _t2);
}

// SIGNAL 1
void Fooyin::PlaylistOrganiserModel::filesDroppedOnGroup(const QList<QUrl> & _t1, const QString & _t2, int _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1, _t2, _t3);
}

// SIGNAL 2
void Fooyin::PlaylistOrganiserModel::tracksDroppedOnPlaylist(const std::vector<int> & _t1, const Fooyin::UId & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1, _t2);
}

// SIGNAL 3
void Fooyin::PlaylistOrganiserModel::tracksDroppedOnGroup(const std::vector<int> & _t1, const QString & _t2, int _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1, _t2, _t3);
}
QT_WARNING_POP
