/*
 * This file was generated by qdbusxml2cpp version 0.8
 * Source file was org.mpris.MediaPlayer2.xml
 *
 * qdbusxml2cpp is Copyright (C) The Qt Company Ltd. and other contributors.
 *
 * This is an auto-generated file.
 * Do not edit! All changes made to it will be lost.
 */

#include "mprisroot.h"
#include <QtCore/QMetaObject>
#include <QtCore/QByteArray>
#include <QtCore/QList>
#include <QtCore/QMap>
#include <QtCore/QString>
#include <QtCore/QStringList>
#include <QtCore/QVariant>

/*
 * Implementation of adaptor class MprisRoot
 */

MprisRoot::MprisRoot(Fooyin::Mpris::MprisPlugin *parent)
    : QDBusAbstractAdaptor(parent)
{
    // constructor
    setAutoRelaySignals(true);
}

MprisRoot::~MprisRoot()
{
    // destructor
}

bool MprisRoot::canQuit() const
{
    // get the value of property CanQuit
    return qvariant_cast< bool >(parent()->property("CanQuit"));
}

bool MprisRoot::canRaise() const
{
    // get the value of property CanRaise
    return qvariant_cast< bool >(parent()->property("CanRaise"));
}

bool MprisRoot::canSetFullscreen() const
{
    // get the value of property CanSetFullscreen
    return qvariant_cast< bool >(parent()->property("CanSetFullscreen"));
}

QString MprisRoot::desktopEntry() const
{
    // get the value of property DesktopEntry
    return qvariant_cast< QString >(parent()->property("DesktopEntry"));
}

bool MprisRoot::fullscreen() const
{
    // get the value of property Fullscreen
    return qvariant_cast< bool >(parent()->property("Fullscreen"));
}

void MprisRoot::setFullscreen(bool value)
{
    // set the value of property Fullscreen
    parent()->setProperty("Fullscreen", QVariant::fromValue(value));
}

QString MprisRoot::identity() const
{
    // get the value of property Identity
    return qvariant_cast< QString >(parent()->property("Identity"));
}

QStringList MprisRoot::supportedMimeTypes() const
{
    // get the value of property SupportedMimeTypes
    return qvariant_cast< QStringList >(parent()->property("SupportedMimeTypes"));
}

QStringList MprisRoot::supportedUriSchemes() const
{
    // get the value of property SupportedUriSchemes
    return qvariant_cast< QStringList >(parent()->property("SupportedUriSchemes"));
}

void MprisRoot::Quit()
{
    // handle method call org.mpris.MediaPlayer2.Quit
    parent()->Quit();
}

void MprisRoot::Raise()
{
    // handle method call org.mpris.MediaPlayer2.Raise
    parent()->Raise();
}


#include "moc_mprisroot.cpp"
