/*
 * Fooyin
 * Copyright © 2024, Luke Taylor <luket@pm.me>
 *
 * Fooyin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fooyin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fooyin.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#pragma once

namespace Fooyin::Gme {
constexpr auto DefaultLengthMinutes        = 3.0;
constexpr auto DefaultLength               = "Gme/MaxLength";
constexpr auto DefaultLoopCount            = 2;
constexpr auto LoopCount                   = "Gme/LoopCount";
constexpr auto DefaultFadeLength           = 4000;
constexpr auto FadeLength                  = "Gme/FadeLength";
constexpr auto DefaultFadeNonLoopingTracks = false;
constexpr auto FadeNonLoopingTracks        = "Gme/FadeNonLoopingTracks";
} // namespace Fooyin::Gme
