/*
 * Fooyin
 * Copyright © 2024, Luke Taylor <luket@pm.me>
 *
 * Fooyin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fooyin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fooyin.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "queuevieweritem.h"

#include <core/scripting/scriptparser.h>
#include <gui/scripting/richtextutils.h>
#include <gui/scripting/scriptformatter.h>

namespace Fooyin {
QueueViewerItem::QueueViewerItem(PlaylistTrack track)
    : m_track{std::move(track)}
{ }

QString QueueViewerItem::title() const
{
    return m_title.joinedText();
}

QString QueueViewerItem::subtitle() const
{
    return m_subtitle.joinedText();
}

const RichText& QueueViewerItem::richTitle() const
{
    return m_title;
}

const RichText& QueueViewerItem::richSubtitle() const
{
    return m_subtitle;
}

PlaylistTrack QueueViewerItem::track() const
{
    return m_track;
}

void QueueViewerItem::generateTitle(ScriptParser* parser, ScriptFormatter* formatter, const QString& leftScript,
                                    const QString& rightScript, const ScriptContext& context)
{
    if(!parser || !formatter) {
        return;
    }

    m_title    = trimRichText(formatter->evaluate(parser->evaluate(leftScript, m_track.track, context)));
    m_subtitle = trimRichText(formatter->evaluate(parser->evaluate(rightScript, m_track.track, context)));
}
} // namespace Fooyin
