/****************************************************************************
** Meta object code from reading C++ file 'settingsentry.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../../../fooyin/include/utils/settings/settingsentry.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'settingsentry.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin13SettingsEntryE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::SettingsEntry::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin13SettingsEntryE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::SettingsEntry",
        "settingChangedVariant",
        "",
        "QVariant",
        "value",
        "settingChangedBool",
        "settingChangedInt",
        "settingChangedFloat",
        "settingChangedDouble",
        "settingChangedString",
        "settingChangedStringList",
        "settingChangedByteArray"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'settingChangedVariant'
        QtMocHelpers::SignalData<void(const QVariant &)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'settingChangedBool'
        QtMocHelpers::SignalData<void(bool)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 4 },
        }}),
        // Signal 'settingChangedInt'
        QtMocHelpers::SignalData<void(int)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 4 },
        }}),
        // Signal 'settingChangedFloat'
        QtMocHelpers::SignalData<void(float)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 4 },
        }}),
        // Signal 'settingChangedDouble'
        QtMocHelpers::SignalData<void(double)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Double, 4 },
        }}),
        // Signal 'settingChangedString'
        QtMocHelpers::SignalData<void(const QString &)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 4 },
        }}),
        // Signal 'settingChangedStringList'
        QtMocHelpers::SignalData<void(const QStringList &)>(10, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QStringList, 4 },
        }}),
        // Signal 'settingChangedByteArray'
        QtMocHelpers::SignalData<void(const QByteArray &)>(11, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QByteArray, 4 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<SettingsEntry, qt_meta_tag_ZN6Fooyin13SettingsEntryE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::SettingsEntry::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin13SettingsEntryE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin13SettingsEntryE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin13SettingsEntryE_t>.metaTypes,
    nullptr
} };

void Fooyin::SettingsEntry::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<SettingsEntry *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->settingChangedVariant((*reinterpret_cast<std::add_pointer_t<QVariant>>(_a[1]))); break;
        case 1: _t->settingChangedBool((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 2: _t->settingChangedInt((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 3: _t->settingChangedFloat((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 4: _t->settingChangedDouble((*reinterpret_cast<std::add_pointer_t<double>>(_a[1]))); break;
        case 5: _t->settingChangedString((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 6: _t->settingChangedStringList((*reinterpret_cast<std::add_pointer_t<QStringList>>(_a[1]))); break;
        case 7: _t->settingChangedByteArray((*reinterpret_cast<std::add_pointer_t<QByteArray>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (SettingsEntry::*)(const QVariant & )>(_a, &SettingsEntry::settingChangedVariant, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (SettingsEntry::*)(bool )>(_a, &SettingsEntry::settingChangedBool, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (SettingsEntry::*)(int )>(_a, &SettingsEntry::settingChangedInt, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (SettingsEntry::*)(float )>(_a, &SettingsEntry::settingChangedFloat, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (SettingsEntry::*)(double )>(_a, &SettingsEntry::settingChangedDouble, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (SettingsEntry::*)(const QString & )>(_a, &SettingsEntry::settingChangedString, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (SettingsEntry::*)(const QStringList & )>(_a, &SettingsEntry::settingChangedStringList, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (SettingsEntry::*)(const QByteArray & )>(_a, &SettingsEntry::settingChangedByteArray, 7))
            return;
    }
}

const QMetaObject *Fooyin::SettingsEntry::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Fooyin::SettingsEntry::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin13SettingsEntryE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int Fooyin::SettingsEntry::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 8)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 8;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 8)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 8;
    }
    return _id;
}

// SIGNAL 0
void Fooyin::SettingsEntry::settingChangedVariant(const QVariant & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void Fooyin::SettingsEntry::settingChangedBool(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void Fooyin::SettingsEntry::settingChangedInt(int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void Fooyin::SettingsEntry::settingChangedFloat(float _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void Fooyin::SettingsEntry::settingChangedDouble(double _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void Fooyin::SettingsEntry::settingChangedString(const QString & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1);
}

// SIGNAL 6
void Fooyin::SettingsEntry::settingChangedStringList(const QStringList & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1);
}

// SIGNAL 7
void Fooyin::SettingsEntry::settingChangedByteArray(const QByteArray & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1);
}
QT_WARNING_POP
