/****************************************************************************
** Meta object code from reading C++ file 'guisettings.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../fooyin/include/gui/guisettings.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'guisettings.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin8Settings3GuiE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::Settings::Gui::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin8Settings3GuiE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::Settings::Gui",
        "GuiSettings",
        "LayoutEditing",
        "StartupBehaviour",
        "WaitForTracks",
        "IconTheme",
        "ResolvedAppStyle",
        "CursorFollowsPlayback",
        "PlaybackFollowsCursor",
        "ToolButtonStyle",
        "MainWindowPixelRatio",
        "StarRatingSize",
        "Style",
        "SeekStepSmall",
        "SeekStepLarge",
        "ShowStatusTips",
        "CustomTheme",
        "ShowSplitterHandles",
        "LockSplitterHandles",
        "SplitterHandleSize",
        "VolumeStep",
        "SearchSuccessClear",
        "SearchAutoDelay",
        "SearchPlaylistName",
        "SearchPlaylistAppendSearch",
        "SearchSuccessFocus",
        "SearchErrorBg",
        "SearchErrorFg",
        "SearchSuccessClose",
        "ShowMenuBar",
        "RatingFullStarSymbol",
        "RatingHalfStarSymbol",
        "RatingEmptyStarSymbol",
        "SeekBarMouseFocus",
        "PlaylistIntegratedSearch",
        "PlaylistSearchMode",
        "PlaylistSearchScript",
        "DragOnlyAfterSelect"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'GuiSettings'
        QtMocHelpers::EnumData<GuiSettings>(1, 1, QMC::EnumFlags{}).add({
            {    2, GuiSettings::LayoutEditing },
            {    3, GuiSettings::StartupBehaviour },
            {    4, GuiSettings::WaitForTracks },
            {    5, GuiSettings::IconTheme },
            {    6, GuiSettings::ResolvedAppStyle },
            {    7, GuiSettings::CursorFollowsPlayback },
            {    8, GuiSettings::PlaybackFollowsCursor },
            {    9, GuiSettings::ToolButtonStyle },
            {   10, GuiSettings::MainWindowPixelRatio },
            {   11, GuiSettings::StarRatingSize },
            {   12, GuiSettings::Style },
            {   13, GuiSettings::SeekStepSmall },
            {   14, GuiSettings::SeekStepLarge },
            {   15, GuiSettings::ShowStatusTips },
            {   16, GuiSettings::CustomTheme },
            {   17, GuiSettings::ShowSplitterHandles },
            {   18, GuiSettings::LockSplitterHandles },
            {   19, GuiSettings::SplitterHandleSize },
            {   20, GuiSettings::VolumeStep },
            {   21, GuiSettings::SearchSuccessClear },
            {   22, GuiSettings::SearchAutoDelay },
            {   23, GuiSettings::SearchPlaylistName },
            {   24, GuiSettings::SearchPlaylistAppendSearch },
            {   25, GuiSettings::SearchSuccessFocus },
            {   26, GuiSettings::SearchErrorBg },
            {   27, GuiSettings::SearchErrorFg },
            {   28, GuiSettings::SearchSuccessClose },
            {   29, GuiSettings::ShowMenuBar },
            {   30, GuiSettings::RatingFullStarSymbol },
            {   31, GuiSettings::RatingHalfStarSymbol },
            {   32, GuiSettings::RatingEmptyStarSymbol },
            {   33, GuiSettings::SeekBarMouseFocus },
            {   34, GuiSettings::PlaylistIntegratedSearch },
            {   35, GuiSettings::PlaylistSearchMode },
            {   36, GuiSettings::PlaylistSearchScript },
            {   37, GuiSettings::DragOnlyAfterSelect },
        }),
    };
    return QtMocHelpers::metaObjectData<void, qt_meta_tag_ZN6Fooyin8Settings3GuiE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}

static constexpr auto qt_staticMetaObjectContent_ZN6Fooyin8Settings3GuiE =
    Fooyin::Settings::Gui::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin8Settings3GuiE_t>();
static constexpr auto qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings3GuiE =
    qt_staticMetaObjectContent_ZN6Fooyin8Settings3GuiE.staticData;
static constexpr auto qt_staticMetaObjectRelocatingContent_ZN6Fooyin8Settings3GuiE =
    qt_staticMetaObjectContent_ZN6Fooyin8Settings3GuiE.relocatingData;

Q_CONSTINIT const QMetaObject Fooyin::Settings::Gui::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings3GuiE.stringdata,
    qt_staticMetaObjectStaticContent_ZN6Fooyin8Settings3GuiE.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent_ZN6Fooyin8Settings3GuiE.metaTypes,
    nullptr
} };

QT_WARNING_POP
