/****************************************************************************
** Meta object code from reading C++ file 'mprisplugin.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../../fooyin/src/plugins/mpris/mprisplugin.h"
#include <QtCore/qmetatype.h>
#include <QtCore/qplugin.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'mprisplugin.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin5Mpris11MprisPluginE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::Mpris::MprisPlugin::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin5Mpris11MprisPluginE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::Mpris::MprisPlugin",
        "Seeked",
        "",
        "position",
        "fullscreenChanged",
        "fullscreen",
        "volumeChanged",
        "volume",
        "loopStatusChanged",
        "status",
        "shuffleChanged",
        "shuffle",
        "reloadMetadata",
        "Identity",
        "DesktopEntry",
        "CanRaise",
        "CanQuit",
        "Fullscreen",
        "CanSetFullscreen",
        "SupportedUriSchemes",
        "SupportedMimeTypes",
        "CanControl",
        "CanGoNext",
        "CanGoPrevious",
        "CanPause",
        "CanPlay",
        "CanSeek",
        "Volume",
        "PlaybackStatus",
        "LoopStatus",
        "Shuffle",
        "Metadata",
        "QVariantMap",
        "Position",
        "int64_t",
        "MinimumRate",
        "MaximumRate"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'Seeked'
        QtMocHelpers::SignalData<void(qlonglong)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::LongLong, 3 },
        }}),
        // Signal 'fullscreenChanged'
        QtMocHelpers::SignalData<void(bool)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 5 },
        }}),
        // Signal 'volumeChanged'
        QtMocHelpers::SignalData<void(double)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Double, 7 },
        }}),
        // Signal 'loopStatusChanged'
        QtMocHelpers::SignalData<void(const QString &)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 9 },
        }}),
        // Signal 'shuffleChanged'
        QtMocHelpers::SignalData<void(bool)>(10, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 11 },
        }}),
        // Signal 'reloadMetadata'
        QtMocHelpers::SignalData<void()>(12, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'Identity'
        QtMocHelpers::PropertyData<QString>(13, QMetaType::QString, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'DesktopEntry'
        QtMocHelpers::PropertyData<QString>(14, QMetaType::QString, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'CanRaise'
        QtMocHelpers::PropertyData<bool>(15, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'CanQuit'
        QtMocHelpers::PropertyData<bool>(16, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'Fullscreen'
        QtMocHelpers::PropertyData<bool>(17, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 1),
        // property 'CanSetFullscreen'
        QtMocHelpers::PropertyData<bool>(18, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'SupportedUriSchemes'
        QtMocHelpers::PropertyData<QStringList>(19, QMetaType::QStringList, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'SupportedMimeTypes'
        QtMocHelpers::PropertyData<QStringList>(20, QMetaType::QStringList, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'CanControl'
        QtMocHelpers::PropertyData<bool>(21, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'CanGoNext'
        QtMocHelpers::PropertyData<bool>(22, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'CanGoPrevious'
        QtMocHelpers::PropertyData<bool>(23, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'CanPause'
        QtMocHelpers::PropertyData<bool>(24, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'CanPlay'
        QtMocHelpers::PropertyData<bool>(25, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'CanSeek'
        QtMocHelpers::PropertyData<bool>(26, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'Volume'
        QtMocHelpers::PropertyData<double>(27, QMetaType::Double, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 2),
        // property 'PlaybackStatus'
        QtMocHelpers::PropertyData<QString>(28, QMetaType::QString, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'LoopStatus'
        QtMocHelpers::PropertyData<QString>(29, QMetaType::QString, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 3),
        // property 'Shuffle'
        QtMocHelpers::PropertyData<bool>(30, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 4),
        // property 'Metadata'
        QtMocHelpers::PropertyData<QVariantMap>(31, 0x80000000 | 32, QMC::DefaultPropertyFlags | QMC::EnumOrFlag | QMC::Constant),
        // property 'Position'
        QtMocHelpers::PropertyData<int64_t>(33, 0x80000000 | 34, QMC::DefaultPropertyFlags | QMC::EnumOrFlag | QMC::Constant),
        // property 'MinimumRate'
        QtMocHelpers::PropertyData<double>(35, QMetaType::Double, QMC::DefaultPropertyFlags | QMC::Constant),
        // property 'MaximumRate'
        QtMocHelpers::PropertyData<double>(36, QMetaType::Double, QMC::DefaultPropertyFlags | QMC::Constant),
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<MprisPlugin, qt_meta_tag_ZN6Fooyin5Mpris11MprisPluginE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::Mpris::MprisPlugin::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin5Mpris11MprisPluginE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin5Mpris11MprisPluginE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin5Mpris11MprisPluginE_t>.metaTypes,
    nullptr
} };

void Fooyin::Mpris::MprisPlugin::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<MprisPlugin *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->Seeked((*reinterpret_cast<std::add_pointer_t<qlonglong>>(_a[1]))); break;
        case 1: _t->fullscreenChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 2: _t->volumeChanged((*reinterpret_cast<std::add_pointer_t<double>>(_a[1]))); break;
        case 3: _t->loopStatusChanged((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 4: _t->shuffleChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 5: _t->reloadMetadata(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (MprisPlugin::*)(qlonglong )>(_a, &MprisPlugin::Seeked, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (MprisPlugin::*)(bool )>(_a, &MprisPlugin::fullscreenChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (MprisPlugin::*)(double )>(_a, &MprisPlugin::volumeChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (MprisPlugin::*)(const QString & )>(_a, &MprisPlugin::loopStatusChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (MprisPlugin::*)(bool )>(_a, &MprisPlugin::shuffleChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (MprisPlugin::*)()>(_a, &MprisPlugin::reloadMetadata, 5))
            return;
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<QString*>(_v) = _t->identity(); break;
        case 1: *reinterpret_cast<QString*>(_v) = _t->desktopEntry(); break;
        case 2: *reinterpret_cast<bool*>(_v) = _t->canRaise(); break;
        case 3: *reinterpret_cast<bool*>(_v) = _t->canQuit(); break;
        case 4: *reinterpret_cast<bool*>(_v) = _t->fullscreen(); break;
        case 5: *reinterpret_cast<bool*>(_v) = _t->canSetFullscreen(); break;
        case 6: *reinterpret_cast<QStringList*>(_v) = _t->supportedUriSchemes(); break;
        case 7: *reinterpret_cast<QStringList*>(_v) = _t->supportedMimeTypes(); break;
        case 8: *reinterpret_cast<bool*>(_v) = _t->canControl(); break;
        case 9: *reinterpret_cast<bool*>(_v) = _t->canGoNext(); break;
        case 10: *reinterpret_cast<bool*>(_v) = _t->canGoPrevious(); break;
        case 11: *reinterpret_cast<bool*>(_v) = _t->canPause(); break;
        case 12: *reinterpret_cast<bool*>(_v) = _t->canPlay(); break;
        case 13: *reinterpret_cast<bool*>(_v) = _t->canSeek(); break;
        case 14: *reinterpret_cast<double*>(_v) = _t->volume(); break;
        case 15: *reinterpret_cast<QString*>(_v) = _t->playbackStatus(); break;
        case 16: *reinterpret_cast<QString*>(_v) = _t->loopStatus(); break;
        case 17: *reinterpret_cast<bool*>(_v) = _t->shuffle(); break;
        case 18: *reinterpret_cast<QVariantMap*>(_v) = _t->metadata(); break;
        case 19: *reinterpret_cast<int64_t*>(_v) = _t->position(); break;
        case 20: *reinterpret_cast<double*>(_v) = _t->minimumRate(); break;
        case 21: *reinterpret_cast<double*>(_v) = _t->maximumRate(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 4: _t->setFullscreen(*reinterpret_cast<bool*>(_v)); break;
        case 14: _t->setVolume(*reinterpret_cast<double*>(_v)); break;
        case 16: _t->setLoopStatus(*reinterpret_cast<QString*>(_v)); break;
        case 17: _t->setShuffle(*reinterpret_cast<bool*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *Fooyin::Mpris::MprisPlugin::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Fooyin::Mpris::MprisPlugin::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin5Mpris11MprisPluginE_t>.strings))
        return static_cast<void*>(this);
    if (!strcmp(_clname, "Plugin"))
        return static_cast< Plugin*>(this);
    if (!strcmp(_clname, "CorePlugin"))
        return static_cast< CorePlugin*>(this);
    if (!strcmp(_clname, "GuiPlugin"))
        return static_cast< GuiPlugin*>(this);
    if (!strcmp(_clname, "org.fooyin.fooyin.plugin/1.0"))
        return static_cast< Fooyin::Plugin*>(this);
    if (!strcmp(_clname, "org.fooyin.fooyin.plugin.core"))
        return static_cast< Fooyin::CorePlugin*>(this);
    if (!strcmp(_clname, "org.fooyin.fooyin.plugin.gui"))
        return static_cast< Fooyin::GuiPlugin*>(this);
    return QObject::qt_metacast(_clname);
}

int Fooyin::Mpris::MprisPlugin::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 6)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 6;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 6)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 6;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 22;
    }
    return _id;
}

// SIGNAL 0
void Fooyin::Mpris::MprisPlugin::Seeked(qlonglong _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void Fooyin::Mpris::MprisPlugin::fullscreenChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void Fooyin::Mpris::MprisPlugin::volumeChanged(double _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void Fooyin::Mpris::MprisPlugin::loopStatusChanged(const QString & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void Fooyin::Mpris::MprisPlugin::shuffleChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void Fooyin::Mpris::MprisPlugin::reloadMetadata()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}
using namespace Fooyin;
using namespace Fooyin::Mpris;

#ifdef QT_MOC_EXPORT_PLUGIN_V2
static constexpr unsigned char qt_pluginMetaDataV2_MprisPlugin[] = {
    0xbf, 
    // "IID"
    0x02,  0x78,  0x1c,  'o',  'r',  'g',  '.',  'f', 
    'o',  'o',  'y',  'i',  'n',  '.',  'f',  'o', 
    'o',  'y',  'i',  'n',  '.',  'p',  'l',  'u', 
    'g',  'i',  'n',  '/',  '1',  '.',  '0', 
    // "className"
    0x03,  0x6b,  'M',  'p',  'r',  'i',  's',  'P', 
    'l',  'u',  'g',  'i',  'n', 
    // "MetaData"
    0x04,  0xa8,  0x66,  'A',  'u',  't',  'h',  'o', 
    'r',  0x66,  'f',  'o',  'o',  'y',  'i',  'n', 
    0x68,  'C',  'a',  't',  'e',  'g',  'o',  'r', 
    'y',  0x72,  'S',  'y',  's',  't',  'e',  'm', 
    ' ',  'I',  'n',  't',  'e',  'g',  'r',  'a', 
    't',  'i',  'o',  'n',  0x69,  'C',  'o',  'p', 
    'y',  'r',  'i',  'g',  'h',  't',  0x78,  0x2c, 
    'C',  'o',  'p',  'y',  'r',  'i',  'g',  'h', 
    't',  ' ',  uchar('\xc2'), uchar('\xa9'), ' ',  '2',  '0',  '2', 
    '4',  ',',  ' ',  'L',  'u',  'k',  'e',  ' ', 
    'T',  'a',  'y',  'l',  'o',  'r',  ' ',  '<', 
    'l',  'u',  'k',  'e',  't',  '@',  'p',  'm', 
    '.',  'm',  'e',  '>',  0x6b,  'D',  'e',  's', 
    'c',  'r',  'i',  'p',  't',  'i',  'o',  'n', 
    0x72,  'A',  'd',  'd',  's',  ' ',  'M',  'P', 
    'R',  'I',  'S',  ' ',  's',  'u',  'p',  'p', 
    'o',  'r',  't',  0x67,  'L',  'i',  'c',  'e', 
    'n',  's',  'e',  0x85,  0x78,  0xe9,  'F',  'o', 
    'o',  'y',  'i',  'n',  ' ',  'i',  's',  ' ', 
    'f',  'r',  'e',  'e',  ' ',  's',  'o',  'f', 
    't',  'w',  'a',  'r',  'e',  ':',  ' ',  'y', 
    'o',  'u',  ' ',  'c',  'a',  'n',  ' ',  'r', 
    'e',  'd',  'i',  's',  't',  'r',  'i',  'b', 
    'u',  't',  'e',  ' ',  'i',  't',  ' ',  'a', 
    'n',  'd',  '/',  'o',  'r',  ' ',  'm',  'o', 
    'd',  'i',  'f',  'y',  ' ',  'i',  't',  ' ', 
    'u',  'n',  'd',  'e',  'r',  ' ',  't',  'h', 
    'e',  ' ',  't',  'e',  'r',  'm',  's',  ' ', 
    'o',  'f',  ' ',  't',  'h',  'e',  ' ',  'G', 
    'N',  'U',  ' ',  'G',  'e',  'n',  'e',  'r', 
    'a',  'l',  ' ',  'P',  'u',  'b',  'l',  'i', 
    'c',  ' ',  'L',  'i',  'c',  'e',  'n',  's', 
    'e',  ' ',  'a',  's',  ' ',  'p',  'u',  'b', 
    'l',  'i',  's',  'h',  'e',  'd',  ' ',  'b', 
    'y',  ' ',  't',  'h',  'e',  ' ',  'F',  'r', 
    'e',  'e',  ' ',  'S',  'o',  'f',  't',  'w', 
    'a',  'r',  'e',  ' ',  'F',  'o',  'u',  'n', 
    'd',  'a',  't',  'i',  'o',  'n',  ',',  ' ', 
    'e',  'i',  't',  'h',  'e',  'r',  ' ',  'v', 
    'e',  'r',  's',  'i',  'o',  'n',  ' ',  '3', 
    ' ',  'o',  'f',  ' ',  't',  'h',  'e',  ' ', 
    'L',  'i',  'c',  'e',  'n',  's',  'e',  ',', 
    ' ',  'o',  'r',  ' ',  '(',  'a',  't',  ' ', 
    'y',  'o',  'u',  'r',  ' ',  'o',  'p',  't', 
    'i',  'o',  'n',  ')',  ' ',  'a',  'n',  'y', 
    ' ',  'l',  'a',  't',  'e',  'r',  ' ',  'v', 
    'e',  'r',  's',  'i',  'o',  'n',  '.',  0x60, 
    0x78,  0xe3,  'F',  'o',  'o',  'y',  'i',  'n', 
    ' ',  'i',  's',  ' ',  'd',  'i',  's',  't', 
    'r',  'i',  'b',  'u',  't',  'e',  'd',  ' ', 
    'i',  'n',  ' ',  't',  'h',  'e',  ' ',  'h', 
    'o',  'p',  'e',  ' ',  't',  'h',  'a',  't', 
    ' ',  'i',  't',  ' ',  'w',  'i',  'l',  'l', 
    ' ',  'b',  'e',  ' ',  'u',  's',  'e',  'f', 
    'u',  'l',  ',',  ' ',  'b',  'u',  't',  ' ', 
    'W',  'I',  'T',  'H',  'O',  'U',  'T',  ' ', 
    'A',  'N',  'Y',  ' ',  'W',  'A',  'R',  'R', 
    'A',  'N',  'T',  'Y',  ';',  ' ',  'w',  'i', 
    't',  'h',  'o',  'u',  't',  ' ',  'e',  'v', 
    'e',  'n',  ' ',  't',  'h',  'e',  ' ',  'i', 
    'm',  'p',  'l',  'i',  'e',  'd',  ' ',  'w', 
    'a',  'r',  'r',  'a',  'n',  't',  'y',  ' ', 
    'o',  'f',  ' ',  'M',  'E',  'R',  'C',  'H', 
    'A',  'N',  'T',  'A',  'B',  'I',  'L',  'I', 
    'T',  'Y',  ' ',  'o',  'r',  ' ',  'F',  'I', 
    'T',  'N',  'E',  'S',  'S',  ' ',  'F',  'O', 
    'R',  ' ',  'A',  ' ',  'P',  'A',  'R',  'T', 
    'I',  'C',  'U',  'L',  'A',  'R',  ' ',  'P', 
    'U',  'R',  'P',  'O',  'S',  'E',  '.',  ' ', 
    ' ',  'S',  'e',  'e',  ' ',  't',  'h',  'e', 
    ' ',  'G',  'N',  'U',  ' ',  'G',  'e',  'n', 
    'e',  'r',  'a',  'l',  ' ',  'P',  'u',  'b', 
    'l',  'i',  'c',  ' ',  'L',  'i',  'c',  'e', 
    'n',  's',  'e',  ' ',  'f',  'o',  'r',  ' ', 
    'm',  'o',  'r',  'e',  ' ',  'd',  'e',  't', 
    'a',  'i',  'l',  's',  '.',  0x60,  0x78,  0x80, 
    'Y',  'o',  'u',  ' ',  's',  'h',  'o',  'u', 
    'l',  'd',  ' ',  'h',  'a',  'v',  'e',  ' ', 
    'r',  'e',  'c',  'e',  'i',  'v',  'e',  'd', 
    ' ',  'a',  ' ',  'c',  'o',  'p',  'y',  ' ', 
    'o',  'f',  ' ',  't',  'h',  'e',  ' ',  'G', 
    'N',  'U',  ' ',  'G',  'e',  'n',  'e',  'r', 
    'a',  'l',  ' ',  'P',  'u',  'b',  'l',  'i', 
    'c',  ' ',  'L',  'i',  'c',  'e',  'n',  's', 
    'e',  ' ',  'a',  'l',  'o',  'n',  'g',  ' ', 
    'w',  'i',  't',  'h',  ' ',  'F',  'o',  'o', 
    'y',  'i',  'n',  '.',  ' ',  ' ',  'I',  'f', 
    ' ',  'n',  'o',  't',  ',',  ' ',  's',  'e', 
    'e',  ' ',  '<',  'h',  't',  't',  'p',  ':', 
    '/',  '/',  'w',  'w',  'w',  '.',  'g',  'n', 
    'u',  '.',  'o',  'r',  'g',  '/',  'l',  'i', 
    'c',  'e',  'n',  's',  'e',  's',  '/',  '>', 
    0x64,  'N',  'a',  'm',  'e',  0x65,  'M',  'P', 
    'R',  'I',  'S',  0x63,  'U',  'r',  'l',  0x78, 
    0x20,  'h',  't',  't',  'p',  's',  ':',  '/', 
    '/',  'g',  'i',  't',  'h',  'u',  'b',  '.', 
    'c',  'o',  'm',  '/',  'f',  'o',  'o',  'y', 
    'i',  'n',  '/',  'f',  'o',  'o',  'y',  'i', 
    'n',  0x67,  'V',  'e',  'r',  's',  'i',  'o', 
    'n',  0x66,  '0',  '.',  '1',  '2',  '.',  '0', 
    0xff, 
};
QT_MOC_EXPORT_PLUGIN_V2(Fooyin::Mpris::MprisPlugin, MprisPlugin, qt_pluginMetaDataV2_MprisPlugin)
#else
QT_PLUGIN_METADATA_SECTION
Q_CONSTINIT static constexpr unsigned char qt_pluginMetaData_MprisPlugin[] = {
    'Q', 'T', 'M', 'E', 'T', 'A', 'D', 'A', 'T', 'A', ' ', '!',
    // metadata version, Qt version, architectural requirements
    0, QT_VERSION_MAJOR, QT_VERSION_MINOR, qPluginArchRequirements(),
    0xbf, 
    // "IID"
    0x02,  0x78,  0x1c,  'o',  'r',  'g',  '.',  'f', 
    'o',  'o',  'y',  'i',  'n',  '.',  'f',  'o', 
    'o',  'y',  'i',  'n',  '.',  'p',  'l',  'u', 
    'g',  'i',  'n',  '/',  '1',  '.',  '0', 
    // "className"
    0x03,  0x6b,  'M',  'p',  'r',  'i',  's',  'P', 
    'l',  'u',  'g',  'i',  'n', 
    // "MetaData"
    0x04,  0xa8,  0x66,  'A',  'u',  't',  'h',  'o', 
    'r',  0x66,  'f',  'o',  'o',  'y',  'i',  'n', 
    0x68,  'C',  'a',  't',  'e',  'g',  'o',  'r', 
    'y',  0x72,  'S',  'y',  's',  't',  'e',  'm', 
    ' ',  'I',  'n',  't',  'e',  'g',  'r',  'a', 
    't',  'i',  'o',  'n',  0x69,  'C',  'o',  'p', 
    'y',  'r',  'i',  'g',  'h',  't',  0x78,  0x2c, 
    'C',  'o',  'p',  'y',  'r',  'i',  'g',  'h', 
    't',  ' ',  uchar('\xc2'), uchar('\xa9'), ' ',  '2',  '0',  '2', 
    '4',  ',',  ' ',  'L',  'u',  'k',  'e',  ' ', 
    'T',  'a',  'y',  'l',  'o',  'r',  ' ',  '<', 
    'l',  'u',  'k',  'e',  't',  '@',  'p',  'm', 
    '.',  'm',  'e',  '>',  0x6b,  'D',  'e',  's', 
    'c',  'r',  'i',  'p',  't',  'i',  'o',  'n', 
    0x72,  'A',  'd',  'd',  's',  ' ',  'M',  'P', 
    'R',  'I',  'S',  ' ',  's',  'u',  'p',  'p', 
    'o',  'r',  't',  0x67,  'L',  'i',  'c',  'e', 
    'n',  's',  'e',  0x85,  0x78,  0xe9,  'F',  'o', 
    'o',  'y',  'i',  'n',  ' ',  'i',  's',  ' ', 
    'f',  'r',  'e',  'e',  ' ',  's',  'o',  'f', 
    't',  'w',  'a',  'r',  'e',  ':',  ' ',  'y', 
    'o',  'u',  ' ',  'c',  'a',  'n',  ' ',  'r', 
    'e',  'd',  'i',  's',  't',  'r',  'i',  'b', 
    'u',  't',  'e',  ' ',  'i',  't',  ' ',  'a', 
    'n',  'd',  '/',  'o',  'r',  ' ',  'm',  'o', 
    'd',  'i',  'f',  'y',  ' ',  'i',  't',  ' ', 
    'u',  'n',  'd',  'e',  'r',  ' ',  't',  'h', 
    'e',  ' ',  't',  'e',  'r',  'm',  's',  ' ', 
    'o',  'f',  ' ',  't',  'h',  'e',  ' ',  'G', 
    'N',  'U',  ' ',  'G',  'e',  'n',  'e',  'r', 
    'a',  'l',  ' ',  'P',  'u',  'b',  'l',  'i', 
    'c',  ' ',  'L',  'i',  'c',  'e',  'n',  's', 
    'e',  ' ',  'a',  's',  ' ',  'p',  'u',  'b', 
    'l',  'i',  's',  'h',  'e',  'd',  ' ',  'b', 
    'y',  ' ',  't',  'h',  'e',  ' ',  'F',  'r', 
    'e',  'e',  ' ',  'S',  'o',  'f',  't',  'w', 
    'a',  'r',  'e',  ' ',  'F',  'o',  'u',  'n', 
    'd',  'a',  't',  'i',  'o',  'n',  ',',  ' ', 
    'e',  'i',  't',  'h',  'e',  'r',  ' ',  'v', 
    'e',  'r',  's',  'i',  'o',  'n',  ' ',  '3', 
    ' ',  'o',  'f',  ' ',  't',  'h',  'e',  ' ', 
    'L',  'i',  'c',  'e',  'n',  's',  'e',  ',', 
    ' ',  'o',  'r',  ' ',  '(',  'a',  't',  ' ', 
    'y',  'o',  'u',  'r',  ' ',  'o',  'p',  't', 
    'i',  'o',  'n',  ')',  ' ',  'a',  'n',  'y', 
    ' ',  'l',  'a',  't',  'e',  'r',  ' ',  'v', 
    'e',  'r',  's',  'i',  'o',  'n',  '.',  0x60, 
    0x78,  0xe3,  'F',  'o',  'o',  'y',  'i',  'n', 
    ' ',  'i',  's',  ' ',  'd',  'i',  's',  't', 
    'r',  'i',  'b',  'u',  't',  'e',  'd',  ' ', 
    'i',  'n',  ' ',  't',  'h',  'e',  ' ',  'h', 
    'o',  'p',  'e',  ' ',  't',  'h',  'a',  't', 
    ' ',  'i',  't',  ' ',  'w',  'i',  'l',  'l', 
    ' ',  'b',  'e',  ' ',  'u',  's',  'e',  'f', 
    'u',  'l',  ',',  ' ',  'b',  'u',  't',  ' ', 
    'W',  'I',  'T',  'H',  'O',  'U',  'T',  ' ', 
    'A',  'N',  'Y',  ' ',  'W',  'A',  'R',  'R', 
    'A',  'N',  'T',  'Y',  ';',  ' ',  'w',  'i', 
    't',  'h',  'o',  'u',  't',  ' ',  'e',  'v', 
    'e',  'n',  ' ',  't',  'h',  'e',  ' ',  'i', 
    'm',  'p',  'l',  'i',  'e',  'd',  ' ',  'w', 
    'a',  'r',  'r',  'a',  'n',  't',  'y',  ' ', 
    'o',  'f',  ' ',  'M',  'E',  'R',  'C',  'H', 
    'A',  'N',  'T',  'A',  'B',  'I',  'L',  'I', 
    'T',  'Y',  ' ',  'o',  'r',  ' ',  'F',  'I', 
    'T',  'N',  'E',  'S',  'S',  ' ',  'F',  'O', 
    'R',  ' ',  'A',  ' ',  'P',  'A',  'R',  'T', 
    'I',  'C',  'U',  'L',  'A',  'R',  ' ',  'P', 
    'U',  'R',  'P',  'O',  'S',  'E',  '.',  ' ', 
    ' ',  'S',  'e',  'e',  ' ',  't',  'h',  'e', 
    ' ',  'G',  'N',  'U',  ' ',  'G',  'e',  'n', 
    'e',  'r',  'a',  'l',  ' ',  'P',  'u',  'b', 
    'l',  'i',  'c',  ' ',  'L',  'i',  'c',  'e', 
    'n',  's',  'e',  ' ',  'f',  'o',  'r',  ' ', 
    'm',  'o',  'r',  'e',  ' ',  'd',  'e',  't', 
    'a',  'i',  'l',  's',  '.',  0x60,  0x78,  0x80, 
    'Y',  'o',  'u',  ' ',  's',  'h',  'o',  'u', 
    'l',  'd',  ' ',  'h',  'a',  'v',  'e',  ' ', 
    'r',  'e',  'c',  'e',  'i',  'v',  'e',  'd', 
    ' ',  'a',  ' ',  'c',  'o',  'p',  'y',  ' ', 
    'o',  'f',  ' ',  't',  'h',  'e',  ' ',  'G', 
    'N',  'U',  ' ',  'G',  'e',  'n',  'e',  'r', 
    'a',  'l',  ' ',  'P',  'u',  'b',  'l',  'i', 
    'c',  ' ',  'L',  'i',  'c',  'e',  'n',  's', 
    'e',  ' ',  'a',  'l',  'o',  'n',  'g',  ' ', 
    'w',  'i',  't',  'h',  ' ',  'F',  'o',  'o', 
    'y',  'i',  'n',  '.',  ' ',  ' ',  'I',  'f', 
    ' ',  'n',  'o',  't',  ',',  ' ',  's',  'e', 
    'e',  ' ',  '<',  'h',  't',  't',  'p',  ':', 
    '/',  '/',  'w',  'w',  'w',  '.',  'g',  'n', 
    'u',  '.',  'o',  'r',  'g',  '/',  'l',  'i', 
    'c',  'e',  'n',  's',  'e',  's',  '/',  '>', 
    0x64,  'N',  'a',  'm',  'e',  0x65,  'M',  'P', 
    'R',  'I',  'S',  0x63,  'U',  'r',  'l',  0x78, 
    0x20,  'h',  't',  't',  'p',  's',  ':',  '/', 
    '/',  'g',  'i',  't',  'h',  'u',  'b',  '.', 
    'c',  'o',  'm',  '/',  'f',  'o',  'o',  'y', 
    'i',  'n',  '/',  'f',  'o',  'o',  'y',  'i', 
    'n',  0x67,  'V',  'e',  'r',  's',  'i',  'o', 
    'n',  0x66,  '0',  '.',  '1',  '2',  '.',  '0', 
    0xff, 
};
QT_MOC_EXPORT_PLUGIN(Fooyin::Mpris::MprisPlugin, MprisPlugin)
#endif  // QT_MOC_EXPORT_PLUGIN_V2

QT_WARNING_POP
