/****************************************************************************
** Meta object code from reading C++ file 'playlisttabs.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../fooyin/src/gui/playlist/playlisttabs.h"
#include <QtCore/qmetatype.h>
#include <QtCore/QList>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'playlisttabs.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin12PlaylistTabsE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::PlaylistTabs::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin12PlaylistTabsE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::PlaylistTabs",
        "filesDropped",
        "",
        "QList<QUrl>",
        "urls",
        "Fooyin::UId",
        "playlistId",
        "tracksDropped",
        "data",
        "trackListDropped",
        "Fooyin::TrackList",
        "tracks",
        "savePlaylistRequested"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'filesDropped'
        QtMocHelpers::SignalData<void(const QList<QUrl> &, const Fooyin::UId &)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 }, { 0x80000000 | 5, 6 },
        }}),
        // Signal 'tracksDropped'
        QtMocHelpers::SignalData<void(const QByteArray &, const Fooyin::UId &)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QByteArray, 8 }, { 0x80000000 | 5, 6 },
        }}),
        // Signal 'trackListDropped'
        QtMocHelpers::SignalData<void(const Fooyin::TrackList &, const Fooyin::UId &)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 10, 11 }, { 0x80000000 | 5, 6 },
        }}),
        // Signal 'savePlaylistRequested'
        QtMocHelpers::SignalData<void(const Fooyin::UId &)>(12, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 5, 6 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<PlaylistTabs, qt_meta_tag_ZN6Fooyin12PlaylistTabsE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::PlaylistTabs::staticMetaObject = { {
    QMetaObject::SuperData::link<WidgetContainer::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin12PlaylistTabsE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin12PlaylistTabsE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin12PlaylistTabsE_t>.metaTypes,
    nullptr
} };

void Fooyin::PlaylistTabs::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<PlaylistTabs *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->filesDropped((*reinterpret_cast<std::add_pointer_t<QList<QUrl>>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::UId>>(_a[2]))); break;
        case 1: _t->tracksDropped((*reinterpret_cast<std::add_pointer_t<QByteArray>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::UId>>(_a[2]))); break;
        case 2: _t->trackListDropped((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::UId>>(_a[2]))); break;
        case 3: _t->savePlaylistRequested((*reinterpret_cast<std::add_pointer_t<Fooyin::UId>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 2:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (PlaylistTabs::*)(const QList<QUrl> & , const Fooyin::UId & )>(_a, &PlaylistTabs::filesDropped, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistTabs::*)(const QByteArray & , const Fooyin::UId & )>(_a, &PlaylistTabs::tracksDropped, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistTabs::*)(const Fooyin::TrackList & , const Fooyin::UId & )>(_a, &PlaylistTabs::trackListDropped, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (PlaylistTabs::*)(const Fooyin::UId & )>(_a, &PlaylistTabs::savePlaylistRequested, 3))
            return;
    }
}

const QMetaObject *Fooyin::PlaylistTabs::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Fooyin::PlaylistTabs::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin12PlaylistTabsE_t>.strings))
        return static_cast<void*>(this);
    return WidgetContainer::qt_metacast(_clname);
}

int Fooyin::PlaylistTabs::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = WidgetContainer::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 4)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 4;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 4)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 4;
    }
    return _id;
}

// SIGNAL 0
void Fooyin::PlaylistTabs::filesDropped(const QList<QUrl> & _t1, const Fooyin::UId & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1, _t2);
}

// SIGNAL 1
void Fooyin::PlaylistTabs::tracksDropped(const QByteArray & _t1, const Fooyin::UId & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1, _t2);
}

// SIGNAL 2
void Fooyin::PlaylistTabs::trackListDropped(const Fooyin::TrackList & _t1, const Fooyin::UId & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1, _t2);
}

// SIGNAL 3
void Fooyin::PlaylistTabs::savePlaylistRequested(const Fooyin::UId & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}
QT_WARNING_POP
