/*
 * Fooyin
 * Copyright © 2024, Luke Taylor <luket@pm.me>
 *
 * Fooyin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fooyin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fooyin.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "dirdelegate.h"

#include "dirproxymodel.h"

#include <gui/iconloader.h>

#include <QApplication>
#include <QPainter>

namespace Fooyin {
void DirDelegate::paint(QPainter* painter, const QStyleOptionViewItem& option, const QModelIndex& index) const
{
    QStyleOptionViewItem opt = option;
    initStyleOption(&opt, index);

    QStyle* style           = option.widget ? option.widget->style() : QApplication::style();
    const bool isPlaying    = index.data(DirProxyModel::IsPlaying).toBool();
    const QIcon playingIcon = isPlaying ? opt.icon : QIcon{};
    const QRect iconRect
        = isPlaying ? style->subElementRect(QStyle::SE_ItemViewItemDecoration, &opt, option.widget) : QRect{};

    painter->save();

    if(opt.backgroundBrush.style() != Qt::NoBrush) {
        painter->fillRect(option.rect, opt.backgroundBrush);
        opt.backgroundBrush = Qt::NoBrush;
    }

    if(isPlaying) {
        opt.icon = {};
    }
    style->drawControl(QStyle::CE_ItemViewItem, &opt, painter, option.widget);

    if(isPlaying) {
        Gui::drawItemViewIcon(painter, opt, playingIcon, iconRect);
    }

    painter->restore();
}
} // namespace Fooyin
