/*
 * Fooyin
 * Copyright © 2024, Luke Taylor <luket@pm.me>
 *
 * Fooyin is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fooyin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fooyin.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#pragma once

#include <QObject>

#include <vector>

class QAction;
class QActionGroup;

namespace Fooyin {
class ActionContainer;
class ActionManager;
class Command;
class FyLayout;
class LayoutProvider;
class SettingsManager;

class LayoutMenu : public QObject
{
    Q_OBJECT

public:
    LayoutMenu(ActionManager* actionManager, LayoutProvider* layoutProvider, SettingsManager* settings,
               QObject* parent = nullptr);

    void setup();

Q_SIGNALS:
    void changeLayout(const Fooyin::FyLayout& layout);
    void newLayout();
    void clearLayout();
    void importLayout();
    void exportLayout();

private:
    void refreshLayouts();
    void updateCurrentLayout();
    void resetCurrentLayout();

    ActionManager* m_actionManager;
    LayoutProvider* m_layoutProvider;
    SettingsManager* m_settings;

    ActionContainer* m_layoutMenu;
    QAction* m_layoutEditing;
    Command* m_layoutEditingCmd;
    QAction* m_lockSplitters;
    Command* m_lockSplittersCmd;
    QAction* m_newLayout;
    QAction* m_clearLayout;
    QAction* m_resetLayout;
    QActionGroup* m_layoutActionGroup;
    std::vector<QAction*> m_layoutActions;
};
} // namespace Fooyin
