/****************************************************************************
** Meta object code from reading C++ file 'enginedefs.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../fooyin/include/core/engine/enginedefs.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'enginedefs.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin6EngineE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::Engine::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin6EngineE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::Engine",
        "RGProcessing",
        "RGProcess",
        "NoProcessing",
        "ApplyGain",
        "PreventClipping",
        "AnalysisDataTypes",
        "AnalysisDataType",
        "NoAnalysisData",
        "LevelFrameData",
        "PcmFrameData"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // flag 'RGProcessing'
        QtMocHelpers::EnumData<RGProcessing>(1, 2, QMC::EnumIsFlag).add({
            {    3, RGProcess::NoProcessing },
            {    4, RGProcess::ApplyGain },
            {    5, RGProcess::PreventClipping },
        }),
        // flag 'AnalysisDataTypes'
        QtMocHelpers::EnumData<AnalysisDataTypes>(6, 7, QMC::EnumIsFlag).add({
            {    8, AnalysisDataType::NoAnalysisData },
            {    9, AnalysisDataType::LevelFrameData },
            {   10, AnalysisDataType::PcmFrameData },
        }),
    };
    return QtMocHelpers::metaObjectData<void, qt_meta_tag_ZN6Fooyin6EngineE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}

static constexpr auto qt_staticMetaObjectContent_ZN6Fooyin6EngineE =
    Fooyin::Engine::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin6EngineE_t>();
static constexpr auto qt_staticMetaObjectStaticContent_ZN6Fooyin6EngineE =
    qt_staticMetaObjectContent_ZN6Fooyin6EngineE.staticData;
static constexpr auto qt_staticMetaObjectRelocatingContent_ZN6Fooyin6EngineE =
    qt_staticMetaObjectContent_ZN6Fooyin6EngineE.relocatingData;

Q_CONSTINIT const QMetaObject Fooyin::Engine::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent_ZN6Fooyin6EngineE.stringdata,
    qt_staticMetaObjectStaticContent_ZN6Fooyin6EngineE.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent_ZN6Fooyin6EngineE.metaTypes,
    nullptr
} };

QT_WARNING_POP
