/****************************************************************************
** Meta object code from reading C++ file 'mprisplayer.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "mprisplayer.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'mprisplayer.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN11MprisPlayerE_t {};
} // unnamed namespace

template <> constexpr inline auto MprisPlayer::qt_create_metaobjectdata<qt_meta_tag_ZN11MprisPlayerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "MprisPlayer",
        "D-Bus Interface",
        "org.mpris.MediaPlayer2.Player",
        "D-Bus Introspection",
        "  <interface name=\"org.mpris.MediaPlayer2.Player\">\n    <method "
        "name=\"Next\"/>\n    <method name=\"Previous\"/>\n    <method name"
        "=\"Pause\"/>\n    <method name=\"PlayPause\"/>\n    <method name=\""
        "Stop\"/>\n    <method name=\"Play\"/>\n    <method name=\"Seek\">\n"
        "      <arg direction=\"in\" type=\"x\" name=\"Offset\"/>\n    </me"
        "thod>\n    <method name=\"SetPosition\">\n      <arg direction=\"i"
        "n\" type=\"o\" name=\"TrackId\"/>\n      <arg direction=\"in\" typ"
        "e=\"x\" name=\"Position\"/>\n    </method>\n    <signal name=\"See"
        "ked\">\n      <arg direction=\"out\" type=\"x\" name=\"Position\"/"
        ">\n    </signal>\n    <property access=\"read\" type=\"b\" name=\""
        "CanGoNext\"/>\n    <property access=\"read\" type=\"b\" name=\"Can"
        "GoPrevious\"/>\n    <property access=\"read\" type=\"b\" name=\"Ca"
        "nPlay\"/>\n    <property access=\"read\" type=\"b\" name=\"CanPaus"
        "e\"/>\n    <property access=\"read\" type=\"b\" name=\"CanSeek\"/>"
        "\n    <property access=\"read\" type=\"b\" name=\"CanControl\"/>\n"
        "    <property access=\"read\" type=\"d\" name=\"MinimumRate\"/>\n "
        "   <property access=\"read\" type=\"d\" name=\"MaximumRate\"/>\n  "
        "  <property access=\"read\" type=\"s\" name=\"PlaybackStatus\"/>\n"
        "    <property access=\"readwrite\" type=\"s\" name=\"LoopStatus\"/"
        ">\n    <property access=\"readwrite\" type=\"b\" name=\"Shuffle\"/"
        ">\n    <property access=\"read\" type=\"a{sv}\" name=\"Metadata\">"
        "\n      <annotation value=\"QVariantMap\" name=\"org.qtproject.QtD"
        "Bus.QtTypeName\"/>\n    </property>\n    <property access=\"readwr"
        "ite\" type=\"d\" name=\"Volume\"/>\n    <property access=\"read\" "
        "type=\"x\" name=\"Position\"/>\n  </interface>\n",
        "Seeked",
        "",
        "Position",
        "Next",
        "Pause",
        "Play",
        "PlayPause",
        "Previous",
        "Seek",
        "Offset",
        "SetPosition",
        "QDBusObjectPath",
        "TrackId",
        "Stop",
        "CanControl",
        "CanGoNext",
        "CanGoPrevious",
        "CanPause",
        "CanPlay",
        "CanSeek",
        "LoopStatus",
        "MaximumRate",
        "Metadata",
        "QVariantMap",
        "MinimumRate",
        "PlaybackStatus",
        "Shuffle",
        "Volume"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'Seeked'
        QtMocHelpers::SignalData<void(qlonglong)>(5, 6, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::LongLong, 7 },
        }}),
        // Slot 'Next'
        QtMocHelpers::SlotData<void()>(8, 6, QMC::AccessPublic, QMetaType::Void),
        // Slot 'Pause'
        QtMocHelpers::SlotData<void()>(9, 6, QMC::AccessPublic, QMetaType::Void),
        // Slot 'Play'
        QtMocHelpers::SlotData<void()>(10, 6, QMC::AccessPublic, QMetaType::Void),
        // Slot 'PlayPause'
        QtMocHelpers::SlotData<void()>(11, 6, QMC::AccessPublic, QMetaType::Void),
        // Slot 'Previous'
        QtMocHelpers::SlotData<void()>(12, 6, QMC::AccessPublic, QMetaType::Void),
        // Slot 'Seek'
        QtMocHelpers::SlotData<void(qlonglong)>(13, 6, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::LongLong, 14 },
        }}),
        // Slot 'SetPosition'
        QtMocHelpers::SlotData<void(const QDBusObjectPath &, qlonglong)>(15, 6, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 16, 17 }, { QMetaType::LongLong, 7 },
        }}),
        // Slot 'Stop'
        QtMocHelpers::SlotData<void()>(18, 6, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'CanControl'
        QtMocHelpers::PropertyData<bool>(19, QMetaType::Bool, QMC::DefaultPropertyFlags),
        // property 'CanGoNext'
        QtMocHelpers::PropertyData<bool>(20, QMetaType::Bool, QMC::DefaultPropertyFlags),
        // property 'CanGoPrevious'
        QtMocHelpers::PropertyData<bool>(21, QMetaType::Bool, QMC::DefaultPropertyFlags),
        // property 'CanPause'
        QtMocHelpers::PropertyData<bool>(22, QMetaType::Bool, QMC::DefaultPropertyFlags),
        // property 'CanPlay'
        QtMocHelpers::PropertyData<bool>(23, QMetaType::Bool, QMC::DefaultPropertyFlags),
        // property 'CanSeek'
        QtMocHelpers::PropertyData<bool>(24, QMetaType::Bool, QMC::DefaultPropertyFlags),
        // property 'LoopStatus'
        QtMocHelpers::PropertyData<QString>(25, QMetaType::QString, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet),
        // property 'MaximumRate'
        QtMocHelpers::PropertyData<double>(26, QMetaType::Double, QMC::DefaultPropertyFlags),
        // property 'Metadata'
        QtMocHelpers::PropertyData<QVariantMap>(27, 0x80000000 | 28, QMC::DefaultPropertyFlags | QMC::EnumOrFlag),
        // property 'MinimumRate'
        QtMocHelpers::PropertyData<double>(29, QMetaType::Double, QMC::DefaultPropertyFlags),
        // property 'PlaybackStatus'
        QtMocHelpers::PropertyData<QString>(30, QMetaType::QString, QMC::DefaultPropertyFlags),
        // property 'Position'
        QtMocHelpers::PropertyData<qlonglong>(7, QMetaType::LongLong, QMC::DefaultPropertyFlags),
        // property 'Shuffle'
        QtMocHelpers::PropertyData<bool>(31, QMetaType::Bool, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet),
        // property 'Volume'
        QtMocHelpers::PropertyData<double>(32, QMetaType::Double, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet),
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
            {    3,    4 },
    });
    return QtMocHelpers::metaObjectData<MprisPlayer, qt_meta_tag_ZN11MprisPlayerE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject MprisPlayer::staticMetaObject = { {
    QMetaObject::SuperData::link<QDBusAbstractAdaptor::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN11MprisPlayerE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN11MprisPlayerE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN11MprisPlayerE_t>.metaTypes,
    nullptr
} };

void MprisPlayer::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<MprisPlayer *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->Seeked((*reinterpret_cast<std::add_pointer_t<qlonglong>>(_a[1]))); break;
        case 1: _t->Next(); break;
        case 2: _t->Pause(); break;
        case 3: _t->Play(); break;
        case 4: _t->PlayPause(); break;
        case 5: _t->Previous(); break;
        case 6: _t->Seek((*reinterpret_cast<std::add_pointer_t<qlonglong>>(_a[1]))); break;
        case 7: _t->SetPosition((*reinterpret_cast<std::add_pointer_t<QDBusObjectPath>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<qlonglong>>(_a[2]))); break;
        case 8: _t->Stop(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (MprisPlayer::*)(qlonglong )>(_a, &MprisPlayer::Seeked, 0))
            return;
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<bool*>(_v) = _t->canControl(); break;
        case 1: *reinterpret_cast<bool*>(_v) = _t->canGoNext(); break;
        case 2: *reinterpret_cast<bool*>(_v) = _t->canGoPrevious(); break;
        case 3: *reinterpret_cast<bool*>(_v) = _t->canPause(); break;
        case 4: *reinterpret_cast<bool*>(_v) = _t->canPlay(); break;
        case 5: *reinterpret_cast<bool*>(_v) = _t->canSeek(); break;
        case 6: *reinterpret_cast<QString*>(_v) = _t->loopStatus(); break;
        case 7: *reinterpret_cast<double*>(_v) = _t->maximumRate(); break;
        case 8: *reinterpret_cast<QVariantMap*>(_v) = _t->metadata(); break;
        case 9: *reinterpret_cast<double*>(_v) = _t->minimumRate(); break;
        case 10: *reinterpret_cast<QString*>(_v) = _t->playbackStatus(); break;
        case 11: *reinterpret_cast<qlonglong*>(_v) = _t->position(); break;
        case 12: *reinterpret_cast<bool*>(_v) = _t->shuffle(); break;
        case 13: *reinterpret_cast<double*>(_v) = _t->volume(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 6: _t->setLoopStatus(*reinterpret_cast<QString*>(_v)); break;
        case 12: _t->setShuffle(*reinterpret_cast<bool*>(_v)); break;
        case 13: _t->setVolume(*reinterpret_cast<double*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *MprisPlayer::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *MprisPlayer::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN11MprisPlayerE_t>.strings))
        return static_cast<void*>(this);
    return QDBusAbstractAdaptor::qt_metacast(_clname);
}

int MprisPlayer::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QDBusAbstractAdaptor::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 9)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 9;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 9)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 9;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 14;
    }
    return _id;
}

// SIGNAL 0
void MprisPlayer::Seeked(qlonglong _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}
QT_WARNING_POP
