/****************************************************************************
** Meta object code from reading C++ file 'filterwidget.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../../fooyin/src/plugins/filters/filterwidget.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'filterwidget.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin7Filters12FilterWidgetE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::Filters::FilterWidget::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin7Filters12FilterWidgetE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::Filters::FilterWidget",
        "doubleClicked",
        "",
        "middleClicked",
        "filterDeleted",
        "filterUpdated",
        "selectionKeysChanged",
        "std::vector<Fooyin::Filters::RowKey>",
        "keys",
        "configChanged",
        "requestHeaderMenu",
        "Fooyin::AutoHeaderView*",
        "header",
        "QPoint",
        "pos",
        "requestContextMenu",
        "requestEditConnections",
        "searchTextChanged",
        "search"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'doubleClicked'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'middleClicked'
        QtMocHelpers::SignalData<void()>(3, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'filterDeleted'
        QtMocHelpers::SignalData<void()>(4, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'filterUpdated'
        QtMocHelpers::SignalData<void()>(5, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'selectionKeysChanged'
        QtMocHelpers::SignalData<void(const std::vector<Fooyin::Filters::RowKey> &)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 7, 8 },
        }}),
        // Signal 'configChanged'
        QtMocHelpers::SignalData<void()>(9, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'requestHeaderMenu'
        QtMocHelpers::SignalData<void(Fooyin::AutoHeaderView *, const QPoint &)>(10, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 11, 12 }, { 0x80000000 | 13, 14 },
        }}),
        // Signal 'requestContextMenu'
        QtMocHelpers::SignalData<void(const QPoint &)>(15, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 13, 14 },
        }}),
        // Signal 'requestEditConnections'
        QtMocHelpers::SignalData<void()>(16, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'searchTextChanged'
        QtMocHelpers::SignalData<void(const QString &)>(17, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 18 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<FilterWidget, qt_meta_tag_ZN6Fooyin7Filters12FilterWidgetE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::Filters::FilterWidget::staticMetaObject = { {
    QMetaObject::SuperData::link<FyWidget::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin7Filters12FilterWidgetE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin7Filters12FilterWidgetE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin7Filters12FilterWidgetE_t>.metaTypes,
    nullptr
} };

void Fooyin::Filters::FilterWidget::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<FilterWidget *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->doubleClicked(); break;
        case 1: _t->middleClicked(); break;
        case 2: _t->filterDeleted(); break;
        case 3: _t->filterUpdated(); break;
        case 4: _t->selectionKeysChanged((*reinterpret_cast<std::add_pointer_t<std::vector<Fooyin::Filters::RowKey>>>(_a[1]))); break;
        case 5: _t->configChanged(); break;
        case 6: _t->requestHeaderMenu((*reinterpret_cast<std::add_pointer_t<Fooyin::AutoHeaderView*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[2]))); break;
        case 7: _t->requestContextMenu((*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[1]))); break;
        case 8: _t->requestEditConnections(); break;
        case 9: _t->searchTextChanged((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (FilterWidget::*)()>(_a, &FilterWidget::doubleClicked, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (FilterWidget::*)()>(_a, &FilterWidget::middleClicked, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (FilterWidget::*)()>(_a, &FilterWidget::filterDeleted, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (FilterWidget::*)()>(_a, &FilterWidget::filterUpdated, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (FilterWidget::*)(const std::vector<Fooyin::Filters::RowKey> & )>(_a, &FilterWidget::selectionKeysChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (FilterWidget::*)()>(_a, &FilterWidget::configChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (FilterWidget::*)(Fooyin::AutoHeaderView * , const QPoint & )>(_a, &FilterWidget::requestHeaderMenu, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (FilterWidget::*)(const QPoint & )>(_a, &FilterWidget::requestContextMenu, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (FilterWidget::*)()>(_a, &FilterWidget::requestEditConnections, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (FilterWidget::*)(const QString & )>(_a, &FilterWidget::searchTextChanged, 9))
            return;
    }
}

const QMetaObject *Fooyin::Filters::FilterWidget::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Fooyin::Filters::FilterWidget::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin7Filters12FilterWidgetE_t>.strings))
        return static_cast<void*>(this);
    return FyWidget::qt_metacast(_clname);
}

int Fooyin::Filters::FilterWidget::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = FyWidget::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 10)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 10;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 10)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 10;
    }
    return _id;
}

// SIGNAL 0
void Fooyin::Filters::FilterWidget::doubleClicked()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void Fooyin::Filters::FilterWidget::middleClicked()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void Fooyin::Filters::FilterWidget::filterDeleted()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void Fooyin::Filters::FilterWidget::filterUpdated()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void Fooyin::Filters::FilterWidget::selectionKeysChanged(const std::vector<Fooyin::Filters::RowKey> & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void Fooyin::Filters::FilterWidget::configChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}

// SIGNAL 6
void Fooyin::Filters::FilterWidget::requestHeaderMenu(Fooyin::AutoHeaderView * _t1, const QPoint & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1, _t2);
}

// SIGNAL 7
void Fooyin::Filters::FilterWidget::requestContextMenu(const QPoint & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1);
}

// SIGNAL 8
void Fooyin::Filters::FilterWidget::requestEditConnections()
{
    QMetaObject::activate(this, &staticMetaObject, 8, nullptr);
}

// SIGNAL 9
void Fooyin::Filters::FilterWidget::searchTextChanged(const QString & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 9, nullptr, _t1);
}
QT_WARNING_POP
