/****************************************************************************
** Meta object code from reading C++ file 'audioinput.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../fooyin/include/core/engine/audioinput.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'audioinput.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin12AudioDecoderE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::AudioDecoder::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin12AudioDecoderE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::AudioDecoder",
        "DecoderOptions",
        "DecoderFlag",
        "None",
        "NoSeeking",
        "NoLooping",
        "NoInfiniteLooping",
        "UpdateTracks",
        "PlaybackHints",
        "PlaybackHint",
        "NoHints",
        "RepeatTrackEnabled"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // flag 'DecoderOptions'
        QtMocHelpers::EnumData<DecoderOptions>(1, 2, QMC::EnumIsFlag).add({
            {    3, DecoderFlag::None },
            {    4, DecoderFlag::NoSeeking },
            {    5, DecoderFlag::NoLooping },
            {    6, DecoderFlag::NoInfiniteLooping },
            {    7, DecoderFlag::UpdateTracks },
        }),
        // flag 'PlaybackHints'
        QtMocHelpers::EnumData<PlaybackHints>(8, 9, QMC::EnumIsFlag).add({
            {   10, PlaybackHint::NoHints },
            {   11, PlaybackHint::RepeatTrackEnabled },
        }),
    };
    return QtMocHelpers::metaObjectData<AudioDecoder, qt_meta_tag_ZN6Fooyin12AudioDecoderE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::AudioDecoder::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin12AudioDecoderE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin12AudioDecoderE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin12AudioDecoderE_t>.metaTypes,
    nullptr
} };

namespace {
struct qt_meta_tag_ZN6Fooyin11AudioReaderE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::AudioReader::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin11AudioReaderE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::AudioReader",
        "WriteOptions",
        "WriteFlag",
        "None",
        "Metadata",
        "Rating",
        "Playcount",
        "PreserveTimestamps"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // flag 'WriteOptions'
        QtMocHelpers::EnumData<WriteOptions>(1, 2, QMC::EnumIsFlag).add({
            {    3, WriteFlag::None },
            {    4, WriteFlag::Metadata },
            {    5, WriteFlag::Rating },
            {    6, WriteFlag::Playcount },
            {    7, WriteFlag::PreserveTimestamps },
        }),
    };
    return QtMocHelpers::metaObjectData<AudioReader, qt_meta_tag_ZN6Fooyin11AudioReaderE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::AudioReader::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin11AudioReaderE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin11AudioReaderE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin11AudioReaderE_t>.metaTypes,
    nullptr
} };

QT_WARNING_POP
