/****************************************************************************
** Meta object code from reading C++ file 'trackdatabasemanager.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../fooyin/src/core/library/trackdatabasemanager.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'trackdatabasemanager.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN6Fooyin20TrackDatabaseManagerE_t {};
} // unnamed namespace

template <> constexpr inline auto Fooyin::TrackDatabaseManager::qt_create_metaobjectdata<qt_meta_tag_ZN6Fooyin20TrackDatabaseManagerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "Fooyin::TrackDatabaseManager",
        "gotTracks",
        "",
        "Fooyin::TrackList",
        "tracks",
        "updatedTracks",
        "availabilityChecked",
        "updatedTracksStats",
        "removedTracks",
        "trackWriteCompleted",
        "operationId",
        "failed",
        "cancelled",
        "trackCoverWriteCompleted",
        "tracksDeleted",
        "unavailableTracksRemoved",
        "getAllTracks",
        "checkTrackAvailability",
        "updateTracks",
        "write",
        "updateTrackStats",
        "track",
        "AudioReader::WriteOptions",
        "writeOptions",
        "writeCovers",
        "Fooyin::TrackCoverData",
        "deleteTracks",
        "TrackList",
        "removeUnavailbleTracks",
        "cleanupTracks",
        "std::stop_token",
        "stopToken"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'gotTracks'
        QtMocHelpers::SignalData<void(const Fooyin::TrackList &)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'updatedTracks'
        QtMocHelpers::SignalData<void(const Fooyin::TrackList &)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'availabilityChecked'
        QtMocHelpers::SignalData<void(const Fooyin::TrackList &)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'updatedTracksStats'
        QtMocHelpers::SignalData<void(const Fooyin::TrackList &)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'removedTracks'
        QtMocHelpers::SignalData<void(const Fooyin::TrackList &)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'trackWriteCompleted'
        QtMocHelpers::SignalData<void(int, const Fooyin::TrackList &, int, bool)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 10 }, { 0x80000000 | 3, 4 }, { QMetaType::Int, 11 }, { QMetaType::Bool, 12 },
        }}),
        // Signal 'trackCoverWriteCompleted'
        QtMocHelpers::SignalData<void(int, const Fooyin::TrackList &, int, bool)>(13, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 10 }, { 0x80000000 | 3, 4 }, { QMetaType::Int, 11 }, { QMetaType::Bool, 12 },
        }}),
        // Signal 'tracksDeleted'
        QtMocHelpers::SignalData<void(int, const Fooyin::TrackList &, int, bool)>(14, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 10 }, { 0x80000000 | 3, 4 }, { QMetaType::Int, 11 }, { QMetaType::Bool, 12 },
        }}),
        // Signal 'unavailableTracksRemoved'
        QtMocHelpers::SignalData<void(int, const Fooyin::TrackList &, int, bool)>(15, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 10 }, { 0x80000000 | 3, 4 }, { QMetaType::Int, 11 }, { QMetaType::Bool, 12 },
        }}),
        // Slot 'getAllTracks'
        QtMocHelpers::SlotData<void()>(16, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'checkTrackAvailability'
        QtMocHelpers::SlotData<void(const Fooyin::TrackList &)>(17, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Slot 'updateTracks'
        QtMocHelpers::SlotData<void(const Fooyin::TrackList &, bool)>(18, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 }, { QMetaType::Bool, 19 },
        }}),
        // Slot 'updateTrackStats'
        QtMocHelpers::SlotData<void(const Fooyin::TrackList &, AudioReader::WriteOptions)>(20, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 21 }, { 0x80000000 | 22, 23 },
        }}),
        // Slot 'writeCovers'
        QtMocHelpers::SlotData<void(const Fooyin::TrackCoverData &)>(24, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 25, 4 },
        }}),
        // Slot 'deleteTracks'
        QtMocHelpers::SlotData<void(const TrackList &)>(26, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 27, 4 },
        }}),
        // Slot 'removeUnavailbleTracks'
        QtMocHelpers::SlotData<void(const TrackList &)>(28, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 27, 4 },
        }}),
        // Slot 'cleanupTracks'
        QtMocHelpers::SlotData<void()>(29, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'updateTracks'
        QtMocHelpers::SlotData<void(const Fooyin::TrackList &, bool, int, std::stop_token)>(18, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 }, { QMetaType::Bool, 19 }, { QMetaType::Int, 10 }, { 0x80000000 | 30, 31 },
        }}),
        // Slot 'writeCovers'
        QtMocHelpers::SlotData<void(const Fooyin::TrackCoverData &, int, std::stop_token)>(24, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 25, 4 }, { QMetaType::Int, 10 }, { 0x80000000 | 30, 31 },
        }}),
        // Slot 'deleteTracks'
        QtMocHelpers::SlotData<void(const TrackList &, int, std::stop_token)>(26, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 27, 4 }, { QMetaType::Int, 10 }, { 0x80000000 | 30, 31 },
        }}),
        // Slot 'removeUnavailbleTracks'
        QtMocHelpers::SlotData<void(const TrackList &, int, std::stop_token)>(28, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 27, 4 }, { QMetaType::Int, 10 }, { 0x80000000 | 30, 31 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<TrackDatabaseManager, qt_meta_tag_ZN6Fooyin20TrackDatabaseManagerE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject Fooyin::TrackDatabaseManager::staticMetaObject = { {
    QMetaObject::SuperData::link<Worker::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin20TrackDatabaseManagerE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin20TrackDatabaseManagerE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN6Fooyin20TrackDatabaseManagerE_t>.metaTypes,
    nullptr
} };

void Fooyin::TrackDatabaseManager::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<TrackDatabaseManager *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->gotTracks((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1]))); break;
        case 1: _t->updatedTracks((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1]))); break;
        case 2: _t->availabilityChecked((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1]))); break;
        case 3: _t->updatedTracksStats((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1]))); break;
        case 4: _t->removedTracks((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1]))); break;
        case 5: _t->trackWriteCompleted((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[4]))); break;
        case 6: _t->trackCoverWriteCompleted((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[4]))); break;
        case 7: _t->tracksDeleted((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[4]))); break;
        case 8: _t->unavailableTracksRemoved((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[4]))); break;
        case 9: _t->getAllTracks(); break;
        case 10: _t->checkTrackAvailability((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1]))); break;
        case 11: _t->updateTracks((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 12: _t->updateTrackStats((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<AudioReader::WriteOptions>>(_a[2]))); break;
        case 13: _t->writeCovers((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackCoverData>>(_a[1]))); break;
        case 14: _t->deleteTracks((*reinterpret_cast<std::add_pointer_t<TrackList>>(_a[1]))); break;
        case 15: _t->removeUnavailbleTracks((*reinterpret_cast<std::add_pointer_t<TrackList>>(_a[1]))); break;
        case 16: _t->cleanupTracks(); break;
        case 17: _t->updateTracks((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackList>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<std::stop_token>>(_a[4]))); break;
        case 18: _t->writeCovers((*reinterpret_cast<std::add_pointer_t<Fooyin::TrackCoverData>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<std::stop_token>>(_a[3]))); break;
        case 19: _t->deleteTracks((*reinterpret_cast<std::add_pointer_t<TrackList>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<std::stop_token>>(_a[3]))); break;
        case 20: _t->removeUnavailbleTracks((*reinterpret_cast<std::add_pointer_t<TrackList>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<std::stop_token>>(_a[3]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 0:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 1:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 2:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 3:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 4:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 5:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 6:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 7:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 8:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 10:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 11:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 12:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        case 17:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< Fooyin::TrackList >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (TrackDatabaseManager::*)(const Fooyin::TrackList & )>(_a, &TrackDatabaseManager::gotTracks, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (TrackDatabaseManager::*)(const Fooyin::TrackList & )>(_a, &TrackDatabaseManager::updatedTracks, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (TrackDatabaseManager::*)(const Fooyin::TrackList & )>(_a, &TrackDatabaseManager::availabilityChecked, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (TrackDatabaseManager::*)(const Fooyin::TrackList & )>(_a, &TrackDatabaseManager::updatedTracksStats, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (TrackDatabaseManager::*)(const Fooyin::TrackList & )>(_a, &TrackDatabaseManager::removedTracks, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (TrackDatabaseManager::*)(int , const Fooyin::TrackList & , int , bool )>(_a, &TrackDatabaseManager::trackWriteCompleted, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (TrackDatabaseManager::*)(int , const Fooyin::TrackList & , int , bool )>(_a, &TrackDatabaseManager::trackCoverWriteCompleted, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (TrackDatabaseManager::*)(int , const Fooyin::TrackList & , int , bool )>(_a, &TrackDatabaseManager::tracksDeleted, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (TrackDatabaseManager::*)(int , const Fooyin::TrackList & , int , bool )>(_a, &TrackDatabaseManager::unavailableTracksRemoved, 8))
            return;
    }
}

const QMetaObject *Fooyin::TrackDatabaseManager::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *Fooyin::TrackDatabaseManager::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN6Fooyin20TrackDatabaseManagerE_t>.strings))
        return static_cast<void*>(this);
    return Worker::qt_metacast(_clname);
}

int Fooyin::TrackDatabaseManager::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = Worker::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 21)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 21;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 21)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 21;
    }
    return _id;
}

// SIGNAL 0
void Fooyin::TrackDatabaseManager::gotTracks(const Fooyin::TrackList & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void Fooyin::TrackDatabaseManager::updatedTracks(const Fooyin::TrackList & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void Fooyin::TrackDatabaseManager::availabilityChecked(const Fooyin::TrackList & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void Fooyin::TrackDatabaseManager::updatedTracksStats(const Fooyin::TrackList & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void Fooyin::TrackDatabaseManager::removedTracks(const Fooyin::TrackList & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void Fooyin::TrackDatabaseManager::trackWriteCompleted(int _t1, const Fooyin::TrackList & _t2, int _t3, bool _t4)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1, _t2, _t3, _t4);
}

// SIGNAL 6
void Fooyin::TrackDatabaseManager::trackCoverWriteCompleted(int _t1, const Fooyin::TrackList & _t2, int _t3, bool _t4)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1, _t2, _t3, _t4);
}

// SIGNAL 7
void Fooyin::TrackDatabaseManager::tracksDeleted(int _t1, const Fooyin::TrackList & _t2, int _t3, bool _t4)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1, _t2, _t3, _t4);
}

// SIGNAL 8
void Fooyin::TrackDatabaseManager::unavailableTracksRemoved(int _t1, const Fooyin::TrackList & _t2, int _t3, bool _t4)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 8, nullptr, _t1, _t2, _t3, _t4);
}
QT_WARNING_POP
